(begin
(define cadr
  (lambda (l)
    (car (cdr l))))

(define equal?
  (lambda (obj1 obj2)
    (if (and (pair? obj1) (pair? obj2))
	(and (equal? (car obj1) (car obj2))
	     (equal? (cdr obj1) (cdr obj2)))
	(if (or (pair? obj1) (pair? obj2)) #f
	    (eqv? obj1 obj2)))))

(define memgeneric
  (lambda (obj l pred)
    (if (null? l) #f
	(if (pred obj (car l)) l
	    (memgeneric obj (cdr l) pred)))))

(define memq
  (lambda (obj l) (memgeneric obj l eq?)))
(define memv
  (lambda (obj l) (memgeneric obj l eqv?)))
(define member
  (lambda (obj l) (memgeneric obj l equal?)))

(define filter 
  (lambda (f l) 
    (if (null? l) '() 
	(if (f (car l)) 
	    (cons (car l) (filter f (cdr l)))
	    (filter f (cdr l))))))

(define map-over-single-list
  (lambda (p l)
    (if (null? l) '()
	(cons (p (car l)) 
	      (map-over-single-list p (cdr l))))))

(define map
  (lambda (proc . lists)
    (if (memq '() lists) '()
	(cons
	 (apply proc 
		(map-over-single-list car lists))
	 (apply map 
		(cons proc (map-over-single-list cdr lists)))))))

;; intersection of 2 lists
(define intersection
  (lambda (l1 l2 . accum)
    (let ((acc (if (null? accum) '() (car accum))))
      (if (null? l1)
          acc
          (intersection
           (cdr l1) l2
           (if (not (member (car l1) l2))
               acc
               (cons (car l1) acc)))))))

;; a list of (taxname type) matching homezip:
(define *home-taxes* '())

;; str is a comma-delimited list of tax names:
(define add-home-tax
  (lambda (str type)
    (set! *home-taxes* (cons (list str type) *home-taxes*))))

;; a list of (taxname type) matching workzip:
(define *work-taxes* '())

;; str is a comma-delimited list of tax names:
(define add-work-tax
  (lambda (str type)
    (set! *work-taxes* (cons (list str type) *work-taxes*))))

;; used in rules:
(define resident?
  (lambda ()
    (member *current-tax* (map car *home-taxes*))))

(define company-resident?
  (lambda ()
    (member *current-tax* (map car *work-taxes*))))

(define worktax
  (lambda ()
    (not
     (null?
      (filter 
       (lambda (tax)
         (equal?
          "S"
          (cadr tax)))
       *work-taxes*)))))

(define restax
  (lambda ()
    (not
     (null?
      (filter 
       (lambda (tax)
         (equal?
          "S"
          (cadr tax)))
       *home-taxes*)))))

;; traditionally, zip would be a zip code, but we have
;; (homezip) and (workzip) return *home-taxes* and
;; *work-taxes*.  Any intersection returns #t
(define within
  (lambda (zip . states)
    (not (null? (intersection zip states)))))

;; now return *home-taxes*/*work-taxes* rather than a ZIP code:
(define homezip
  (lambda () (map car *home-taxes*)))

(define workzip
  (lambda () (map car *work-taxes*))))
