<?php
// This is the TaxControls PHP table distribution.
// Copyright 1997-2024 by Caliber Technology LLC.
// All rights reserved.

// WARNING
// When writing your application, DO NOT assume that the internal workings of this
// file will remain unchanged from one version to the next!  You should only use the
// functions as described in the TaxControls help reference:
//      https://www.taxupdate.com/help/php/reference.html

$table_version = "2023-12-16v";

include_once('scheme.php');
$zips_db = false;  // sqlite PDO required only if taxNamesForZip is called
$unemployment_rate = 0.05;    // set this!

if (!function_exists('cal_days_in_month'))
{
	function cal_days_in_month($calendar, $month, $year)
	{
		return date('t', mktime(0, 0, 0, $month, 1, $year));
	}
}

if (!defined('CAL_GREGORIAN'))
    define('CAL_GREGORIAN', 1);

date_default_timezone_set ('UTC');
function compute($taxname,$earnings,$ytdearnings=0,$filingstatus=0,$exemptions=0,
                 $stateexemptions=0,$payperiods=12,$paydate=0,$misc=0,$aux=0,
                 $ytdtax=0){
  global $taxlist;
  $d = $paydate==0 ? time() : $paydate;
  $remaining = payperiodsRemaining($payperiods, $d);
  $emp = new emp($earnings,$ytdearnings,$filingstatus,$exemptions,
                  $stateexemptions,$misc,$aux,$payperiods,$d,$ytdtax,$remaining);
  if (array_key_exists($taxname, $taxlist)){
    $tax = $taxlist[$taxname];
    $val = max(0, $tax->compute($emp) / $payperiods);
    $res = round($val, ($tax->round ? 0 : 2));
    return $res;
  } else
    return 0.0;
}

function taxable_wages($taxname,$earnings,$ytdearnings=0,$filingstatus=0,$exemptions=0,
                 $stateexemptions=0,$payperiods=12,$paydate=0,$misc=0,$aux=0,
                 $ytdtax=0){
  global $taxlist;
  $d = $paydate==0 ? time() : $paydate;
  $remaining = payperiodsRemaining($payperiods, $d);
  $emp = new emp($earnings,$ytdearnings,$filingstatus,$exemptions,
                  $stateexemptions,$misc,$aux,1,$d,$ytdtax,$remaining);  // annual number returned

  $wb = $taxname;
  if ($taxname == 'FICA')
	 return $earnings;
  elseif ($taxname == 'FICA Med Additional'){
	 $base = 200000;
	 $wb1 ='FICA Med Additional Wage Base'; 
	 if (array_key_exists($wb1, $taxlist)){
	     $tax = $taxlist[$wb1];
	     $base = $tax->compute($emp);
	 }
	 $res = max(0, min($earnings, $ytdearnings + $earnings - $base));
	 return $res;
  } elseif ($taxname == 'FICA SS')
	 $wb = 'FICA Wage Base';
  else
	 $wb .= ' Wage Base';

  if (array_key_exists($wb, $taxlist)){
    $tax = $taxlist[$wb];
    $base = $tax->compute($emp);
    $res = min($earnings, max(0, $base - $ytdearnings));
    return $res;
  } else
    return $earnings;
}

// filing status values:
$filingSingle = 0;
$filingWidow =  1;
$filingHeadOfHousehold = 2;
$filingMarried = 3;
$filingSeparate = 4;
$filingBothEmployed = 5;

class Tax {
	public $name;
	public $type;
	public $round = false;
	public $withheld = false;
	public $title = false;
	public $state_exemption_instructions = false;
	public $miscellaneous_instructions = false;
	public $auxiliary_instructions = false;
	public $homerule = false;
	public $workrule = false;
	public $stateexemption_tag = false;
	public $miscellaneous_tag = false;
	public $auxiliary_tag = false;
	function compute($emp){
	  return (0.0);
	}
}	

// You can call this to get an array of tax names based on an employee's
// work and home ZIP codes.  Federal taxes such as 'FIT', 'FICA SS' and 
// 'FICA Med' apply regardless of ZIP Code, so they are not included in
// the results.
// Taxes have (scheme) rules which get evaluated here.  This allows us
// to consider reciprocal agreements, for example.
function taxNamesForZip($workzip, $homezip){
  global $scheme_loaded, $scheme_base, $zips_db, $taxlist;
  if (!$scheme_loaded){
    scheme_init();
    scheme_evaluate($scheme_base);
    $zips_db = new PDO('sqlite:zips.db');
    $scheme_loaded = true;
  }
  $workdata = zipToNames($workzip);
  $homedata = zipToNames($homezip);

  scheme_evaluate("(set! *work-taxes* '())");
  scheme_evaluate("(set! *home-taxes* '())");
  $data = array();

  // add all taxes before evaluating rules:
  foreach($workdata as $name){
    $tax = $taxlist[$name];
    scheme_evaluate("(add-work-tax \"$name\" \"$tax->type\")");
  }
  foreach($homedata as $name){
    $tax = $taxlist[$name];
    scheme_evaluate("(add-home-tax \"$name\" \"$tax->type\")");
  }

  foreach($workdata as $name){
    $tax = $taxlist[$name];
    if ($tax->type=="P")
      continue;
    $rule = $tax->workrule;
    if (strstr($rule, "(use ")){
      $use = extract_from_quotes($rule);
      $rule = getrule(1, $use);
    }
    scheme_evaluate("(define *current-tax* \"$name\")");
    if (!$rule || $rule=="" || scheme_evaluate_bool($rule))
      $data[] = $name;
  }

  foreach($homedata as $name){
    $tax = $taxlist[$name];
    if ($tax->type=="P")
      continue;
    $rule = $tax->homerule;
    if (strstr($rule, "(use ")){
      $use = extract_from_quotes($rule);
      $rule = getrule(0, $use);
    }
    scheme_evaluate("(define *current-tax* \"$name\")");
    if ($rule && $rule!="" && scheme_evaluate_bool($rule))
      $data[] = $name;
  }

  return $data;
}

function getSSLPage($url) {
  $arrContext = array(
	'ssl' => array(
	    'verify_peer' => false,
	    'verify_peer_name' => false));
  return file_get_contents($url, false, stream_context_create($arrContext));
}

function taxQuery($work, $home, $serviceDomain, $nexus = "") {
  $parse = parse_url($serviceDomain);
  $domain = 'https://' . $parse['host'] . ':' . $parse['port'];
  $query = 'work_addr=' . urlencode($work) . '&home_addr=' . urlencode($home);
  if ($nexus != ""){
  	$query .= '&nexus=' . urlencode($nexus);
  }
  $url = $domain . '/taxes.php?' . $query;
  $resp = json_decode(getSSLPage($url));
  if ($resp->status == 'ok'){
    return $resp->value;
  } else {
    throw new Exception($resp->value);
  }
}

class emp {
  public $earnings;
  public $ytdearnings;
  public $filingstatus;
  public $exemptions;
  public $stateexemptions;
  public $misc;
  public $aux;
  public $payperiods;
  public $paydate;
  public $ytdtax;
  public $periods_remaining;
  function __construct($earnings, $ytdearnings, $filingstatus, $exemptions,
               $stateexemptions, $misc, $aux, $payperiods, $paydate, $ytdtax, $periods_remaining){
    $this->earnings = $earnings * $payperiods;
    $this->ytdearnings = $ytdearnings;
    $this->filingstatus = $filingstatus;
    $this->exemptions = $exemptions;
    $this->payperiods = $payperiods;
    $this->stateexemptions = $stateexemptions;
    $this->misc = $misc;
    $this->aux = $aux;
    $this->paydate = $paydate;
    $this->ytdtax = $ytdtax;
    $this->periods_remaining = $periods_remaining;
    return $this;
  }
}

function tableLookup($input, $ar){
  $len = count($ar);
  for ($i=0; $i < $len; $i++){
    $cur = $ar[$i];
    if ($cur[0] > $input){
      if ($i==0)
	return 0;
      else {
	$ndx = $ar[$i-1];
	return $ndx[2] + $ndx[1] * ($input - $ndx[0]);
      }
    }
  }
  $ndx = $ar[$len-1];
  return $ndx[2] + $ndx[1] * ($input - $ndx[0]);
}

function isTrue($amt) { if (! $amt) return False; else return abs($amt) > .00001; }

$scheme_loaded = false;

function zipToNames($zip){
  global $zips_db;
  $stmt = $zips_db->prepare('SELECT name FROM zips INNER JOIN names ON name_id=id WHERE zip=:zip');
  $stmt->bindParam(':zip', $zip, PDO::PARAM_INT);
  $stmt->execute();
  return $stmt->fetchAll(PDO::FETCH_COLUMN);
}

function getrule($iswork, $name){
  global $taxlist;
  $tax = $taxlist[$name];
  if ($iswork)
    return $tax->workrule;
  else
    return $tax->homerule;
}

// gets the portion of str between the quotes;
function extract_from_quotes($str){
  $start = strstr($str, '"');
  if (!$start)
    return $str;
  $end = strrchr($start, '"');
  if (!$end)
    return $str;

  return substr($start, 1, -strlen($end));
}

function payperiodsRemaining($date, $periodsperyear){
  $year = date('Y', $date);
  $mon = date('m', $date);
  $day = date('d', $date);

  switch ($periodsperyear){
    case 1: return 1;
    case 2: return $mon <= 6 ? 2 : 1;
    case 4:
      if ($mon <= 3) return 4;
      else if ($mon <= 6) return 3;
      else if ($mon <= 9) return 2;
      else return 1;
    case 12: return 13 - $mon;
    case 24:
      $months = 12 - $mon;
      $ndays = cal_days_in_month(CAL_GREGORIAN, $mon, $year);
      if ($day < $ndays/2)
        return $months * 2 + 2;
      else
        return $months * 2 + 1;
    case 26:
      $rem = daysLeftInYear($day, $mon, $year);
      return 1 + intval($rem / 14);
    case 52:
      $rem = daysLeftInYear($day, $mon, $year);
      return 1 + intval($rem / 7);
    default:
      $rem = daysLeftInYear($day, $mon, $year);
      return 1 + intval($rem * $periodsperyear / 360);
  }
}

function daysLeftInYear($day, $mon, $year){
   $ndays = cal_days_in_month(CAL_GREGORIAN, $mon, $year) - $day;
   for ($i = $mon+1; $i < 13; $i++){
     $ndays = $ndays + cal_days_in_month(CAL_GREGORIAN, $i, $year);
   }
   return $ndays;
}

function taxList(){
	global $taxlist;
	return $taxlist;
}

$scheme_base = "(begin(define cadr  (lambda (l)    (car (cdr l))))(define equal?  (lambda (obj1 obj2)    (if (and (pair? obj1) (pair? obj2))\t(and (equal? (car obj1) (car obj2))\t     (equal? (cdr obj1) (cdr obj2)))\t(if (or (pair? obj1) (pair? obj2)) #f\t    (eqv? obj1 obj2)))))(define memgeneric  (lambda (obj l pred)    (if (null? l) #f\t(if (pred obj (car l)) l\t    (memgeneric obj (cdr l) pred)))))(define memq  (lambda (obj l) (memgeneric obj l eq?)))(define memv  (lambda (obj l) (memgeneric obj l eqv?)))(define member  (lambda (obj l) (memgeneric obj l equal?)))(define filter   (lambda (f l)     (if (null? l) '() \t(if (f (car l)) \t    (cons (car l) (filter f (cdr l)))\t    (filter f (cdr l))))))(define map-over-single-list  (lambda (p l)    (if (null? l) '()\t(cons (p (car l)) \t      (map-over-single-list p (cdr l))))))(define map  (lambda (proc . lists)    (if (memq '() lists) '()\t(cons\t (apply proc \t\t(map-over-single-list car lists))\t (apply map \t\t(cons proc (map-over-single-list cdr lists)))))))(define intersection  (lambda (l1 l2 . accum)    (let ((acc (if (null? accum) '() (car accum))))      (if (null? l1)          acc          (intersection           (cdr l1) l2           (if (not (member (car l1) l2))               acc               (cons (car l1) acc)))))))(define *home-taxes* '())(define add-home-tax  (lambda (str type)    (set! *home-taxes* (cons (list str type) *home-taxes*))))(define *work-taxes* '())(define add-work-tax  (lambda (str type)    (set! *work-taxes* (cons (list str type) *work-taxes*))))(define resident?  (lambda ()    (member *current-tax* (map car *home-taxes*))))(define company-resident?  (lambda ()    (member *current-tax* (map car *work-taxes*))))(define worktax  (lambda ()    (not     (null?      (filter        (lambda (tax)         (equal?          \"S\"          (cadr tax)))       *work-taxes*)))))(define restax  (lambda ()    (not     (null?      (filter        (lambda (tax)         (equal?          \"S\"          (cadr tax)))       *home-taxes*)))))(define within  (lambda (zip . states)    (not (null? (intersection zip states)))))(define homezip  (lambda () (map car *home-taxes*)))(define workzip  (lambda () (map car *work-taxes*))))";
class _ak extends Tax { 
	public $name = "AK";
	public $type = "S";
	public $title = "Alaska";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ak_unemployment extends Tax { 
	public $name = "AK Unemployment";
	public $type = "O";
	public $title = "Alaska Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["AK Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ak_unemployment_employee extends Tax { 
	public $name = "AK Unemployment Employee";
	public $type = "O";
	public $withheld = true;
	public $title = "Alaska Unemployment Tax (Employee)";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["AK Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0056)),0);
	 }
}
class _ak_unemployment_wage_base extends Tax { 
	public $name = "AK Unemployment Wage Base";
	public $type = "P";
	public $title = "AK Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 47100;
	 }
}
class _al extends Tax { 
	public $name = "AL";
	public $type = "S";
	public $withheld = true;
	public $title = "Alabama Income Tax";
	public $state_exemption_instructions = "Enter dependents other than spouse from form A-4 line 4.";
	public $miscellaneous_instructions = "Enter 0 if not claiming personal exemption (Form A-4 line 1).\nEnter 1 if claiming personal exemption (Line 2 'S' or Line 3 'M' or 'H').\nEnter -1 (minus 1) if married but claiming only personal exemption for employee (Line 2 'MS').";
	public $auxiliary_instructions = "Enter Federal Income Tax withheld this pay period.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Dependents";
	public $miscellaneous_tag = "Personal exemption";
	public $auxiliary_tag = "Federal Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3))||($emp->filingstatus==(4))||(($emp->misc<(1))))?(tableLookup(max(0.0,$taxlist["AL Taxable Income"]->compute($emp)),array(array(0,0.02,0),array(500,0.04,10),array(3000,0.05,110)))):(tableLookup(max(0.0,$taxlist["AL Taxable Income"]->compute($emp)),array(array(0,0.02,0),array(1000,0.04,20),array(6000,0.05,220)))));
	 }
}
class _al_attalla extends Tax { 
	public $name = "AL Attalla";
	public $type = "L";
	public $withheld = true;
	public $title = "Attalla Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _al_auburn extends Tax { 
	public $name = "AL Auburn";
	public $type = "L";
	public $withheld = true;
	public $title = "Auburn Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_bear_creek extends Tax { 
	public $name = "AL Bear Creek";
	public $type = "L";
	public $withheld = true;
	public $title = "Bear Creek Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_beaverton extends Tax { 
	public $name = "AL Beaverton";
	public $type = "L";
	public $withheld = true;
	public $title = "Beaverton Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_bessemer extends Tax { 
	public $name = "AL Bessemer";
	public $type = "L";
	public $withheld = true;
	public $title = "Bessemer Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_birmingham extends Tax { 
	public $name = "AL Birmingham";
	public $type = "L";
	public $withheld = true;
	public $title = "Birmingham Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_brilliant extends Tax { 
	public $name = "AL Brilliant";
	public $type = "L";
	public $withheld = true;
	public $title = "Brilliant Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_fairfield extends Tax { 
	public $name = "AL Fairfield";
	public $type = "L";
	public $withheld = true;
	public $title = "Fairfield Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_gadsden extends Tax { 
	public $name = "AL Gadsden";
	public $type = "L";
	public $withheld = true;
	public $title = "Gadsden Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _al_glencoe extends Tax { 
	public $name = "AL Glencoe";
	public $type = "L";
	public $withheld = true;
	public $title = "Glencoe Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _al_goodwater extends Tax { 
	public $name = "AL Goodwater";
	public $type = "L";
	public $withheld = true;
	public $title = "Goodwater Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((20000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0075));
	 }
}
class _al_guin extends Tax { 
	public $name = "AL Guin";
	public $type = "L";
	public $withheld = true;
	public $title = "Guin Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_hackleburg extends Tax { 
	public $name = "AL Hackleburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Hackleburg Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_haleyville extends Tax { 
	public $name = "AL Haleyville";
	public $type = "L";
	public $withheld = true;
	public $title = "Haleyville Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_hamilton extends Tax { 
	public $name = "AL Hamilton";
	public $type = "L";
	public $withheld = true;
	public $title = "Hamilton Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_hobson extends Tax { 
	public $name = "AL Hobson";
	public $type = "L";
	public $withheld = true;
	public $title = "Hobson Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _al_irondale extends Tax { 
	public $name = "AL Irondale";
	public $type = "L";
	public $withheld = true;
	public $title = "Irondale Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _al_leeds extends Tax { 
	public $name = "AL Leeds";
	public $type = "L";
	public $withheld = true;
	public $title = "Leeds Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_lynn extends Tax { 
	public $name = "AL Lynn";
	public $type = "L";
	public $withheld = true;
	public $title = "Lynn Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_macon_co_ extends Tax { 
	public $name = "AL Macon Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Macon Co. Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_midfield extends Tax { 
	public $name = "AL Midfield";
	public $type = "L";
	public $withheld = true;
	public $title = "Midfield Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_mosses extends Tax { 
	public $name = "AL Mosses";
	public $type = "L";
	public $withheld = true;
	public $title = "Mosses Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_mountain_brook extends Tax { 
	public $name = "AL Mountain Brook";
	public $type = "L";
	public $withheld = true;
	public $title = "Mountain Brook Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _al_opelika extends Tax { 
	public $name = "AL Opelika";
	public $type = "L";
	public $withheld = true;
	public $title = "Opelika Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _al_rainbow_city extends Tax { 
	public $name = "AL Rainbow City";
	public $type = "L";
	public $withheld = true;
	public $title = "Rainbow City Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _al_red_bay extends Tax { 
	public $name = "AL Red Bay";
	public $type = "L";
	public $withheld = true;
	public $title = "Red Bay Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _al_shorter extends Tax { 
	public $name = "AL Shorter";
	public $type = "L";
	public $withheld = true;
	public $title = "Shorter Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_southside extends Tax { 
	public $name = "AL Southside";
	public $type = "L";
	public $withheld = true;
	public $title = "Southside Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _al_sulligent extends Tax { 
	public $name = "AL Sulligent";
	public $type = "L";
	public $withheld = true;
	public $title = "Sulligent Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _al_tarrant extends Tax { 
	public $name = "AL Tarrant";
	public $type = "L";
	public $withheld = true;
	public $title = "Tarrant Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _al_taxable_income extends Tax { 
	public $name = "AL Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "AL Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (((($emp->earnings-((isTrue(($emp->filingstatus<(1))||($emp->filingstatus==(2))||($emp->misc==(0)))?(max((3000-((((max(($emp->earnings-(25999)),0)+(499.99))/(500))*(25)))),2500)):((isTrue($emp->filingstatus==(1))?(max((5200-((((max(($emp->earnings-(25999)),0)+(499.99))/(500))*(135)))),2500)):((isTrue(($emp->filingstatus>(3)))?(max((4250-((((max(($emp->earnings-(12999)),0)+(249.99))/(250))*(88)))),2500)):(max((8500-((((max(($emp->earnings-(25999)),0)+(499.99))/(500))*(175)))),5000)))))))))-((isTrue($emp->aux)?(($emp->aux*($emp->payperiods))):(max(0.0,$taxlist["FIT"]->compute($emp))))))-((isTrue($emp->misc==(0))?(0):((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(2))||($emp->filingstatus==(4))||(($emp->misc<(0))))?(1500):(3000))))))-(($emp->stateexemptions*((isTrue(($emp->earnings<(50000)))?(1000):((isTrue(($emp->earnings<(100000)))?(500):(300))))))));
	 }
}
class _al_tuskegee extends Tax { 
	public $name = "AL Tuskegee";
	public $type = "L";
	public $withheld = true;
	public $title = "Tuskegee Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.03));
	 }
}
class _al_unemployment extends Tax { 
	public $name = "AL Unemployment";
	public $type = "O";
	public $title = "Alabama Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["AL Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _al_unemployment_wage_base extends Tax { 
	public $name = "AL Unemployment Wage Base";
	public $type = "P";
	public $title = "AL Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 8000;
	 }
}
class _al_wadley extends Tax { 
	public $name = "AL Wadley";
	public $type = "L";
	public $withheld = true;
	public $title = "Wadley Alabama License Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ar extends Tax { 
	public $name = "AR";
	public $type = "S";
	public $withheld = true;
	public $title = "Arkansas Income Tax";
	public $miscellaneous_instructions = "Set to 1 for the Inflationary Relief credit from Oct through Dec 2022";
	public $homerule = "(and (not (worktax)) (company-resident?) (not (within (homezip) \"AR Texarkana\")))";
	public $workrule = "(and (not (and (within (workzip) \"AR Texarkana\") (within (homezip) \"TX Texarkana\"))) (not (within (homezip) \"AR Texarkana\")))";
	public $miscellaneous_tag = "Relief credit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (((tableLookup((isTrue((max(0.0,$taxlist["AR Taxable Income"]->compute($emp))<(91301)))?((((max(0.0,$taxlist["AR Taxable Income"]->compute($emp))/(100))*(100))+(50))):(max(0.0,$taxlist["AR Taxable Income"]->compute($emp)))),array(array(0,0,0),array(5100,0.02,0.02),array(10300,0.03,104.03),array(14700,0.034,236.03),array(24300,0.047,562.45),array(87001,0.047,3509.45),array(87101,0.047,3524.15),array(87201,0.047,3538.85),array(87301,0.047,3553.55),array(87401,0.047,3568.25),array(87501,0.047,3582.95),array(87601,0.047,3597.65),array(87701,0.047,3612.35),array(87801,0.047,3627.05),array(87901,0.047,3641.75),array(88001,0.047,3656.45),array(88101,0.047,3671.15),array(88201,0.047,3685.85),array(88301,0.047,3700.55),array(88401,0.047,3715.25),array(88501,0.047,3729.95),array(88601,0.047,3744.65),array(88701,0.047,3759.35),array(88801,0.047,3774.05),array(88901,0.047,3788.75),array(89001,0.047,3803.45),array(89101,0.047,3818.15),array(89201,0.047,3832.85),array(89301,0.047,3847.55),array(89401,0.047,3862.25),array(89501,0.047,3876.95),array(89601,0.047,3891.65),array(89701,0.047,3906.35),array(89801,0.047,3921.05),array(89901,0.047,3935.75),array(90001,0.047,3950.45),array(90101,0.047,3965.15),array(90201,0.047,3979.85),array(90301,0.047,3994.55),array(90401,0.047,4009.25),array(90501,0.047,4023.95),array(90601,0.047,4038.65),array(90701,0.047,4053.35),array(90801,0.047,4068.05),array(90901,0.047,4082.75),array(91001,0.047,4097.45),array(91101,0.047,4112.15),array(91201,0.047,4126.85),array(91301,0.047,4141.55)))+(0.5))-(max(0.0,$taxlist["AR temp credit"]->compute($emp))))-(($emp->stateexemptions*(29))));
	 }
}
class _ar_taxable_income extends Tax { 
	public $name = "AR Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "AR Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(2270));
	 }
}
class _ar_texarkana extends Tax { 
	public $name = "AR Texarkana";
	public $type = "P";
	public $withheld = true;
	public $title = "used for zip code lookup only";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ar_unemployment extends Tax { 
	public $name = "AR Unemployment";
	public $type = "O";
	public $title = "Arkansas Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["AR Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ar_unemployment_wage_base extends Tax { 
	public $name = "AR Unemployment Wage Base";
	public $type = "P";
	public $title = "AR Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 7000;
	 }
}
class _ar_temp_credit extends Tax { 
	public $name = "AR temp credit";
	public $type = "O";
	public $withheld = true;
	public $title = "Arkansas temporary credit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc&&(($emp->earnings<(87000))))?(600):(0));
	 }
}
class _az extends Tax { 
	public $name = "AZ";
	public $type = "S";
	public $withheld = true;
	public $title = "Arizona Income Tax";
	public $miscellaneous_instructions = "Enter withholding rate from form A-4.  For example, enter 3.6% as .036 (if 0 or empty, 2.0% will be used). Enter a negative value for 0%.";
	public $workrule = "(not (within (homezip) \"CA\" \"IN\" \"OR\" \"VA\"))";
	public $miscellaneous_tag = "Withholding rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*((isTrue($emp->misc)?($emp->misc):(0.02))));
	 }
}
class _az_job_training_tax extends Tax { 
	public $name = "AZ Job Training Tax";
	public $type = "O";
	public $title = "Dropped";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _az_unemployment extends Tax { 
	public $name = "AZ Unemployment";
	public $type = "O";
	public $title = "Arizona Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["AZ Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _az_unemployment_wage_base extends Tax { 
	public $name = "AZ Unemployment Wage Base";
	public $type = "P";
	public $title = "AZ Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 8000;
	 }
}
class _benefits_dependent_care_exclusion extends Tax { 
	public $name = "Benefits Dependent Care Exclusion";
	public $type = "P";
	public $title = "Benefits Dependent Care Exclusion";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->payperiods*((isTrue($emp->filingstatus==(4))?(2500):(5000))));
	 }
}
class _ca extends Tax { 
	public $name = "CA";
	public $type = "S";
	public $withheld = true;
	public $title = "California Income Tax";
	public $state_exemption_instructions = "Enter regular withholding allowances only (not for estimated deductions).";
	public $miscellaneous_instructions = "Enter additional allowances for estimated deductions.";
	public $homerule = "(company-resident?)";
	public $stateexemption_tag = "Regular allowances";
	public $miscellaneous_tag = "Additional allowances";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((isTrue(($emp->earnings>((isTrue(max(0.0,$taxlist["CA Married 2"]->compute($emp)))?(34503):(17252)))))?((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(2))||(($emp->filingstatus>(3))))?(tableLookup(max(0.0,$taxlist["CA Taxable Income"]->compute($emp)),array(array(0,0.011,0),array(10099,0.022,111.09),array(23942,0.044,415.64),array(37788,0.066,1024.86),array(52455,0.088,1992.88),array(66295,0.1023,3210.8),array(338639,0.1133,31071.59),array(406364,0.1243,38744.83),array(677275,0.1353,72419.07),array(1000000,0.1463,116083.76)))):((isTrue($emp->filingstatus==(1)||($emp->filingstatus==(2)))?(tableLookup(max(0.0,$taxlist["CA Taxable Income"]->compute($emp)),array(array(0,0.011,0),array(20212,0.022,222.33),array(47887,0.044,831.18),array(61730,0.066,1440.27),array(76397,0.088,2408.29),array(90240,0.1023,3626.47),array(460547,0.1133,41508.88),array(552658,0.1243,51945.06),array(921095,0.1353,97741.78),array(1000000,0.1463,108417.63)))):(tableLookup(max(0.0,$taxlist["CA Taxable Income"]->compute($emp)),array(array(0,0.011,0),array(20198,0.022,222.18),array(47884,0.044,831.27),array(75576,0.066,2049.72),array(104910,0.088,3985.76),array(132590,0.1023,6421.6),array(677278,0.1133,62143.18),array(812728,0.1243,77489.67),array(1000000,0.1353,100767.58),array(1354550,0.1463,148738.2)))))))):(0))-(($emp->stateexemptions*(154)))),0);
	 }
}
class _ca_disability extends Tax { 
	public $name = "CA Disability";
	public $type = "O";
	public $withheld = true;
	public $title = "California State Disability Insurance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["CA SDI"]->compute($emp));
	 }
}
class _ca_disability_se extends Tax { 
	public $name = "CA Disability SE";
	public $type = "O";
	public $withheld = true;
	public $title = "California Disability Insurance (Self-Employed)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["CA Disability SE Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0256)),0);
	 }
}
class _ca_disability_se_wage_base extends Tax { 
	public $name = "CA Disability SE Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "CA Disability SE Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 31767;
	 }
}
class _ca_disability_wage_base extends Tax { 
	public $name = "CA Disability Wage Base";
	public $type = "P";
	public $title = "CA Disability Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 153164;
	 }
}
class _ca_employment_training_fund extends Tax { 
	public $name = "CA Employment Training Fund";
	public $type = "O";
	public $title = "California Employment Training Fund";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["CA Employment Training Fund Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.001)),0);
	 }
}
class _ca_employment_training_fund_wage_base extends Tax { 
	public $name = "CA Employment Training Fund Wage Base";
	public $type = "P";
	public $title = "CA Employment Training Fund Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["CA Unemployment Wage Base"]->compute($emp));
	 }
}
class _ca_la extends Tax { 
	public $name = "CA LA";
	public $type = "L";
	public $title = "Los Angeles Payroll Expense Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ca_married_2 extends Tax { 
	public $name = "CA Married 2";
	public $type = "P";
	public $withheld = true;
	public $title = "CA Married with 2 or more allowances.  Used 2x in CA";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->filingstatus==(3)||($emp->filingstatus==(4))&&(($emp->stateexemptions>(1)))||($emp->filingstatus==(1)))?(1):(0));
	 }
}
class _ca_sdi extends Tax { 
	public $name = "CA SDI";
	public $type = "O";
	public $withheld = true;
	public $title = "California State Disability Insurance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["CA SDI Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.009)),0);
	 }
}
class _ca_sdi_wage_base extends Tax { 
	public $name = "CA SDI Wage Base";
	public $type = "P";
	public $title = "CA SDI Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["CA Disability Wage Base"]->compute($emp));
	 }
}
class _ca_taxable_income extends Tax { 
	public $name = "CA Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "CA Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->misc*(1000))))-((isTrue(max(0.0,$taxlist["CA Married 2"]->compute($emp)))?(10404):(5202))));
	 }
}
class _ca_unemployment extends Tax { 
	public $name = "CA Unemployment";
	public $type = "O";
	public $title = "California Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["CA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ca_unemployment_wage_base extends Tax { 
	public $name = "CA Unemployment Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "CA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 7000;
	 }
}
class _co extends Tax { 
	public $name = "CO";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Colorado Income Tax";
	public $state_exemption_instructions = "Enter Withholding Allowance from form DR 0004 Line 2.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Withholding allowance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-((isTrue($emp->stateexemptions)?($emp->stateexemptions):((isTrue($emp->filingstatus==(3))?(9000):(4500))))))*(0.044));
	 }
}
class _co_aurora extends Tax { 
	public $name = "CO Aurora";
	public $type = "L";
	public $withheld = true;
	public $title = "Aurora Colorado Occupational Privilege Tax";
	public $miscellaneous_instructions = "Enter the month-to-date Aurora taxes withheld for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(249.99)))?(((2-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_aurora_employer extends Tax { 
	public $name = "CO Aurora Employer";
	public $type = "L";
	public $title = "Aurora Colorado Occupational Privilege Tax (Employer Contribution)";
	public $miscellaneous_instructions = "Enter the month-to-date Aurora taxes paid for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(249.99)))?(((2-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_denver extends Tax { 
	public $name = "CO Denver";
	public $type = "L";
	public $withheld = true;
	public $title = "Denver Colorado Occupational Privilege Tax";
	public $miscellaneous_instructions = "Enter the month-to-date Denver taxes withheld for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(499.99)))?(((5.75-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_denver_employer extends Tax { 
	public $name = "CO Denver Employer";
	public $type = "L";
	public $title = "Denver Colorado Occupational Privilege Tax (Employer Contribution)";
	public $miscellaneous_instructions = "Enter the month-to-date Denver taxes paid for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(499.99)))?(((4-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_family_leave extends Tax { 
	public $name = "CO Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "Colorado Family Leave Tax";
	public $auxiliary_instructions = "Enter employee share of premiums (default/maximum is 50%) as a number between 0 and 0.5";
	public $auxiliary_tag = "Employer share";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((min($emp->earnings,max((max(0.0,$taxlist["CO Family Leave Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.009))*((isTrue($emp->aux)?(min($emp->aux,0.5)):(0.5)))),0);
	 }
}
class _co_family_leave_employer extends Tax { 
	public $name = "CO Family Leave Employer";
	public $type = "O";
	public $title = "Colorado Family Leave Employer Tax";
	public $miscellaneous_instructions = "Enter a non-zero value if employer has 9 or more employees.";
	public $auxiliary_instructions = "Enter employer share of premiums (default/minimum is 50%) as a number between 0.5 and 1.0";
	public $miscellaneous_tag = "9+ employees";
	public $auxiliary_tag = "Employer share";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc)?(max(((min($emp->earnings,max((max(0.0,$taxlist["CO Family Leave Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.009))*((isTrue($emp->aux)?(max($emp->aux,0.5)):(0.5)))),0)):(0));
	 }
}
class _co_family_leave_employer_wage_base extends Tax { 
	public $name = "CO Family Leave Employer Wage Base";
	public $type = "P";
	public $title = "employer wage base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["CO Family Leave Wage Base"]->compute($emp));
	 }
}
class _co_family_leave_wage_base extends Tax { 
	public $name = "CO Family Leave Wage Base";
	public $type = "P";
	public $title = "CO Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _co_glendale extends Tax { 
	public $name = "CO Glendale";
	public $type = "L";
	public $withheld = true;
	public $title = "Glendale Colorado Occupational Privilege Tax";
	public $miscellaneous_instructions = "Enter the month-to-date Glendale taxes withheld for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(749.99)))?(((5-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_glendale_employer extends Tax { 
	public $name = "CO Glendale Employer";
	public $type = "L";
	public $title = "Glendale Colorado Occupational Privilege Tax (Employer Contribution)";
	public $miscellaneous_instructions = "Enter the month-to-date Glendale taxes paid for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(749.99)))?(((5-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_greenwood_village extends Tax { 
	public $name = "CO Greenwood Village";
	public $type = "L";
	public $withheld = true;
	public $title = "Greenwood Village Colorado Occupational Privilege Tax";
	public $miscellaneous_instructions = "Enter the month-to-date Greenwood taxes withheld for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(249.99)))?(((2-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_greenwood_village_employer extends Tax { 
	public $name = "CO Greenwood Village Employer";
	public $type = "L";
	public $title = "Greenwood Village Colorado Occupational Privilege Tax (Employer Contribution)";
	public $miscellaneous_instructions = "Enter the month-to-date Greenwood taxes paid for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(249.99)))?(((2-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_sheridan extends Tax { 
	public $name = "CO Sheridan";
	public $type = "L";
	public $withheld = true;
	public $title = "Sheridan Colorado Occupational Privilege Tax";
	public $miscellaneous_instructions = "Enter the month-to-date Sheridan taxes withheld for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(499.99)))?(((3-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_sheridan_employer extends Tax { 
	public $name = "CO Sheridan Employer";
	public $type = "L";
	public $title = "Sheridan Colorado Occupational Privilege Tax (Employer Contribution)";
	public $miscellaneous_instructions = "Enter the month-to-date Sheridan taxes paid for this employee.";
	public $auxiliary_instructions = "Enter the month-to-date earnings (not including current earnings) for this employee.";
	public $miscellaneous_tag = "MTD Taxes";
	public $auxiliary_tag = "MTD Earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+($emp->aux))>(499.99)))?(((3-($emp->misc))*($emp->payperiods))):(0));
	 }
}
class _co_unemployment extends Tax { 
	public $name = "CO Unemployment";
	public $type = "O";
	public $title = "Colorado Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["CO Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _co_unemployment_wage_base extends Tax { 
	public $name = "CO Unemployment Wage Base";
	public $type = "P";
	public $title = "CO Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 20400;
	 }
}
class _ct extends Tax { 
	public $name = "CT";
	public $type = "S";
	public $withheld = true;
	public $title = "Connecticut Income Tax";
	public $miscellaneous_instructions = "Using Form CT-W4 line 1: Enter 1 for category A. Enter 2 for category B. Enter 3 for category C. Enter 4 for category D. Enter 5 for category F.  (Category E is exempt from withholding).";
	public $miscellaneous_tag = "Category";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc==(-1))?(0):(((((isTrue($emp->misc==(1)||($emp->misc==(4))||($emp->misc==(5))||($emp->misc==(0)))?(tableLookup(max(0.0,$taxlist["CT Taxable Income"]->compute($emp)),array(array(0,0.03,0),array(10000.01,0.05,300),array(50000.01,0.055,2300),array(100000.01,0.06,5050),array(200000.01,0.065,11050),array(250000.01,0.069,14300),array(500000.01,0.0699,31550)))):((isTrue($emp->misc==(2))?(tableLookup(max(0.0,$taxlist["CT Taxable Income"]->compute($emp)),array(array(0,0.03,0),array(16000.01,0.05,480),array(80000.01,0.055,3680),array(160000.01,0.06,8080),array(320000.01,0.065,17680),array(400000.01,0.069,22880),array(800000.01,0.0699,50480)))):(tableLookup(max(0.0,$taxlist["CT Taxable Income"]->compute($emp)),array(array(0,0.03,0),array(20000.01,0.05,600),array(100000.01,0.055,4600),array(200000.01,0.06,10100),array(400000.01,0.065,22100),array(500000.01,0.069,28600),array(1000000.01,0.0699,63100)))))))+(max(0.0,$taxlist["CT Recapture"]->compute($emp))))+(max(0.0,$taxlist["CT 3% Phase-Out"]->compute($emp))))*((1-(max(0.0,$taxlist["CT Personal Tax Credits"]->compute($emp))))))));
	 }
}
class _ct_3__phase_out extends Tax { 
	public $name = "CT 3% Phase-Out";
	public $type = "P";
	public $withheld = true;
	public $title = "CT 3% Phase-Out";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc==(2))?(tableLookup($emp->earnings,array(array(0,0,0),array(78500.01,0,32),array(82500.01,0,64),array(86500.01,0,96),array(90500.01,0,128),array(94500.01,0,160),array(98500.01,0,192),array(102500.01,0,224),array(106500.01,0,256),array(110500.01,0,288),array(114500.01,0,320)))):((isTrue($emp->misc==(3))?(tableLookup($emp->earnings,array(array(0,0,0),array(100500.01,0,40),array(105500.01,0,80),array(110500.01,0,120),array(115500.01,0,160),array(120500.01,0,200),array(125500.01,0,240),array(130500.01,0,280),array(135500.01,0,320),array(140500.01,0,360),array(145500.01,0,400)))):((isTrue($emp->misc==(5))?(tableLookup($emp->earnings,array(array(0,0,0),array(56500.01,0,20),array(61500.01,0,40),array(66500.01,0,60),array(71500.01,0,80),array(76500.01,0,100),array(81500.01,0,120),array(86500.01,0,140),array(91500.01,0,160),array(96500.01,0,180),array(101500.01,0,200)))):(tableLookup($emp->earnings,array(array(0,0,0),array(50250.01,0,20),array(52750.01,0,40),array(55250.01,0,60),array(57750.01,0,80),array(60250.01,0,100),array(62750.01,0,120),array(65250.01,0,140),array(67750.01,0,160),array(70250.01,0,180),array(72750.01,0,200)))))))));
	 }
}
class _ct_family_leave extends Tax { 
	public $name = "CT Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "Connecticut Family Leave Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max((max(0.0,$taxlist["CT Family Leave Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.005));
	 }
}
class _ct_family_leave_wage_base extends Tax { 
	public $name = "CT Family Leave Wage Base";
	public $type = "P";
	public $title = "CT Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ct_personal_tax_credits extends Tax { 
	public $name = "CT Personal Tax Credits";
	public $type = "P";
	public $withheld = true;
	public $title = "CT Personal Tax Credits";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc==(1))?(tableLookup($emp->earnings,array(array(12000.01,0,0.75),array(15000.01,0,0.7),array(15500.01,0,0.65),array(16000.01,0,0.6),array(16500.01,0,0.55),array(17000.01,0,0.5),array(17500.01,0,0.45),array(18000.01,0,0.4),array(18500.01,0,0.35),array(20000.01,0,0.3),array(20500.01,0,0.25),array(21000.01,0,0.2),array(21500.01,0,0.15),array(25000.01,0,0.14),array(25500.01,0,0.13),array(26000.01,0,0.12),array(26500.01,0,0.11),array(27000.01,0,0.1),array(48000.01,0,0.09),array(48500.01,0,0.08),array(49000.01,0,0.07),array(49500.01,0,0.06),array(50000.01,0,0.05),array(50500.01,0,0.04),array(51000.01,0,0.03),array(51500.01,0,0.02),array(52000.01,0,0.01),array(52500.01,0,0)))):((isTrue($emp->misc==(2))?(tableLookup($emp->earnings,array(array(19000.01,0,0.75),array(24000.01,0,0.7),array(24500.01,0,0.65),array(25000.01,0,0.6),array(25500.01,0,0.55),array(26500.01,0,0.45),array(27000.01,0,0.4),array(27500.01,0,0.35),array(34000.01,0,0.3),array(34500.01,0,0.25),array(35000.01,0,0.2),array(35500.01,0,0.15),array(44000.01,0,0.14),array(44500.01,0,0.13),array(45000.01,0,0.12),array(45500.01,0,0.11),array(46000.01,0,0.1),array(74000.01,0,0.09),array(74500.01,0,0.08),array(75000.01,0,0.07),array(75500.01,0,0.06),array(76000.01,0,0.05),array(76500.01,0,0.04),array(77000.01,0,0.03),array(77500.01,0,0.02),array(78000.01,0,0.01),array(78500.01,0,0)))):((isTrue($emp->misc==(3))?(tableLookup($emp->earnings,array(array(24000.01,0,0.75),array(30000.01,0,0.7),array(30500.01,0,0.65),array(31000.01,0,0.6),array(31500.01,0,0.55),array(32000.01,0,0.5),array(32500.01,0,0.45),array(33000.01,0,0.4),array(33500.01,0,0.35),array(40000.01,0,0.3),array(40500.01,0,0.25),array(41000.01,0,0.2),array(41500.01,0,0.15),array(50000.01,0,0.14),array(50500.01,0,0.13),array(51000.01,0,0.12),array(51500.01,0,0.11),array(52000.01,0,0.1),array(96000.01,0,0.09),array(96500.01,0,0.08),array(97000.01,0,0.07),array(97500.01,0,0.06),array(98000.01,0,0.05),array(98500.01,0,0.04),array(99000.01,0,0.03),array(99500.01,0,0.02),array(100000.01,0,0.01),array(100500.01,0,0)))):((isTrue($emp->misc==(5))?(tableLookup($emp->earnings,array(array(15000.01,0,0.75),array(18800.01,0,0.7),array(19300.01,0,0.65),array(19800.01,0,0.6),array(20300.01,0,0.55),array(20800.01,0,0.5),array(21300.01,0,0.45),array(21800.01,0,0.4),array(22300.01,0,0.35),array(25000.01,0,0.3),array(25500.01,0,0.25),array(26000.01,0,0.2),array(26500.01,0,0.15),array(31300.01,0,0.14),array(31800.01,0,0.13),array(32300.01,0,0.12),array(32800.01,0,0.11),array(33300.01,0,0.1),array(60000.01,0,0.09),array(60500.01,0,0.08),array(61000.01,0,0.07),array(61500.01,0,0.06),array(62000.01,0,0.05),array(62500.01,0,0.04),array(63000.01,0,0.03),array(63500.01,0,0.02),array(64000.01,0,0.01),array(64500.01,0,0)))):(0))))))));
	 }
}
class _ct_recapture extends Tax { 
	public $name = "CT Recapture";
	public $type = "P";
	public $withheld = true;
	public $title = "CT Recapture";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc==(2))?(max(0.0,$taxlist["CT Recapture B"]->compute($emp))):((isTrue($emp->misc==(3))?(max(0.0,$taxlist["CT Recapture C"]->compute($emp))):(max(0.0,$taxlist["CT Recapture A D F"]->compute($emp))))));
	 }
}
class _ct_recapture_a_d_f extends Tax { 
	public $name = "CT Recapture A D F";
	public $type = "P";
	public $withheld = true;
	public $title = "CT Recapture Withholding Code A, D, or F";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["CT Taxable Income"]->compute($emp)),array(array(0,0,0),array(200000,0,90),array(205000,0,180),array(210000,0,270),array(215000,0,360),array(220000,0,450),array(225000,0,540),array(230000,0,630),array(235000,0,720),array(240000,0,810),array(245000,0,900),array(250000,0,990),array(255000,0,1080),array(260000,0,1170),array(265000,0,1260),array(270000,0,1350),array(275000,0,1440),array(280000,0,1530),array(285000,0,1620),array(290000,0,1710),array(295000,0,1800),array(300000,0,1890),array(305000,0,1980),array(310000,0,2070),array(315000,0,2160),array(320000,0,2250),array(325000,0,2340),array(330000,0,2430),array(335000,0,2520),array(340000,0,2610),array(345000,0,2700),array(500000,0,2750),array(505000,0,2800),array(510000,0,2850),array(515000,0,2900),array(520000,0,2950),array(525000,0,3000),array(530000,0,3050),array(535000,0,3100),array(540000,0,3150)));
	 }
}
class _ct_recapture_b extends Tax { 
	public $name = "CT Recapture B";
	public $type = "P";
	public $withheld = true;
	public $title = "CT Recapture Withholding Code B";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["CT Taxable Income"]->compute($emp)),array(array(0,0,0),array(320000,0,140),array(328000,0,280),array(336000,0,420),array(344000,0,560),array(352000,0,700),array(360000,0,840),array(368000,0,980),array(376000,0,1120),array(384000,0,1260),array(392000,0,1400),array(400000,0,1540),array(408000,0,1680),array(416000,0,1820),array(424000,0,1960),array(432000,0,2100),array(440000,0,2240),array(448000,0,2380),array(456000,0,2520),array(464000,0,1660),array(472000,0,2800),array(480000,0,2940),array(488000,0,3080),array(496000,0,3220),array(504000,0,3360),array(512000,0,3500),array(520000,0,3640),array(528000,0,3780),array(536000,0,3920),array(544000,0,4060),array(552000,0,4200),array(800000,0,4280),array(808000,0,4360),array(816000,0,4440),array(824000,0,4520),array(832000,0,4600),array(840000,0,4680),array(848000,0,4760),array(856000,0,4840),array(864000,0,4920)));
	 }
}
class _ct_recapture_c extends Tax { 
	public $name = "CT Recapture C";
	public $type = "P";
	public $withheld = true;
	public $title = "CT Recapture Withholding Code C";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["CT Taxable Income"]->compute($emp)),array(array(0,0,0),array(400000,0,180),array(410000,0,360),array(420000,0,540),array(430000,0,720),array(440000,0,900),array(450000,0,1080),array(460000,0,1260),array(470000,0,1440),array(480000,0,1620),array(490000,0,1800),array(500000,0,1980),array(510000,0,2160),array(520000,0,2340),array(530000,0,2520),array(540000,0,2700),array(550000,0,2880),array(560000,0,3060),array(570000,0,3240),array(580000,0,3420),array(590000,0,3600),array(600000,0,3780),array(610000,0,3960),array(620000,0,4140),array(630000,0,4320),array(640000,0,4500),array(650000,0,4680),array(660000,0,4860),array(670000,0,5040),array(680000,0,5220),array(690000,0,5400),array(1000000,0,5500),array(1010000,0,5600),array(1020000,0,5700),array(1030000,0,5800),array(1040000,0,5900),array(1050000,0,6000),array(1060000,0,6100),array(1070000,0,6200),array(1080000,0,6300)));
	 }
}
class _ct_taxable_income extends Tax { 
	public $name = "CT Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "CT Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-((isTrue($emp->misc==(1))?(tableLookup($emp->earnings,array(array(0,0,12000),array(24000.01,0,11000),array(25000.01,0,10000),array(26000.01,0,9000),array(27000.01,0,8000),array(28000.01,0,7000),array(29000.01,0,6000),array(30000.01,0,5000),array(31000.01,0,4000),array(32000.01,0,3000),array(33000.01,0,2000),array(34000.01,0,1000),array(35000.01,0,0)))):((isTrue($emp->misc==(2))?(tableLookup($emp->earnings,array(array(0,0,19000),array(38000.01,0,18000),array(39000.01,0,17000),array(40000.01,0,16000),array(41000.01,0,15000),array(42000.01,0,14000),array(43000.01,0,13000),array(44000.01,0,12000),array(45000.01,0,11000),array(46000.01,0,10000),array(47000.01,0,9000),array(48000.01,0,8000),array(49000.01,0,7000),array(50000.01,0,6000),array(51000.01,0,5000),array(52000.01,0,4000),array(53000.01,0,3000),array(54000.01,0,2000),array(55000.01,0,1000),array(56000.01,0,0)))):((isTrue($emp->misc==(3))?(tableLookup($emp->earnings,array(array(0,0,24000),array(48000.01,0,23000),array(49000.01,0,22000),array(50000.01,0,21000),array(51000.01,0,20000),array(52000.01,0,19000),array(53000.01,0,18000),array(54000.01,0,17000),array(55000.01,0,16000),array(56000.01,0,15000),array(57000.01,0,14000),array(58000.01,0,13000),array(59000.01,0,12000),array(60000.01,0,11000),array(61000.01,0,10000),array(62000.01,0,9000),array(63000.01,0,8000),array(64000.01,0,7000),array(65000.01,0,6000),array(66000.01,0,5000),array(67000.01,0,4000),array(68000.01,0,3000),array(69000.01,0,2000),array(70000.01,0,1000),array(71000.01,0,0)))):((isTrue($emp->misc==(5))?(tableLookup($emp->earnings,array(array(0,0,15000),array(30000.01,0,14000),array(31000.01,0,13000),array(32000.01,0,12000),array(33000.01,0,11000),array(34000.01,0,10000),array(35000.01,0,9000),array(36000.01,0,8000),array(37000.01,0,7000),array(38000.01,0,6000),array(39000.01,0,5000),array(40000.01,0,4000),array(41000.01,0,3000),array(42000.01,0,2000),array(43000.01,0,1000),array(44000.01,0,0)))):(0))))))))));
	 }
}
class _ct_unemployment extends Tax { 
	public $name = "CT Unemployment";
	public $type = "O";
	public $title = "Connecticut Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["CT Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ct_unemployment_wage_base extends Tax { 
	public $name = "CT Unemployment Wage Base";
	public $type = "P";
	public $title = "CT Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 15000;
	 }
}
class _dc extends Tax { 
	public $name = "DC";
	public $type = "L";
	public $round = true;
	public $withheld = true;
	public $title = "Washington, D.C. Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"DC\" \"MD\" \"VA\"))";
	public $workrule = "(not  (within (homezip) \"MD\" \"VA\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["DC Taxable Income"]->compute($emp)),array(array(0,0.04,0),array(10000,0.06,400),array(40000,0.065,2200),array(60000,0.085,3500),array(350000,0.0875,28150),array(1000000,0.0895,85025)));
	 }
}
class _dc_family_leave extends Tax { 
	public $name = "DC Family Leave";
	public $type = "O";
	public $title = "District of Columbia Family Leave Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0026));
	 }
}
class _dc_taxable_income extends Tax { 
	public $name = "DC Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "DC Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(4150)),0)));
	 }
}
class _dc_unemployment extends Tax { 
	public $name = "DC Unemployment";
	public $type = "O";
	public $title = "District of Columbia Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["DC Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _dc_unemployment_wage_base extends Tax { 
	public $name = "DC Unemployment Wage Base";
	public $type = "P";
	public $title = "DC Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9000;
	 }
}
class _de extends Tax { 
	public $name = "DE";
	public $type = "S";
	public $withheld = true;
	public $title = "Delaware Income Tax";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (tableLookup(($emp->earnings-((isTrue(($emp->filingstatus<(3)))?(3250):((isTrue($emp->filingstatus==(4))?(3250):(6500)))))),array(array(0,0,0),array(2000,0.022,0),array(5000,0.039,66),array(10000,0.048,261),array(20000,0.052,741),array(25000,0.0555,1001),array(60000,0.066,2943.5)))-(max(($emp->stateexemptions*(110)),0)));
	 }
}
class _de_training_tax_assessment extends Tax { 
	public $name = "DE Training Tax Assessment";
	public $type = "O";
	public $title = "Deleware Training Tax Assessment";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["DE Training Tax Assessment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0015)),0);
	 }
}
class _de_training_tax_assessment_wage_base extends Tax { 
	public $name = "DE Training Tax Assessment Wage Base";
	public $type = "O";
	public $title = "Deleware Training Tax Assessment Wage Base";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["DE Unemployment Wage Base"]->compute($emp));
	 }
}
class _de_unemployment extends Tax { 
	public $name = "DE Unemployment";
	public $type = "O";
	public $title = "Deleware Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["DE Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _de_unemployment_wage_base extends Tax { 
	public $name = "DE Unemployment Wage Base";
	public $type = "P";
	public $title = "DE Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 10500;
	 }
}
class _de_wilmington extends Tax { 
	public $name = "DE Wilmington";
	public $type = "L";
	public $withheld = true;
	public $title = "Wilmington Delaware Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _de_wilmington_employer extends Tax { 
	public $name = "DE Wilmington Employer";
	public $type = "L";
	public $title = "Wilmington Delaware Business License Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 120;
	 }
}
class _fica extends Tax { 
	public $name = "FICA";
	public $type = "P";
	public $withheld = true;
	public $title = "FICA Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["FICA Med"]->compute($emp))+(max(0.0,$taxlist["FICA SS"]->compute($emp))));
	 }
}
class _fica_med extends Tax { 
	public $name = "FICA Med";
	public $type = "F";
	public $withheld = true;
	public $title = "FICA-Medicare Tax";
	public $miscellaneous_instructions = "Enter 1 to compute without Medicare Additional tax.";
	public $miscellaneous_tag = "Without Medicare Additional";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings*(0.0145))+((isTrue(($emp->misc>(0)))?(0):(max(0.0,$taxlist["FICA Med Additional"]->compute($emp))))));
	 }
}
class _fica_med_additional extends Tax { 
	public $name = "FICA Med Additional";
	public $type = "P";
	public $withheld = true;
	public $title = "FICA Medicare Additional Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((($emp->earnings-(max((max(0.0,$taxlist["FICA Med Additional Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0)))*(0.009)),0);
	 }
}
class _fica_med_additional_wage_base extends Tax { 
	public $name = "FICA Med Additional Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "FICA Med Additional Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (200000*($emp->payperiods));
	 }
}
class _fica_med_employer extends Tax { 
	public $name = "FICA Med employer";
	public $type = "F";
	public $title = "FICA-Medicare Tax (Employer Contribution)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0145));
	 }
}
class _fica_ss extends Tax { 
	public $name = "FICA SS";
	public $type = "F";
	public $withheld = true;
	public $title = "FICA-Social Security Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min((min($emp->earnings,(max(0.0,$taxlist["FICA SS Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.062)),max(0.0,$taxlist["FICA SS YTDTax limit"]->compute($emp)));
	 }
}
class _fica_ss_wage_base extends Tax { 
	public $name = "FICA SS Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "FICA SS Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _fica_ss_ytdtax_limit extends Tax { 
	public $name = "FICA SS YTDTax limit";
	public $type = "P";
	public $title = "FICA SS YTDTax limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((max(0.0,$taxlist["FICA Wage Base"]->compute($emp))*(0.062))-(($emp->ytdtax*($emp->payperiods)))),0);
	 }
}
class _fica_ss_employer extends Tax { 
	public $name = "FICA SS employer";
	public $type = "F";
	public $title = "FICA-Social Security Tax (employer)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min((min($emp->earnings,(max(0.0,$taxlist["FICA SS employer Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.062)),max(0.0,$taxlist["FICA SS YTDTax limit"]->compute($emp)));
	 }
}
class _fica_ss_employer_wage_base extends Tax { 
	public $name = "FICA SS employer Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "FICA SS employer Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _fica_wage_base extends Tax { 
	public $name = "FICA Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "FICA Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (160200*($emp->payperiods));
	 }
}
class _fica_employer extends Tax { 
	public $name = "FICA employer";
	public $type = "F";
	public $title = "FICA Tax (Employer Contribution)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["FICA Med employer"]->compute($emp))+(max(0.0,$taxlist["FICA SS employer"]->compute($emp))));
	 }
}
class _fit extends Tax { 
	public $name = "FIT";
	public $type = "F";
	public $withheld = true;
	public $title = "Federal Income Tax";
	public $miscellaneous_instructions = "For supplemental wages, the total YTD supplemental wages prior to this payment.";
	public $auxiliary_instructions = "Enter 1 if employee is a nonresident alien, 0 otherwise.";
	public $miscellaneous_tag = "YTD Supplemental Wages";
	public $auxiliary_tag = "Nonresident Alien";
	function compute($emp){ 
	  global $taxlist;
	  if ($emp->paydate >= mktime(0, 0, 0,  1,  1, 2024))
	    return (isTrue($emp->payperiods)?((isTrue($emp->filingstatus==(3))?(tableLookup(max(0.0,$taxlist["FIT Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(16300,0.1,0),array(39500,0.12,2320),array(110600,0.22,10852),array(217350,0.24,34337),array(400200,0.32,78221),array(503750,0.35,111357),array(747500,0.37,196669.5)))):((isTrue($emp->filingstatus==(1))?(tableLookup(max(0.0,$taxlist["FIT Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(13300,0.1,0),array(29850,0.12,1655),array(76400,0.22,7241),array(113800,0.24,15469),array(205250,0.32,37417),array(257000,0.35,53977),array(622650,0.37,181954.5)))):(tableLookup(max(0.0,$taxlist["FIT Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(6000,0.1,0),array(17600,0.12,1160),array(53150,0.22,5426),array(106525,0.24,17168.5),array(197950,0.32,39110.5),array(249725,0.35,55678.5),array(615350,0.37,183647.25)))))))):(0));
	  else return (isTrue($emp->payperiods)?((isTrue($emp->filingstatus==(3))?(tableLookup(max(0.0,$taxlist["FIT Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(14800,0.1,0),array(36800,0.12,2200),array(104250,0.22,10294),array(205550,0.24,32580),array(379000,0.32,74208),array(477300,0.35,105664),array(708550,0.37,186601.5)))):(tableLookup(max(0.0,$taxlist["FIT Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(5250,0.1,0),array(16250,0.12,1100),array(49975,0.22,5147),array(100625,0.24,16290),array(187350,0.32,37104),array(236500,0.35,52832),array(583375,0.37,174238.25)))))):(0));
	 }
}
class _fit_adjusted_earnings extends Tax { 
	public $name = "FIT Adjusted Earnings";
	public $type = "P";
	public $withheld = true;
	public $title = "FIT Adjusted Earnings";
	function compute($emp){ 
	  global $taxlist;
	  if ($emp->paydate >= mktime(0, 0, 0,  1,  1, 2024))
	    return (($emp->earnings-(($emp->exemptions*(4300))))+((isTrue($emp->aux==(1))?(10300):(0))));
	  else return (($emp->earnings-(($emp->exemptions*(4300))))+((isTrue($emp->aux==(1))?(9550):(0))));
	 }
}
class _fit_supplemental_under_limit extends Tax { 
	public $name = "FIT Supplemental under limit";
	public $type = "P";
	public $withheld = true;
	public $title = "FIT Supplemental under limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min($emp->earnings,max(((1000000*($emp->payperiods))-(($emp->misc*($emp->payperiods)))),0));
	 }
}
class _fl extends Tax { 
	public $name = "FL";
	public $type = "S";
	public $title = "Florida";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _fl_unemployment extends Tax { 
	public $name = "FL Unemployment";
	public $type = "O";
	public $title = "Florida Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["FL Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _fl_unemployment_federal_interest_assessment extends Tax { 
	public $name = "FL Unemployment Federal Interest Assessment";
	public $type = "O";
	public $title = "this tax has expired";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _fl_unemployment_wage_base extends Tax { 
	public $name = "FL Unemployment Wage Base";
	public $type = "P";
	public $title = "FL Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 7000;
	 }
}
class _futa extends Tax { 
	public $name = "FUTA";
	public $type = "O";
	public $title = "Federal Unemployment Tax";
	public $miscellaneous_instructions = "The FUTA table now includes logic for calculation of normal and additional credits for state unemployment taxes.  There are three methods for FUTA tax calculation: \n1. No credit for state unemployment taxes: Enter '0' in Auxiliary. \n2. Normal credit only: Enter amount of state unemployment taxes paid this period in Miscellaneous, and enter '1' in Auxiliary. \n3. Normal plus Additional credits: Enter amount state taxes paid this period in Miscellaneous, and enter '2' in Auxiliary.";
	public $miscellaneous_tag = "Mode";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((max(0.0,$taxlist["FUTA Taxable Wages"]->compute($emp))*(max(0.0,$taxlist["FUTA Rate"]->compute($emp))))-(max(0.0,$taxlist["FUTA Credits"]->compute($emp)))),0);
	 }
}
class _futa_credits extends Tax { 
	public $name = "FUTA Credits";
	public $type = "P";
	public $title = "FUTA Credits";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((isTrue($emp->aux==(1))?(max(0.0,$taxlist["FUTA Normal Credit"]->compute($emp))):(0))+((isTrue($emp->aux==(2))?((max(0.0,$taxlist["FUTA Taxable Wages"]->compute($emp))*(0.054))):(0))));
	 }
}
class _futa_normal_credit extends Tax { 
	public $name = "FUTA Normal Credit";
	public $type = "P";
	public $title = "FUTA Normal Credit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->aux)?(min($emp->misc,(max(0.0,$taxlist["FUTA Taxable Wages"]->compute($emp))*(0.054)))):(0));
	 }
}
class _futa_rate extends Tax { 
	public $name = "FUTA Rate";
	public $type = "P";
	public $title = "FUTA Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->aux==(3))?($emp->misc):(0.06));
	 }
}
class _futa_taxable_wages extends Tax { 
	public $name = "FUTA Taxable Wages";
	public $type = "P";
	public $title = "FUTA Taxable Wages";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min($emp->earnings,max((max(0.0,$taxlist["FUTA Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0));
	 }
}
class _futa_wage_base extends Tax { 
	public $name = "FUTA Wage Base";
	public $type = "P";
	public $title = "FUTA Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (7000*($emp->payperiods));
	 }
}
class _federal_adjusted_earnings extends Tax { 
	public $name = "Federal Adjusted Earnings";
	public $type = "P";
	public $title = "Federal Adjusted Earnings";
	function compute($emp){ 
	  global $taxlist;
	  if ($emp->paydate >= mktime(0, 0, 0,  1,  1, 2024))
	    return max(((($emp->earnings-((isTrue($emp->aux)?(0):((isTrue($emp->filingstatus==(3))?(12900):(8600))))))+($emp->misc))+((isTrue(($emp->stateexemptions<(0)))?(14600):(0)))),0);
	  else return max(((($emp->earnings-((isTrue($emp->aux)?(0):((isTrue($emp->filingstatus==(3))?(12900):(8600))))))+($emp->misc))+((isTrue(($emp->stateexemptions<(0)))?(13850):(0)))),0);
	 }
}
class _federal_income_tax extends Tax { 
	public $name = "Federal Income Tax";
	public $type = "F";
	public $withheld = true;
	public $title = "Federal Income Tax";
	public $state_exemption_instructions = "Enter -1 for Nonresident Alien";
	public $miscellaneous_instructions = "Enter Form W-4 line 4(a) minus line 4(b)";
	public $auxiliary_instructions = "Enter 1 if Form W-4 Step 2, line (c) is checked, 0 otherwise.";
	public $stateexemption_tag = "Nonresident Alien";
	public $miscellaneous_tag = "Other Adjustments";
	public $auxiliary_tag = "Multiple Jobs";
	function compute($emp){ 
	  global $taxlist;
	  if ($emp->paydate >= mktime(0, 0, 0,  1,  1, 2024))
	    return ((isTrue($emp->aux)?((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(4)))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(7300,0.1,0),array(13100,0.12,580),array(30875,0.22,2713),array(57563,0.24,8584.25),array(103275,0.32,19555.25),array(129163,0.35,27839.25),array(311975,0.37,91823.63)))):((isTrue($emp->filingstatus==(1))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(10950,0.1,0),array(19225,0.12,827.5),array(42500,0.22,3620.5),array(61200,0.24,7734.5),array(106925,0.32,18708.5),array(132800,0.35,26988.5),array(315625,0.37,90977.25)))):(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(14600,0.1,0),array(26200,0.12,1160),array(61750,0.22,5426),array(115125,0.24,17168.5),array(206550,0.32,39110.5),array(258325,0.35,55678.5),array(380200,0.37,98334.75)))))))):((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(4)))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(6000,0.1,0),array(17600,0.12,1160),array(53150,0.22,5426),array(106525,0.24,17168.5),array(197950,0.32,39110.5),array(249725,0.35,55678.5),array(615350,0.37,183647.25)))):((isTrue($emp->filingstatus==(1))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(13300,0.1,0),array(29850,0.12,1655),array(76400,0.22,7241),array(113800,0.24,15469),array(205250,0.32,37417),array(257000,0.35,53977),array(622650,0.37,181954.5)))):(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(16300,0.1,0),array(39500,0.12,2320),array(110600,0.22,10852),array(217350,0.24,34337),array(400200,0.32,78221),array(503750,0.35,111357),array(747500,0.37,196669.5)))))))))-($emp->exemptions));
	  else return ((isTrue($emp->aux)?((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(4)))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(6925,0.1,0),array(12425,0.12,550),array(29288,0.22,2573.5),array(54613,0.24,8145),array(97975,0.32,18552),array(122550,0.35,26416),array(295988,0.37,87119.13)))):((isTrue($emp->filingstatus==(1))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(10400,0.1,0),array(18250,0.12,785),array(40325,0.22,3434),array(58075,0.24,7339),array(101450,0.32,17749),array(126025,0.35,25613),array(299450,0.37,86311.75)))):(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(13850,0.1,0),array(24850,0.12,1100),array(58575,0.22,5147),array(109225,0.24,16290),array(195950,0.32,37104),array(245100,0.35,52832),array(360725,0.37,93300.75)))))))):((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(4)))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(5250,0.1,0),array(16250,0.12,1100),array(49975,0.22,5147),array(100625,0.24,16290),array(187350,0.32,37104),array(236500,0.35,52832),array(583375,0.37,174238.25)))):((isTrue($emp->filingstatus==(1))?(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(12200,0.1,0),array(27900,0.12,1570),array(72050,0.22,6868),array(107550,0.24,14678),array(194300,0.32,35498),array(243450,0.35,51226),array(590300,0.37,172623.5)))):(tableLookup(max(0.0,$taxlist["Federal Adjusted Earnings"]->compute($emp)),array(array(0,0,0),array(14800,0.1,0),array(36800,0.12,2200),array(104250,0.22,10294),array(205550,0.24,32580),array(379000,0.32,74208),array(477300,0.35,105664),array(708550,0.37,186601.5)))))))))-($emp->exemptions));
	 }
}
class _federal_supplemental_under_limit extends Tax { 
	public $name = "Federal Supplemental under limit";
	public $type = "P";
	public $withheld = true;
	public $title = "Federal Income Tax Supplemental under limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min($emp->earnings,max(((1000000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0));
	 }
}
class _federal_tax_lien_exemption extends Tax { 
	public $name = "Federal Tax Lien Exemption";
	public $type = "O";
	public $title = "Federal Tax Lien Exemption";
	public $miscellaneous_instructions = "Enter the number of additional exemptions from Parts 3, 4, and 5 of the levy.";
	public $miscellaneous_tag = "Additional Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (((tableLookup($emp->filingstatus,array(array(0,0,13000),array(1,0,18650),array(3,0,24800),array(4,0,12400)))*($emp->exemptions))+(4200))+(($emp->misc*((isTrue(($emp->filingstatus<(2)))?(1650):(1300))))));
	 }
}
class _ga extends Tax { 
	public $name = "GA";
	public $type = "S";
	public $withheld = true;
	public $title = "Georgia Income Tax";
	public $state_exemption_instructions = "Enter number of dependents + additional allowances claimed from form G-4 line 4 + line 5.  Do not include employee or spouse in this number.";
	public $miscellaneous_instructions = "Enter 1 or 2 to drop one or both personal allowance claims.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Dependents";
	public $miscellaneous_tag = "Drop Allowances";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->filingstatus==(0))?(tableLookup(max(0.0,$taxlist["GA Taxable Income"]->compute($emp)),array(array(0,0.01,0),array(750,0.02,7.5),array(2250,0.03,37.5),array(3750,0.04,82.5),array(5250,0.05,142.5),array(7000,0.0575,230)))):((isTrue($emp->filingstatus==(3)||($emp->filingstatus==(1))||($emp->filingstatus==(2)))?(tableLookup(max(0.0,$taxlist["GA Taxable Income"]->compute($emp)),array(array(0,0.01,0),array(1000,0.02,10),array(3000,0.03,50),array(5000,0.04,110),array(7000,0.05,190),array(10000,0.0575,340)))):(tableLookup(max(0.0,$taxlist["GA Taxable Income"]->compute($emp)),array(array(0,0.01,0),array(500,0.02,5),array(1500,0.03,25),array(2500,0.04,55),array(3500,0.05,95),array(5000,0.0575,170)))))));
	 }
}
class _ga_taxable_income extends Tax { 
	public $name = "GA Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "GA Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((($emp->earnings-((isTrue($emp->filingstatus==(3))?(7100):((isTrue(($emp->filingstatus<(3)))?(5400):(3550))))))-((isTrue($emp->filingstatus==(3))?((isTrue($emp->misc==(0))?(7400):((isTrue(($emp->misc>(1)))?(0):(3700))))):((isTrue($emp->misc==(0))?((isTrue(($emp->filingstatus<(3)))?(2700):(3700))):(0))))))-(max(($emp->stateexemptions*(3000)),0)));
	 }
}
class _ga_unemployment extends Tax { 
	public $name = "GA Unemployment";
	public $type = "O";
	public $title = "Georgia Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["GA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ga_unemployment_wage_base extends Tax { 
	public $name = "GA Unemployment Wage Base";
	public $type = "P";
	public $title = "GA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9500;
	 }
}
class _hi extends Tax { 
	public $name = "HI";
	public $type = "S";
	public $withheld = true;
	public $title = "Hawaii State Income Tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(0)))?(0):((isTrue(($emp->filingstatus<(2)))?(tableLookup(max(0.0,$taxlist["HI Taxable Income"]->compute($emp)),array(array(0,0.014,0),array(2400,0.032,34),array(4800,0.055,110),array(9600,0.064,374),array(14400,0.068,682),array(19200,0.072,1008),array(24000,0.076,1354),array(36000,0.079,2266)))):(tableLookup(max(0.0,$taxlist["HI Taxable Income"]->compute($emp)),array(array(0,0.014,0),array(4800,0.032,67),array(9600,0.055,221),array(19200,0.064,749),array(28800,0.068,1363),array(38400,0.072,2016),array(48000,0.076,2707),array(72000,0.079,4531)))))));
	 }
}
class _hi_disability extends Tax { 
	public $name = "HI Disability";
	public $type = "O";
	public $withheld = true;
	public $title = "Hawaii Disability Insurance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min(min($emp->earnings,(max(0.0,$taxlist["HI Disability Wage Limit"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods))))),max(0.0,$taxlist["HI Disability Wage Limit"]->compute($emp)))*(0.005));
	 }
}
class _hi_disability_wage_limit extends Tax { 
	public $name = "HI Disability Wage Limit";
	public $type = "P";
	public $withheld = true;
	public $title = "HI Disability Wage Limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 68560.96;
	 }
}
class _hi_eta extends Tax { 
	public $name = "HI ETA";
	public $type = "O";
	public $title = "Hawaii Employment and Training Fund Assessment";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["HI Taxable Income"]->compute($emp))*(0.0001));
	 }
}
class _hi_taxable_income extends Tax { 
	public $name = "HI Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "HI Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(1144)),0)));
	 }
}
class _hi_unemployment extends Tax { 
	public $name = "HI Unemployment";
	public $type = "O";
	public $title = "Hawaii Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["HI Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _hi_unemployment_wage_base extends Tax { 
	public $name = "HI Unemployment Wage Base";
	public $type = "P";
	public $title = "HI Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 56700;
	 }
}
class _ia extends Tax { 
	public $name = "IA";
	public $type = "S";
	public $withheld = true;
	public $title = "Iowa Income Tax";
	public $auxiliary_instructions = "Enter Federal Income Tax withheld this pay period.";
	public $homerule = "(and (company-resident?) (within (workzip) \"IL\"))";
	public $workrule = "(not (within (homezip) \"IL\"))";
	public $auxiliary_tag = "Federal Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (tableLookup(($emp->earnings-((isTrue(($emp->stateexemptions<(2)))?(13850):(27700)))),array(array(0,0.044,0),array(4800,0.0482,211.2),array(24000,0.057,1136.64),array(50000,0.06,2618.64)))-(max(($emp->stateexemptions*(40)),0)));
	 }
}
class _ia_unemployment extends Tax { 
	public $name = "IA Unemployment";
	public $type = "O";
	public $title = "Iowa Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["IA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ia_unemployment_wage_base extends Tax { 
	public $name = "IA Unemployment Wage Base";
	public $type = "P";
	public $title = "IA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 36100;
	 }
}
class _id extends Tax { 
	public $name = "ID";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Idaho Income Tax";
	public $state_exemption_instructions = "Enter number from form ID W-4 line 1.";
	public $stateexemption_tag = "Withholding Allowance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["ID Taxable Income"]->compute($emp)),array(array(0,0,0),array(13850,0.058,0)))):(tableLookup(max(0.0,$taxlist["ID Taxable Income"]->compute($emp)),array(array(0,0,0),array(27700,0.058,0)))));
	 }
}
class _id_taxable_income extends Tax { 
	public $name = "ID Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "ID Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(3534)),0)));
	 }
}
class _id_unemployment extends Tax { 
	public $name = "ID Unemployment";
	public $type = "O";
	public $title = "Idaho Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["ID Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _id_unemployment_wage_base extends Tax { 
	public $name = "ID Unemployment Wage Base";
	public $type = "P";
	public $title = "ID Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 49900;
	 }
}
class _il extends Tax { 
	public $name = "IL";
	public $type = "S";
	public $withheld = true;
	public $title = "Illinois Income Tax";
	public $miscellaneous_instructions = "Enter exemptions from form IL-W-4, line 2.";
	public $homerule = "(and (company-resident?) (within (workzip) \"WI\" \"IA\" \"KY\" \"MI\"))";
	public $workrule = "(not (within (homezip) \"WI\" \"IA\" \"KY\" \"MI\"))";
	public $miscellaneous_tag = "IL W-4 line 2";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max((max(($emp->earnings-(($emp->stateexemptions*(2425)))),0)-(($emp->misc*(1000)))),0)*(0.0495));
	 }
}
class _il_unemployment extends Tax { 
	public $name = "IL Unemployment";
	public $type = "O";
	public $title = "Illinois Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["IL Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _il_unemployment_wage_base extends Tax { 
	public $name = "IL Unemployment Wage Base";
	public $type = "P";
	public $title = "IL Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 13271;
	 }
}
class _in extends Tax { 
	public $name = "IN";
	public $type = "S";
	public $withheld = true;
	public $title = "Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 5.";
	public $miscellaneous_instructions = "Enter number of additional dependent exemptions from form WH-4 line 6.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\"))";
	public $workrule = "(not (within (homezip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0315));
	 }
}
class _in_adams_co_ extends Tax { 
	public $name = "IN Adams Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Adams County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.016));
	 }
}
class _in_adams_co__nonresident extends Tax { 
	public $name = "IN Adams Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Adams County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Adams Co."]->compute($emp));
	 }
}
class _in_allen_co_ extends Tax { 
	public $name = "IN Allen Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Allen County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0148));
	 }
}
class _in_allen_co__nonresident extends Tax { 
	public $name = "IN Allen Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Allen County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Allen Co."]->compute($emp));
	 }
}
class _in_bartholomew_co_ extends Tax { 
	public $name = "IN Bartholomew Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Bartholomew County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0175));
	 }
}
class _in_bartholomew_co__nonresident extends Tax { 
	public $name = "IN Bartholomew Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Bartholomew County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Bartholomew Co."]->compute($emp));
	 }
}
class _in_benton_co_ extends Tax { 
	public $name = "IN Benton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Benton County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0179));
	 }
}
class _in_benton_co__nonresident extends Tax { 
	public $name = "IN Benton Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Benton County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Benton Co."]->compute($emp));
	 }
}
class _in_blackford_co_ extends Tax { 
	public $name = "IN Blackford Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Blackford County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_blackford_co__nonresident extends Tax { 
	public $name = "IN Blackford Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Blackford County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Blackford Co."]->compute($emp));
	 }
}
class _in_boone_co_ extends Tax { 
	public $name = "IN Boone Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Boone County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.017));
	 }
}
class _in_boone_co__nonresident extends Tax { 
	public $name = "IN Boone Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Boone County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Boone Co."]->compute($emp));
	 }
}
class _in_brown_co_ extends Tax { 
	public $name = "IN Brown Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Brown County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02523));
	 }
}
class _in_brown_co__nonresident extends Tax { 
	public $name = "IN Brown Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Brown County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Brown Co."]->compute($emp));
	 }
}
class _in_carroll_co_ extends Tax { 
	public $name = "IN Carroll Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Carroll County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02273));
	 }
}
class _in_carroll_co__nonresident extends Tax { 
	public $name = "IN Carroll Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Carroll County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Carroll Co."]->compute($emp));
	 }
}
class _in_cass_co_ extends Tax { 
	public $name = "IN Cass Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Cass County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0295));
	 }
}
class _in_cass_co__nonresident extends Tax { 
	public $name = "IN Cass Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Cass County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Cass Co."]->compute($emp));
	 }
}
class _in_clark_co_ extends Tax { 
	public $name = "IN Clark Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Clark County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02));
	 }
}
class _in_clark_co__nonresident extends Tax { 
	public $name = "IN Clark Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Clark County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Clark Co."]->compute($emp));
	 }
}
class _in_clay_co_ extends Tax { 
	public $name = "IN Clay Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Clay County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0235));
	 }
}
class _in_clay_co__nonresident extends Tax { 
	public $name = "IN Clay Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Clay County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Clay Co."]->compute($emp));
	 }
}
class _in_clinton_co_ extends Tax { 
	public $name = "IN Clinton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Clinton County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0265));
	 }
}
class _in_clinton_co__nonresident extends Tax { 
	public $name = "IN Clinton Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Clinton County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Clinton Co."]->compute($emp));
	 }
}
class _in_county_tax extends Tax { 
	public $name = "IN County Tax";
	public $type = "L";
	public $withheld = true;
	public $title = "Indiana County Income Tax";
	public $auxiliary_instructions = "Enter rate for tax (CAGIT and CEDIT)";
	public $auxiliary_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*($emp->aux));
	 }
}
class _in_crawford_co_ extends Tax { 
	public $name = "IN Crawford Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Crawford County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.01));
	 }
}
class _in_crawford_co__nonresident extends Tax { 
	public $name = "IN Crawford Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Crawford County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Crawford Co."]->compute($emp));
	 }
}
class _in_daviess_co_ extends Tax { 
	public $name = "IN Daviess Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Daviess County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_daviess_co__nonresident extends Tax { 
	public $name = "IN Daviess Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Daviess County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Daviess Co."]->compute($emp));
	 }
}
class _in_dekalb_co_ extends Tax { 
	public $name = "IN DeKalb Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "DeKalb County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0213));
	 }
}
class _in_dekalb_co__nonresident extends Tax { 
	public $name = "IN DeKalb Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "DeKalb County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN DeKalb Co."]->compute($emp));
	 }
}
class _in_dearborn_co_ extends Tax { 
	public $name = "IN Dearborn Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Dearborn County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.014));
	 }
}
class _in_dearborn_co__nonresident extends Tax { 
	public $name = "IN Dearborn Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Dearborn County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Dearborn Co."]->compute($emp));
	 }
}
class _in_decatur_co_ extends Tax { 
	public $name = "IN Decatur Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Decatur County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.025));
	 }
}
class _in_decatur_co__nonresident extends Tax { 
	public $name = "IN Decatur Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Decatur County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Decatur Co."]->compute($emp));
	 }
}
class _in_delaware_co_ extends Tax { 
	public $name = "IN Delaware Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Delaware County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_delaware_co__nonresident extends Tax { 
	public $name = "IN Delaware Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Delaware County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Delaware Co."]->compute($emp));
	 }
}
class _in_dubois_co_ extends Tax { 
	public $name = "IN Dubois Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Dubois County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.012));
	 }
}
class _in_dubois_co__nonresident extends Tax { 
	public $name = "IN Dubois Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Dubois County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Dubois Co."]->compute($emp));
	 }
}
class _in_elkhart_co_ extends Tax { 
	public $name = "IN Elkhart Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Elkhart County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02));
	 }
}
class _in_elkhart_co__nonresident extends Tax { 
	public $name = "IN Elkhart Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Elkhart County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Elkhart Co."]->compute($emp));
	 }
}
class _in_fayette_co_ extends Tax { 
	public $name = "IN Fayette Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Fayette County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0257));
	 }
}
class _in_fayette_co__nonresident extends Tax { 
	public $name = "IN Fayette Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Fayette County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Fayette Co."]->compute($emp));
	 }
}
class _in_floyd_co_ extends Tax { 
	public $name = "IN Floyd Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Floyd County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0135));
	 }
}
class _in_floyd_co__nonresident extends Tax { 
	public $name = "IN Floyd Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Floyd County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Floyd Co."]->compute($emp));
	 }
}
class _in_fountain_co_ extends Tax { 
	public $name = "IN Fountain Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Fountain County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.021));
	 }
}
class _in_fountain_co__nonresident extends Tax { 
	public $name = "IN Fountain Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Fountain County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Fountain Co."]->compute($emp));
	 }
}
class _in_franklin_co_ extends Tax { 
	public $name = "IN Franklin Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Franklin County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_franklin_co__nonresident extends Tax { 
	public $name = "IN Franklin Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Franklin County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Franklin Co."]->compute($emp));
	 }
}
class _in_fulton_co_ extends Tax { 
	public $name = "IN Fulton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Fulton County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0268));
	 }
}
class _in_fulton_co__nonresident extends Tax { 
	public $name = "IN Fulton Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Fulton County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Fulton Co."]->compute($emp));
	 }
}
class _in_gibson_co_ extends Tax { 
	public $name = "IN Gibson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Gibson County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.009));
	 }
}
class _in_gibson_co__nonresident extends Tax { 
	public $name = "IN Gibson Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Gibson County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Gibson Co."]->compute($emp));
	 }
}
class _in_grant_co_ extends Tax { 
	public $name = "IN Grant Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Grant County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0255));
	 }
}
class _in_grant_co__nonresident extends Tax { 
	public $name = "IN Grant Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Grant County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Grant Co."]->compute($emp));
	 }
}
class _in_greene_co_ extends Tax { 
	public $name = "IN Greene Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Greene County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0215));
	 }
}
class _in_greene_co__nonresident extends Tax { 
	public $name = "IN Greene Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Greene County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Greene Co."]->compute($emp));
	 }
}
class _in_hamilton_co_ extends Tax { 
	public $name = "IN Hamilton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Hamilton County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.011));
	 }
}
class _in_hamilton_co__nonresident extends Tax { 
	public $name = "IN Hamilton Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Hamilton County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Hamilton Co."]->compute($emp));
	 }
}
class _in_hancock_co_ extends Tax { 
	public $name = "IN Hancock Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Hancock County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0194));
	 }
}
class _in_hancock_co__nonresident extends Tax { 
	public $name = "IN Hancock Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Hancock County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Hancock Co."]->compute($emp));
	 }
}
class _in_harrison_co_ extends Tax { 
	public $name = "IN Harrison Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Harrison County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.01));
	 }
}
class _in_harrison_co__nonresident extends Tax { 
	public $name = "IN Harrison Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Harrison County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Harrison Co."]->compute($emp));
	 }
}
class _in_hendricks_co_ extends Tax { 
	public $name = "IN Hendricks Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Hendricks County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.017));
	 }
}
class _in_hendricks_co__nonresident extends Tax { 
	public $name = "IN Hendricks Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Hendricks County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Hendricks Co."]->compute($emp));
	 }
}
class _in_henry_co_ extends Tax { 
	public $name = "IN Henry Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Henry County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.018));
	 }
}
class _in_henry_co__nonresident extends Tax { 
	public $name = "IN Henry Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Henry County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Henry Co."]->compute($emp));
	 }
}
class _in_howard_co_ extends Tax { 
	public $name = "IN Howard Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Howard County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0175));
	 }
}
class _in_howard_co__nonresident extends Tax { 
	public $name = "IN Howard Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Howard County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Howard Co."]->compute($emp));
	 }
}
class _in_huntington_co_ extends Tax { 
	public $name = "IN Huntington Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Huntington County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0195));
	 }
}
class _in_huntington_co__nonresident extends Tax { 
	public $name = "IN Huntington Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Huntington County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Huntington Co."]->compute($emp));
	 }
}
class _in_jackson_co_ extends Tax { 
	public $name = "IN Jackson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jackson County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.021));
	 }
}
class _in_jackson_co__nonresident extends Tax { 
	public $name = "IN Jackson Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jackson County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Jackson Co."]->compute($emp));
	 }
}
class _in_jasper_co_ extends Tax { 
	public $name = "IN Jasper Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jasper County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02864));
	 }
}
class _in_jasper_co__nonresident extends Tax { 
	public $name = "IN Jasper Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jasper County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Jasper Co."]->compute($emp));
	 }
}
class _in_jay_co_ extends Tax { 
	public $name = "IN Jay Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jay County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0295));
	 }
}
class _in_jay_co__nonresident extends Tax { 
	public $name = "IN Jay Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jay County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Jay Co."]->compute($emp));
	 }
}
class _in_jefferson_co_ extends Tax { 
	public $name = "IN Jefferson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jefferson County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.009));
	 }
}
class _in_jefferson_co__nonresident extends Tax { 
	public $name = "IN Jefferson Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jefferson County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Jefferson Co."]->compute($emp));
	 }
}
class _in_jennings_co_ extends Tax { 
	public $name = "IN Jennings Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jennings County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.025));
	 }
}
class _in_jennings_co__nonresident extends Tax { 
	public $name = "IN Jennings Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jennings County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Jennings Co."]->compute($emp));
	 }
}
class _in_johnson_co_ extends Tax { 
	public $name = "IN Johnson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Johnson County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.014));
	 }
}
class _in_johnson_co__nonresident extends Tax { 
	public $name = "IN Johnson Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Johnson County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Johnson Co."]->compute($emp));
	 }
}
class _in_knox_co_ extends Tax { 
	public $name = "IN Knox Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Knox County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.017));
	 }
}
class _in_knox_co__nonresident extends Tax { 
	public $name = "IN Knox Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Knox County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Knox Co."]->compute($emp));
	 }
}
class _in_kosciusko_co_ extends Tax { 
	public $name = "IN Kosciusko Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Kosciusko County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.01));
	 }
}
class _in_kosciusko_co__nonresident extends Tax { 
	public $name = "IN Kosciusko Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Kosciusko County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Kosciusko Co."]->compute($emp));
	 }
}
class _in_lagrange_co_ extends Tax { 
	public $name = "IN LaGrange Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "LaGrange County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0165));
	 }
}
class _in_lagrange_co__nonresident extends Tax { 
	public $name = "IN LaGrange Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "LaGrange County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN LaGrange Co."]->compute($emp));
	 }
}
class _in_laporte_co_ extends Tax { 
	public $name = "IN LaPorte Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "LaPorte County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0145));
	 }
}
class _in_laporte_co__nonresident extends Tax { 
	public $name = "IN LaPorte Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "LaPorte County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN LaPorte Co."]->compute($emp));
	 }
}
class _in_lake_co_ extends Tax { 
	public $name = "IN Lake Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Lake County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_lake_co__nonresident extends Tax { 
	public $name = "IN Lake Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Lake County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Lake Co."]->compute($emp));
	 }
}
class _in_lawrence_co_ extends Tax { 
	public $name = "IN Lawrence Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Lawrence County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0175));
	 }
}
class _in_lawrence_co__nonresident extends Tax { 
	public $name = "IN Lawrence Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Lawrence County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Lawrence Co."]->compute($emp));
	 }
}
class _in_madison_co_ extends Tax { 
	public $name = "IN Madison Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Madison County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0225));
	 }
}
class _in_madison_co__nonresident extends Tax { 
	public $name = "IN Madison Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Madison County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Madison Co."]->compute($emp));
	 }
}
class _in_marion_co_ extends Tax { 
	public $name = "IN Marion Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Marion County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0202));
	 }
}
class _in_marion_co__nonresident extends Tax { 
	public $name = "IN Marion Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Marion County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Marion Co."]->compute($emp));
	 }
}
class _in_marshall_co_ extends Tax { 
	public $name = "IN Marshall Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Marshall County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0125));
	 }
}
class _in_marshall_co__nonresident extends Tax { 
	public $name = "IN Marshall Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Marshall County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Marshall Co."]->compute($emp));
	 }
}
class _in_martin_co_ extends Tax { 
	public $name = "IN Martin Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Martin County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.025));
	 }
}
class _in_martin_co__nonresident extends Tax { 
	public $name = "IN Martin Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Martin County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Martin Co."]->compute($emp));
	 }
}
class _in_miami_co_ extends Tax { 
	public $name = "IN Miami Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Miami County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0254));
	 }
}
class _in_miami_co__nonresident extends Tax { 
	public $name = "IN Miami Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Miami County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Miami Co."]->compute($emp));
	 }
}
class _in_monroe_co_ extends Tax { 
	public $name = "IN Monroe Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Monroe County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02035));
	 }
}
class _in_monroe_co__nonresident extends Tax { 
	public $name = "IN Monroe Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Monroe County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Monroe Co."]->compute($emp));
	 }
}
class _in_montgomery_co_ extends Tax { 
	public $name = "IN Montgomery Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Montgomery County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0265));
	 }
}
class _in_montgomery_co__nonresident extends Tax { 
	public $name = "IN Montgomery Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Montgomery County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Montgomery Co."]->compute($emp));
	 }
}
class _in_morgan_co_ extends Tax { 
	public $name = "IN Morgan Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Morgan County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0272));
	 }
}
class _in_morgan_co__nonresident extends Tax { 
	public $name = "IN Morgan Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Morgan County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Morgan Co."]->compute($emp));
	 }
}
class _in_newton_co_ extends Tax { 
	public $name = "IN Newton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Newton County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.01));
	 }
}
class _in_newton_co__nonresident extends Tax { 
	public $name = "IN Newton Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Newton County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Newton Co."]->compute($emp));
	 }
}
class _in_noble_co_ extends Tax { 
	public $name = "IN Noble Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Noble County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0175));
	 }
}
class _in_noble_co__nonresident extends Tax { 
	public $name = "IN Noble Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Noble County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Noble Co."]->compute($emp));
	 }
}
class _in_ohio_co_ extends Tax { 
	public $name = "IN Ohio Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Ohio County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_ohio_co__nonresident extends Tax { 
	public $name = "IN Ohio Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Ohio County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Ohio Co."]->compute($emp));
	 }
}
class _in_orange_co_ extends Tax { 
	public $name = "IN Orange Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Orange County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0175));
	 }
}
class _in_orange_co__nonresident extends Tax { 
	public $name = "IN Orange Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Orange County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Orange Co."]->compute($emp));
	 }
}
class _in_owen_co_ extends Tax { 
	public $name = "IN Owen Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Owen County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.025));
	 }
}
class _in_owen_co__nonresident extends Tax { 
	public $name = "IN Owen Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Owen County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Owen Co."]->compute($emp));
	 }
}
class _in_parke_co_ extends Tax { 
	public $name = "IN Parke Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Parke County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0265));
	 }
}
class _in_parke_co__nonresident extends Tax { 
	public $name = "IN Parke Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Parke County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Parke Co."]->compute($emp));
	 }
}
class _in_perry_co_ extends Tax { 
	public $name = "IN Perry Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Perry County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.014));
	 }
}
class _in_perry_co__nonresident extends Tax { 
	public $name = "IN Perry Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Perry County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Perry Co."]->compute($emp));
	 }
}
class _in_pike_co_ extends Tax { 
	public $name = "IN Pike Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Pike County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0075));
	 }
}
class _in_pike_co__nonresident extends Tax { 
	public $name = "IN Pike Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Pike County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Pike Co."]->compute($emp));
	 }
}
class _in_porter_co_ extends Tax { 
	public $name = "IN Porter Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Porter County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.005));
	 }
}
class _in_porter_co__nonresident extends Tax { 
	public $name = "IN Porter Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Porter County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Porter Co."]->compute($emp));
	 }
}
class _in_posey_co_ extends Tax { 
	public $name = "IN Posey Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Posey County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0125));
	 }
}
class _in_posey_co__nonresident extends Tax { 
	public $name = "IN Posey Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Posey County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Posey Co."]->compute($emp));
	 }
}
class _in_pulaski_co_ extends Tax { 
	public $name = "IN Pulaski Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Pulaski County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0285));
	 }
}
class _in_pulaski_co__nonresident extends Tax { 
	public $name = "IN Pulaski Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Pulaski County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Pulaski Co."]->compute($emp));
	 }
}
class _in_putnam_co_ extends Tax { 
	public $name = "IN Putnam Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Putnam County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.021));
	 }
}
class _in_putnam_co__nonresident extends Tax { 
	public $name = "IN Putnam Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Putnam County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Putnam Co."]->compute($emp));
	 }
}
class _in_randolph_co_ extends Tax { 
	public $name = "IN Randolph Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Randolph County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.03));
	 }
}
class _in_randolph_co__nonresident extends Tax { 
	public $name = "IN Randolph Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Randolph County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Randolph Co."]->compute($emp));
	 }
}
class _in_ripley_co_ extends Tax { 
	public $name = "IN Ripley Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Ripley County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0138));
	 }
}
class _in_ripley_co__nonresident extends Tax { 
	public $name = "IN Ripley Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Ripley County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Ripley Co."]->compute($emp));
	 }
}
class _in_rush_co_ extends Tax { 
	public $name = "IN Rush Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Rush County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.021));
	 }
}
class _in_rush_co__nonresident extends Tax { 
	public $name = "IN Rush Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Rush County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Rush Co."]->compute($emp));
	 }
}
class _in_scott_co_ extends Tax { 
	public $name = "IN Scott Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Scott County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0216));
	 }
}
class _in_scott_co__nonresident extends Tax { 
	public $name = "IN Scott Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Scott County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Scott Co."]->compute($emp));
	 }
}
class _in_shelby_co_ extends Tax { 
	public $name = "IN Shelby Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Shelby County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.016));
	 }
}
class _in_shelby_co__nonresident extends Tax { 
	public $name = "IN Shelby Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Shelby County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Shelby Co."]->compute($emp));
	 }
}
class _in_spencer_co_ extends Tax { 
	public $name = "IN Spencer Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Spencer County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.008));
	 }
}
class _in_spencer_co__nonresident extends Tax { 
	public $name = "IN Spencer Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Spencer County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Spencer Co."]->compute($emp));
	 }
}
class _in_st__joseph_co_ extends Tax { 
	public $name = "IN St. Joseph Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "St. Joseph County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0175));
	 }
}
class _in_st__joseph_co__nonresident extends Tax { 
	public $name = "IN St. Joseph Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "St. Joseph County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN St. Joseph Co."]->compute($emp));
	 }
}
class _in_starke_co_ extends Tax { 
	public $name = "IN Starke Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Starke County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0171));
	 }
}
class _in_starke_co__nonresident extends Tax { 
	public $name = "IN Starke Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Starke County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Starke Co."]->compute($emp));
	 }
}
class _in_steuben_co_ extends Tax { 
	public $name = "IN Steuben Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Steuben County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0179));
	 }
}
class _in_steuben_co__nonresident extends Tax { 
	public $name = "IN Steuben Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Steuben County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Steuben Co."]->compute($emp));
	 }
}
class _in_sullivan_co_ extends Tax { 
	public $name = "IN Sullivan Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Steuben County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.017));
	 }
}
class _in_sullivan_co__nonresident extends Tax { 
	public $name = "IN Sullivan Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Steuben County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Sullivan Co."]->compute($emp));
	 }
}
class _in_switzerland_co_ extends Tax { 
	public $name = "IN Switzerland Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Switzerland County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0125));
	 }
}
class _in_switzerland_co__nonresident extends Tax { 
	public $name = "IN Switzerland Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Switzerland County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Switzerland Co."]->compute($emp));
	 }
}
class _in_taxable_income extends Tax { 
	public $name = "IN Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "IN Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (((($emp->earnings-(max(($emp->stateexemptions*(1000)),0)))-(max(($emp->misc*(1500)),0)))-(max(($emp->exemptions*(1500)),0)))-(max(($emp->aux*(3000)),0)));
	 }
}
class _in_tippecanoe_co_ extends Tax { 
	public $name = "IN Tippecanoe Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Tippecanoe County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0128));
	 }
}
class _in_tippecanoe_co__nonresident extends Tax { 
	public $name = "IN Tippecanoe Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Tippecanoe County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Tippecanoe Co."]->compute($emp));
	 }
}
class _in_tipton_co_ extends Tax { 
	public $name = "IN Tipton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Tipton County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.026));
	 }
}
class _in_tipton_co__nonresident extends Tax { 
	public $name = "IN Tipton Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Tipton County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Tipton Co."]->compute($emp));
	 }
}
class _in_unemployment extends Tax { 
	public $name = "IN Unemployment";
	public $type = "O";
	public $title = "Indiana Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["IN Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _in_unemployment_wage_base extends Tax { 
	public $name = "IN Unemployment Wage Base";
	public $type = "P";
	public $title = "IN Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9500;
	 }
}
class _in_union_co_ extends Tax { 
	public $name = "IN Union Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Union County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02));
	 }
}
class _in_union_co__nonresident extends Tax { 
	public $name = "IN Union Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Union County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Union Co."]->compute($emp));
	 }
}
class _in_vanderburgh_co_ extends Tax { 
	public $name = "IN Vanderburgh Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Vanderburgh County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0125));
	 }
}
class _in_vanderburgh_co__nonresident extends Tax { 
	public $name = "IN Vanderburgh Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Vanderburgh County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Vanderburgh Co."]->compute($emp));
	 }
}
class _in_vermillion_co_ extends Tax { 
	public $name = "IN Vermillion Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Vermillion County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.015));
	 }
}
class _in_vermillion_co__nonresident extends Tax { 
	public $name = "IN Vermillion Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Vermillion County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Vermillion Co."]->compute($emp));
	 }
}
class _in_vigo_co_ extends Tax { 
	public $name = "IN Vigo Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Vigo County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02));
	 }
}
class _in_vigo_co__nonresident extends Tax { 
	public $name = "IN Vigo Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Vigo County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Vigo Co."]->compute($emp));
	 }
}
class _in_wabash_co_ extends Tax { 
	public $name = "IN Wabash Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Wabash County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.029));
	 }
}
class _in_wabash_co__nonresident extends Tax { 
	public $name = "IN Wabash Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Wabash County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Wabash Co."]->compute($emp));
	 }
}
class _in_warren_co_ extends Tax { 
	public $name = "IN Warren Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Warren County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0212));
	 }
}
class _in_warren_co__nonresident extends Tax { 
	public $name = "IN Warren Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Warren County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Warren Co."]->compute($emp));
	 }
}
class _in_warrick_co_ extends Tax { 
	public $name = "IN Warrick Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Warrick County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.01));
	 }
}
class _in_warrick_co__nonresident extends Tax { 
	public $name = "IN Warrick Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Warrick County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Warrick Co."]->compute($emp));
	 }
}
class _in_washington_co_ extends Tax { 
	public $name = "IN Washington Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Washington County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.02));
	 }
}
class _in_washington_co__nonresident extends Tax { 
	public $name = "IN Washington Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Washington County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Washington Co."]->compute($emp));
	 }
}
class _in_wayne_co_ extends Tax { 
	public $name = "IN Wayne Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Wayne County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0125));
	 }
}
class _in_wayne_co__nonresident extends Tax { 
	public $name = "IN Wayne Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Wayne County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Wayne Co."]->compute($emp));
	 }
}
class _in_wells_co_ extends Tax { 
	public $name = "IN Wells Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Wells County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0145));
	 }
}
class _in_wells_co__nonresident extends Tax { 
	public $name = "IN Wells Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Wells County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Wells Co."]->compute($emp));
	 }
}
class _in_white_co_ extends Tax { 
	public $name = "IN White Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "White County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.0232));
	 }
}
class _in_white_co__nonresident extends Tax { 
	public $name = "IN White Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "White County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN White Co."]->compute($emp));
	 }
}
class _in_whitley_co_ extends Tax { 
	public $name = "IN Whitley Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Whitley County Indiana Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MI\" \"OH\" \"PA\" \"WI\" \"IN\"))";
	public $workrule = "#f";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["IN Taxable Income"]->compute($emp))*(0.01683));
	 }
}
class _in_whitley_co__nonresident extends Tax { 
	public $name = "IN Whitley Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Whitley County Indiana Nonresident Income Tax";
	public $state_exemption_instructions = "Enter number of personal exemptions from form WH-4 line 4.";
	public $miscellaneous_instructions = "Enter number of dependent exemptions from form WH-4 line 5.";
	public $auxiliary_instructions = "Enter number of adopted children claimed on from WH-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"IN\"))";
	public $stateexemption_tag = "Personal exemptions";
	public $miscellaneous_tag = "Dependent exemptions";
	public $auxiliary_tag = "Adoption exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["IN Whitley Co."]->compute($emp));
	 }
}
class _ks extends Tax { 
	public $name = "KS";
	public $type = "S";
	public $withheld = true;
	public $title = "Kansas Income Tax";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["KS Taxable Income"]->compute($emp)),array(array(0,0,0),array(3500,0.031,0),array(18500,0.0525,465),array(33500,0.057,1252.5)))):(tableLookup(max(0.0,$taxlist["KS Taxable Income"]->compute($emp)),array(array(0,0,0),array(8000,0.031,0),array(38000,0.0525,930),array(68000,0.057,2505)))));
	 }
}
class _ks_taxable_income extends Tax { 
	public $name = "KS Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "KS Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(2250)),0)));
	 }
}
class _ks_unemployment extends Tax { 
	public $name = "KS Unemployment";
	public $type = "O";
	public $title = "Kansas Unemployment Tax";
	public $miscellaneous_instructions = "Enter 1 to ignore wage limit (all earnings are taxable)";
	public $homerule = "#f";
	public $miscellaneous_tag = "No wage limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((isTrue($emp->misc)?(($emp->earnings*($unemployment_rate))):((min($emp->earnings,max(((max(0.0,$taxlist["KS Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)))),0);
	 }
}
class _ks_unemployment_wage_base extends Tax { 
	public $name = "KS Unemployment Wage Base";
	public $type = "P";
	public $title = "KS Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc)?(9999999):(14000));
	 }
}
class _ky extends Tax { 
	public $name = "KY";
	public $type = "S";
	public $withheld = true;
	public $title = "Kentucky Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"IL\" \"IN\" \"MI\" \"OH\" \"VA\" \"WV\" \"WI\"))";
	public $workrule = "(not (within (homezip) \"IL\" \"IN\" \"MI\" \"OH\" \"VA\" \"WV\" \"WI\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((($emp->earnings-(2980))*(0.045)),0);
	 }
}
class _ky_adairville extends Tax { 
	public $name = "KY Adairville";
	public $type = "L";
	public $withheld = true;
	public $title = "Adairville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_albany extends Tax { 
	public $name = "KY Albany";
	public $type = "L";
	public $withheld = true;
	public $title = "Albany Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_alexandria extends Tax { 
	public $name = "KY Alexandria";
	public $type = "L";
	public $withheld = true;
	public $title = "Alexandria Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["KY Alexandria Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.015));
	 }
}
class _ky_alexandria_wage_base extends Tax { 
	public $name = "KY Alexandria Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Alexandria Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_allen_co_ extends Tax { 
	public $name = "KY Allen Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Allen Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_ashland extends Tax { 
	public $name = "KY Ashland";
	public $type = "L";
	public $withheld = true;
	public $title = "Ashland Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_auburn extends Tax { 
	public $name = "KY Auburn";
	public $type = "L";
	public $withheld = true;
	public $title = "Auburn Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_augusta extends Tax { 
	public $name = "KY Augusta";
	public $type = "L";
	public $withheld = true;
	public $title = "Augusta Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((72000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0125));
	 }
}
class _ky_ballard_co_ extends Tax { 
	public $name = "KY Ballard Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Ballard Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_barbourville extends Tax { 
	public $name = "KY Barbourville";
	public $type = "L";
	public $withheld = true;
	public $title = "Barbourville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_bardstown extends Tax { 
	public $name = "KY Bardstown";
	public $type = "L";
	public $withheld = true;
	public $title = "Bardstown Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((15000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.01));
	 }
}
class _ky_bardwell extends Tax { 
	public $name = "KY Bardwell";
	public $type = "L";
	public $withheld = true;
	public $title = "Bardwell Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_bath_co_ extends Tax { 
	public $name = "KY Bath Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Bath Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_beattyville extends Tax { 
	public $name = "KY Beattyville";
	public $type = "L";
	public $withheld = true;
	public $title = "Beattyville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_bellevue extends Tax { 
	public $name = "KY Bellevue";
	public $type = "L";
	public $withheld = true;
	public $title = "Bellevue Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.025));
	 }
}
class _ky_benton extends Tax { 
	public $name = "KY Benton";
	public $type = "L";
	public $withheld = true;
	public $title = "Benton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_benton_nonresident extends Tax { 
	public $name = "KY Benton Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Benton Kentucky Nonresident Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_berea extends Tax { 
	public $name = "KY Berea";
	public $type = "L";
	public $withheld = true;
	public $title = "Berea Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_boone_co_ extends Tax { 
	public $name = "KY Boone Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Boone Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["KY Boone Co. Payroll"]->compute($emp))+(max(0.0,$taxlist["KY Boone Co. Mental Health"]->compute($emp))));
	 }
}
class _ky_boone_co__mental_health extends Tax { 
	public $name = "KY Boone Co. Mental Health";
	public $type = "L";
	public $withheld = true;
	public $title = "Boone Co. Kentucky Mental Health Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->ytdtax)?(min(($emp->earnings*(0.0015)),max(((25-($emp->ytdtax))*($emp->payperiods)),0))):((min($emp->earnings,((16666-($emp->ytdearnings))*($emp->payperiods)))*(0.0015))));
	 }
}
class _ky_boone_co__payroll extends Tax { 
	public $name = "KY Boone Co. Payroll";
	public $type = "L";
	public $withheld = true;
	public $title = "Boone Co. Kentucky Payroll Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->ytdtax)?(min(($emp->earnings*(0.008)),max(((510.07-($emp->ytdtax))*($emp->payperiods)),0))):((min($emp->earnings,((63758-($emp->ytdearnings))*($emp->payperiods)))*(0.008))));
	 }
}
class _ky_boone_co__sd extends Tax { 
	public $name = "KY Boone Co. SD";
	public $type = "L";
	public $withheld = true;
	public $title = "Boone Co. Kentucky School District Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_bourbon_co_ extends Tax { 
	public $name = "KY Bourbon Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Bourbon Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_bowling_green extends Tax { 
	public $name = "KY Bowling Green";
	public $type = "L";
	public $withheld = true;
	public $title = "Bowling Green Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0185));
	 }
}
class _ky_boyd_co_ extends Tax { 
	public $name = "KY Boyd Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Boyd Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Ashland\" \"KY Catlettsburg\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_boyle_co_ extends Tax { 
	public $name = "KY Boyle Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Boyle Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_breathitt_co_ extends Tax { 
	public $name = "KY Breathitt Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Breathitt Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_brodhead extends Tax { 
	public $name = "KY Brodhead";
	public $type = "L";
	public $withheld = true;
	public $title = "Brodhead Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_bromley extends Tax { 
	public $name = "KY Bromley";
	public $type = "L";
	public $withheld = true;
	public $title = "Bromley Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_brooksville extends Tax { 
	public $name = "KY Brooksville";
	public $type = "L";
	public $withheld = true;
	public $title = "Brooksville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0175));
	 }
}
class _ky_brownsville extends Tax { 
	public $name = "KY Brownsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Brownsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_burkesville extends Tax { 
	public $name = "KY Burkesville";
	public $type = "L";
	public $withheld = true;
	public $title = "Burkesville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((37500*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.02));
	 }
}
class _ky_butler_co_ extends Tax { 
	public $name = "KY Butler Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Butler Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_cadiz extends Tax { 
	public $name = "KY Cadiz";
	public $type = "L";
	public $withheld = true;
	public $title = "Cadiz Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_caldwell_co_ extends Tax { 
	public $name = "KY Caldwell Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Caldwell Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_calvert extends Tax { 
	public $name = "KY Calvert";
	public $type = "L";
	public $withheld = true;
	public $title = "Calvert Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_camargo extends Tax { 
	public $name = "KY Camargo";
	public $type = "L";
	public $withheld = true;
	public $title = "Camargo Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_campbell_co_ extends Tax { 
	public $name = "KY Campbell Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Campbell Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["KY Campbell Co. Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0105)),0);
	 }
}
class _ky_campbell_co__wage_base extends Tax { 
	public $name = "KY Campbell Co. Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Campbell Co. Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 38667;
	 }
}
class _ky_campbellsville extends Tax { 
	public $name = "KY Campbellsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Campbellsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_caneyville extends Tax { 
	public $name = "KY Caneyville";
	public $type = "L";
	public $withheld = true;
	public $title = "Caneyville Kentucky Occupational Tax";
	public $miscellaneous_instructions = "Enter 0 for full-time worker.  Enter 1 for part-time worker";
	public $homerule = "#f";
	public $miscellaneous_tag = "Part-time worker";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc)?(52):(104));
	 }
}
class _ky_carlisle extends Tax { 
	public $name = "KY Carlisle";
	public $type = "L";
	public $withheld = true;
	public $title = "Carlisle Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_carroll_co_ extends Tax { 
	public $name = "KY Carroll Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Carroll Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((((($emp->ytdearnings*($emp->payperiods))+($emp->earnings))-((5000*($emp->payperiods))))*(0.01)),0);
	 }
}
class _ky_carrolton extends Tax { 
	public $name = "KY Carrolton";
	public $type = "L";
	public $title = "Dropped";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ky_catlettsburg extends Tax { 
	public $name = "KY Catlettsburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Catlettsburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_cave_city extends Tax { 
	public $name = "KY Cave City";
	public $type = "L";
	public $withheld = true;
	public $title = "Cave City Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_clark_co_ extends Tax { 
	public $name = "KY Clark Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Clark Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_clarkson extends Tax { 
	public $name = "KY Clarkson";
	public $type = "L";
	public $withheld = true;
	public $title = "Clarkson Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.012));
	 }
}
class _ky_clay_city extends Tax { 
	public $name = "KY Clay City";
	public $type = "L";
	public $withheld = true;
	public $title = "Clay City Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_clay_co_ extends Tax { 
	public $name = "KY Clay Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Clay Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_clinton_co_ extends Tax { 
	public $name = "KY Clinton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Clinton Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_coal_run_village extends Tax { 
	public $name = "KY Coal Run Village";
	public $type = "L";
	public $withheld = true;
	public $title = "Coal Run Village Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_cold_spring extends Tax { 
	public $name = "KY Cold Spring";
	public $type = "L";
	public $withheld = true;
	public $title = "Cold Spring Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["FICA Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.01));
	 }
}
class _ky_columbia extends Tax { 
	public $name = "KY Columbia";
	public $type = "L";
	public $withheld = true;
	public $title = "Columbia Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_corbin extends Tax { 
	public $name = "KY Corbin";
	public $type = "L";
	public $withheld = true;
	public $title = "Corbin Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_counties extends Tax { 
	public $name = "KY Counties";
	public $type = "L";
	public $withheld = true;
	public $title = "Kentucky Counties Occupational Tax";
	public $auxiliary_instructions = "Enter municipal tax rate.  For example, enter 1.25% as .0125";
	public $auxiliary_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*($emp->aux));
	 }
}
class _ky_covington extends Tax { 
	public $name = "KY Covington";
	public $type = "L";
	public $withheld = true;
	public $title = "Covington Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["FICA Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0245));
	 }
}
class _ky_crescent_springs extends Tax { 
	public $name = "KY Crescent Springs";
	public $type = "L";
	public $withheld = true;
	public $title = "Crescent Springs Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_cresent_springs extends Tax { 
	public $name = "KY Cresent Springs";
	public $type = "L";
	public $withheld = true;
	public $title = "Crescent Springs Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["KY Crescent Springs"]->compute($emp));
	 }
}
class _ky_crestview_hills extends Tax { 
	public $name = "KY Crestview Hills";
	public $type = "L";
	public $withheld = true;
	public $title = "Crestview Hills Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["KY Crestview Hills Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0115));
	 }
}
class _ky_crestview_hills_wage_base extends Tax { 
	public $name = "KY Crestview Hills Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Crestview Hills Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_cumberland_co_ extends Tax { 
	public $name = "KY Cumberland Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Cumberland Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((60000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0125));
	 }
}
class _ky_cumberland_co__sd extends Tax { 
	public $name = "KY Cumberland Co. SD";
	public $type = "L";
	public $withheld = true;
	public $title = "Cumberland Co. Kentucky School District Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((100000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.005));
	 }
}
class _ky_cynthiana extends Tax { 
	public $name = "KY Cynthiana";
	public $type = "L";
	public $withheld = true;
	public $title = "Cynthiana Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_danville extends Tax { 
	public $name = "KY Danville";
	public $type = "L";
	public $withheld = true;
	public $title = "Danville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.019));
	 }
}
class _ky_daviess_co_ extends Tax { 
	public $name = "KY Daviess Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Daviess Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Owensboro\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_dawson_springs extends Tax { 
	public $name = "KY Dawson Springs";
	public $type = "L";
	public $withheld = true;
	public $title = "Dawson Springs Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_dayton extends Tax { 
	public $name = "KY Dayton";
	public $type = "L";
	public $withheld = true;
	public $title = "Dayton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_dry_ridge extends Tax { 
	public $name = "KY Dry Ridge";
	public $type = "L";
	public $withheld = true;
	public $title = "Dry Ridge Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_earlington extends Tax { 
	public $name = "KY Earlington";
	public $type = "L";
	public $withheld = true;
	public $title = "Earlington Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_eddyville extends Tax { 
	public $name = "KY Eddyville";
	public $type = "L";
	public $withheld = true;
	public $title = "Eddyville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_edgewood extends Tax { 
	public $name = "KY Edgewood";
	public $type = "L";
	public $withheld = true;
	public $title = "Edgewood Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_edmonton extends Tax { 
	public $name = "KY Edmonton";
	public $type = "L";
	public $withheld = true;
	public $title = "Edmonton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_elizabethtown extends Tax { 
	public $name = "KY Elizabethtown";
	public $type = "L";
	public $withheld = true;
	public $title = "Elizabethtown Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0135));
	 }
}
class _ky_elkhorn_city extends Tax { 
	public $name = "KY Elkhorn City";
	public $type = "L";
	public $withheld = true;
	public $title = "Elkhorn City Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_elkton extends Tax { 
	public $name = "KY Elkton";
	public $type = "L";
	public $withheld = true;
	public $title = "Eltkon Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_elsmere extends Tax { 
	public $name = "KY Elsmere";
	public $type = "L";
	public $withheld = true;
	public $title = "Elsmere Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_eminence extends Tax { 
	public $name = "KY Eminence";
	public $type = "L";
	public $withheld = true;
	public $title = "Eminence Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_erlanger extends Tax { 
	public $name = "KY Erlanger";
	public $type = "L";
	public $withheld = true;
	public $title = "Erlanger Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_estill_co_ extends Tax { 
	public $name = "KY Estill Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Estill Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_fayette_co_ extends Tax { 
	public $name = "KY Fayette Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Lexington Fayette Urban County Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0225));
	 }
}
class _ky_fayette_co__sd extends Tax { 
	public $name = "KY Fayette Co. SD";
	public $type = "L";
	public $withheld = true;
	public $title = "Lexington Fayette Urban County School District Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_flemingsburg extends Tax { 
	public $name = "KY Flemingsburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Flemingsburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_florence extends Tax { 
	public $name = "KY Florence";
	public $type = "L";
	public $withheld = true;
	public $title = "Florence Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["FICA Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.02));
	 }
}
class _ky_fort_mitchell extends Tax { 
	public $name = "KY Fort Mitchell";
	public $type = "L";
	public $withheld = true;
	public $title = "Fort Mitchell Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_fort_thomas extends Tax { 
	public $name = "KY Fort Thomas";
	public $type = "L";
	public $withheld = true;
	public $title = "Fort Thomas Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_fort_wright extends Tax { 
	public $name = "KY Fort Wright";
	public $type = "L";
	public $withheld = true;
	public $title = "Fort Wright Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["KY Fort Wright Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0115));
	 }
}
class _ky_fort_wright_wage_base extends Tax { 
	public $name = "KY Fort Wright Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Fort Wright Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_frankfort extends Tax { 
	public $name = "KY Frankfort";
	public $type = "L";
	public $withheld = true;
	public $title = "Frankfort Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0195));
	 }
}
class _ky_franklin extends Tax { 
	public $name = "KY Franklin";
	public $type = "L";
	public $withheld = true;
	public $title = "Franklin Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["KY Franklin Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.01));
	 }
}
class _ky_franklin_co_ extends Tax { 
	public $name = "KY Franklin Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Franklin Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Frankfort\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_franklin_wage_base extends Tax { 
	public $name = "KY Franklin Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Franklin Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_fulton extends Tax { 
	public $name = "KY Fulton";
	public $type = "L";
	public $withheld = true;
	public $title = "Fulton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_gallatin_co_ extends Tax { 
	public $name = "KY Gallatin Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Gallatin Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_gamaliel extends Tax { 
	public $name = "KY Gamaliel";
	public $type = "L";
	public $withheld = true;
	public $title = "Gamaliel Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_garrard_co_ extends Tax { 
	public $name = "KY Garrard Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Garrard Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_georgetown extends Tax { 
	public $name = "KY Georgetown";
	public $type = "L";
	public $withheld = true;
	public $title = "Georgetown Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_glasgow extends Tax { 
	public $name = "KY Glasgow";
	public $type = "L";
	public $withheld = true;
	public $title = "Glasgow Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_grant_co_ extends Tax { 
	public $name = "KY Grant Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Grant County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_graves_co_ extends Tax { 
	public $name = "KY Graves Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Graves Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Mayfield\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_grayson extends Tax { 
	public $name = "KY Grayson";
	public $type = "L";
	public $withheld = true;
	public $title = "Grayson Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_grayson_co_ extends Tax { 
	public $name = "KY Grayson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Grayson Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_greensburg extends Tax { 
	public $name = "KY Greensburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Greensburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_greenup extends Tax { 
	public $name = "KY Greenup";
	public $type = "L";
	public $withheld = true;
	public $title = "Greenup Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_guthrie extends Tax { 
	public $name = "KY Guthrie";
	public $type = "L";
	public $withheld = true;
	public $title = "Guthrie Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_hancock_co_ extends Tax { 
	public $name = "KY Hancock Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Hancock Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_harrison_co_ extends Tax { 
	public $name = "KY Harrison Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Harrison Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_harrodsburg extends Tax { 
	public $name = "KY Harrodsburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Harrodsburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_hart_co_ extends Tax { 
	public $name = "KY Hart Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Hart Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.008));
	 }
}
class _ky_hartford extends Tax { 
	public $name = "KY Hartford";
	public $type = "L";
	public $withheld = true;
	public $title = "Hartford Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_hazard extends Tax { 
	public $name = "KY Hazard";
	public $type = "L";
	public $withheld = true;
	public $title = "Hazard Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max((min($emp->earnings,((60000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0125)),0)+(max((min($emp->earnings,((($emp->ytdearnings*($emp->payperiods))+($emp->earnings))-((60000*($emp->payperiods)))))*(0.005)),0)));
	 }
}
class _ky_henderson extends Tax { 
	public $name = "KY Henderson";
	public $type = "L";
	public $withheld = true;
	public $title = "Henderson Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0165));
	 }
}
class _ky_hickman extends Tax { 
	public $name = "KY Hickman";
	public $type = "L";
	public $withheld = true;
	public $title = "Hickman Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_hickman_co_ extends Tax { 
	public $name = "KY Hickman Co.";
	public $type = "L";
	public $withheld = true;
	public $title = " Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_highland_heights extends Tax { 
	public $name = "KY Highland Heights";
	public $type = "L";
	public $withheld = true;
	public $title = "Highland Heights Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["FICA Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.01));
	 }
}
class _ky_hillview extends Tax { 
	public $name = "KY Hillview";
	public $type = "L";
	public $withheld = true;
	public $title = "Hillview Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.018));
	 }
}
class _ky_hodgenville extends Tax { 
	public $name = "KY Hodgenville";
	public $type = "L";
	public $withheld = true;
	public $title = "Hodgenville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_hopkins_co_ extends Tax { 
	public $name = "KY Hopkins Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Hopkins County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_hopkinsville extends Tax { 
	public $name = "KY Hopkinsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Hopkinsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0195));
	 }
}
class _ky_horse_cave extends Tax { 
	public $name = "KY Horse Cave";
	public $type = "L";
	public $withheld = true;
	public $title = "Horse Cave Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_independence extends Tax { 
	public $name = "KY Independence";
	public $type = "L";
	public $withheld = true;
	public $title = "Independence Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_jackson extends Tax { 
	public $name = "KY Jackson";
	public $type = "L";
	public $withheld = true;
	public $title = "Jackson Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_jackson_co_ extends Tax { 
	public $name = "KY Jackson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jackson County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0185));
	 }
}
class _ky_jamestown extends Tax { 
	public $name = "KY Jamestown";
	public $type = "L";
	public $withheld = true;
	public $title = "Jamestown Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_jefferson_co__nonresident extends Tax { 
	public $name = "KY Jefferson Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jefferson County Kentucky Nonresident Income Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings*(0.0125))+(($emp->earnings*(0.002))));
	 }
}
class _ky_jefferson_co__resident extends Tax { 
	public $name = "KY Jefferson Co. Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "Jefferson County Kentucky Resident Income Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((($emp->earnings*(0.0075))+(($emp->earnings*(0.0125))))+(($emp->earnings*(0.002))));
	 }
}
class _ky_jeffersontown extends Tax { 
	public $name = "KY Jeffersontown";
	public $type = "L";
	public $withheld = true;
	public $title = "Jeffersontown Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_jeffersonville extends Tax { 
	public $name = "KY Jeffersonville";
	public $type = "L";
	public $withheld = true;
	public $title = "Jeffersonville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_jessamine_co_ extends Tax { 
	public $name = "KY Jessamine Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Jessamine Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_johnson_co_ extends Tax { 
	public $name = "KY Johnson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Johnson Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_junction_city extends Tax { 
	public $name = "KY Junction City";
	public $type = "L";
	public $withheld = true;
	public $title = "Junction City Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_kenton_co_ extends Tax { 
	public $name = "KY Kenton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Kenton Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max((min($emp->earnings,max(0.0,$taxlist["KY Kenton Co. bracket 1"]->compute($emp)))*(0.009097)),0)+(max(((min($emp->earnings,max(0.0,$taxlist["KY Kenton Co. bracket 2"]->compute($emp)))-(min($emp->earnings,max(0.0,$taxlist["KY Kenton Co. bracket 1"]->compute($emp)))))*(0.001097)),0)));
	 }
}
class _ky_kenton_co__wage_base extends Tax { 
	public $name = "KY Kenton Co. Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Kenton Co. Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_kenton_co__bracket_1 extends Tax { 
	public $name = "KY Kenton Co. bracket 1";
	public $type = "P";
	public $withheld = true;
	public $title = "Kenton Co. Kentucky ";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(0.0,$taxlist["KY Kenton Co. Wage Base"]->compute($emp))*(0.5))-(($emp->ytdearnings*($emp->payperiods))));
	 }
}
class _ky_kenton_co__bracket_2 extends Tax { 
	public $name = "KY Kenton Co. bracket 2";
	public $type = "P";
	public $withheld = true;
	public $title = "Kenton Co. Kentucky ";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["KY Kenton Co. Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods))));
	 }
}
class _ky_knox_co_ extends Tax { 
	public $name = "KY Knox Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Know Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_la_grange extends Tax { 
	public $name = "KY La Grange";
	public $type = "L";
	public $withheld = true;
	public $title = "La Grange Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_lakeside_park extends Tax { 
	public $name = "KY Lakeside Park";
	public $type = "L";
	public $withheld = true;
	public $title = "Lakeside Park Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_laurel_co_ extends Tax { 
	public $name = "KY Laurel Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Laurel Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_lebanon extends Tax { 
	public $name = "KY Lebanon";
	public $type = "L";
	public $withheld = true;
	public $title = "Lebanon Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_lebanon_junction extends Tax { 
	public $name = "KY Lebanon Junction";
	public $type = "L";
	public $withheld = true;
	public $title = "Lebanon Junction Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_leitchfield extends Tax { 
	public $name = "KY Leitchfield";
	public $type = "L";
	public $withheld = true;
	public $title = "Leitchfield Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.012));
	 }
}
class _ky_leslie_co_ extends Tax { 
	public $name = "KY Leslie Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Leslie Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_lewisburg extends Tax { 
	public $name = "KY Lewisburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Lewisburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_lexington extends Tax { 
	public $name = "KY Lexington";
	public $type = "L";
	public $withheld = true;
	public $title = "use KY Fayette Co.";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["KY Fayette Co."]->compute($emp));
	 }
}
class _ky_lincoln_co_ extends Tax { 
	public $name = "KY Lincoln Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Lincoln Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_livingston_co_ extends Tax { 
	public $name = "KY Livingston Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Livingston Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_logan_co_ extends Tax { 
	public $name = "KY Logan Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Logan Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_louisville_nonresident extends Tax { 
	public $name = "KY Louisville Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "see KY Jefferson Co.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ky_louisville_resident extends Tax { 
	public $name = "KY Louisville Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "see KY Jefferson Co.";
	public $homerule = "#f";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ky_ludlow extends Tax { 
	public $name = "KY Ludlow";
	public $type = "L";
	public $withheld = true;
	public $title = "Ludlow Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_madison_co_ extends Tax { 
	public $name = "KY Madison Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Madison Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Berea\" \"KY Richmond\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_madisonville extends Tax { 
	public $name = "KY Madisonville";
	public $type = "L";
	public $withheld = true;
	public $title = "Madisonville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_magoffin_co_ extends Tax { 
	public $name = "KY Magoffin Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Magoffin County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_marion extends Tax { 
	public $name = "KY Marion";
	public $type = "L";
	public $withheld = true;
	public $title = "Marion Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_marion_co_ extends Tax { 
	public $name = "KY Marion Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Marion Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_marshall_co__nonresident extends Tax { 
	public $name = "KY Marshall Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Marshall Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_marshall_co__resident extends Tax { 
	public $name = "KY Marshall Co. Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "Marshall Co. Kentucky School District Tax";
	public $homerule = "#f";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_martin extends Tax { 
	public $name = "KY Martin";
	public $type = "L";
	public $withheld = true;
	public $title = "Martin Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.013));
	 }
}
class _ky_martin_co_ extends Tax { 
	public $name = "KY Martin Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Martin County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_mayfield extends Tax { 
	public $name = "KY Mayfield";
	public $type = "L";
	public $withheld = true;
	public $title = "Mayfield Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.019));
	 }
}
class _ky_maysville extends Tax { 
	public $name = "KY Maysville";
	public $type = "L";
	public $withheld = true;
	public $title = "Maysville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0199));
	 }
}
class _ky_mccracken_co_ extends Tax { 
	public $name = "KY McCracken Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "McCracken Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Paducah\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_mccreary_co_ extends Tax { 
	public $name = "KY McCreary Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "McCreary Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_mckee extends Tax { 
	public $name = "KY McKee";
	public $type = "L";
	public $withheld = true;
	public $title = "McKee Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_mclean_co_ extends Tax { 
	public $name = "KY McLean Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "McLean Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_menifee_co_ extends Tax { 
	public $name = "KY Menifee Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Menifee Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_mercer_co_ extends Tax { 
	public $name = "KY Mercer Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Mercer Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0045));
	 }
}
class _ky_metcalfe_co_ extends Tax { 
	public $name = "KY Metcalfe Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Metcalfe Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_middlesboro extends Tax { 
	public $name = "KY Middlesboro";
	public $type = "L";
	public $withheld = true;
	public $title = "Middlesboro Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_midway extends Tax { 
	public $name = "KY Midway";
	public $type = "L";
	public $withheld = true;
	public $title = "Midway Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_millersburg extends Tax { 
	public $name = "KY Millersburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Millersburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_monroe_co_ extends Tax { 
	public $name = "KY Monroe Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Monroe Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_montgomery_co_ extends Tax { 
	public $name = "KY Montgomery Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Montgomery Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Camargo\" \"KY Jeffersonville\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_morehead extends Tax { 
	public $name = "KY Morehead";
	public $type = "L";
	public $withheld = true;
	public $title = "Morehead Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_morgan_co_ extends Tax { 
	public $name = "KY Morgan Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Morgan Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_morgantown extends Tax { 
	public $name = "KY Morgantown";
	public $type = "L";
	public $withheld = true;
	public $title = "Morgantown Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_mount_olivet extends Tax { 
	public $name = "KY Mount Olivet";
	public $type = "L";
	public $withheld = true;
	public $title = "Mount Olivet Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_mount_vernon extends Tax { 
	public $name = "KY Mount Vernon";
	public $type = "L";
	public $withheld = true;
	public $title = "Mount Vernon Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_mount_washington extends Tax { 
	public $name = "KY Mount Washington";
	public $type = "L";
	public $withheld = true;
	public $title = "Mount Washington Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_muldraugh extends Tax { 
	public $name = "KY Muldraugh";
	public $type = "L";
	public $withheld = true;
	public $title = "Muldraugh Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0175));
	 }
}
class _ky_munfordville extends Tax { 
	public $name = "KY Munfordville";
	public $type = "L";
	public $withheld = true;
	public $title = "Munfordville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_municipalities extends Tax { 
	public $name = "KY Municipalities";
	public $type = "L";
	public $withheld = true;
	public $title = "Kentucky Municipalities Income Tax";
	public $auxiliary_instructions = "Enter municipal tax rate.  For example, enter 1.25% as .0125";
	public $auxiliary_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*($emp->aux));
	 }
}
class _ky_murray extends Tax { 
	public $name = "KY Murray";
	public $type = "L";
	public $withheld = true;
	public $title = "Murray Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_nelson_co_ extends Tax { 
	public $name = "KY Nelson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Nelson Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((15000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.005));
	 }
}
class _ky_newport extends Tax { 
	public $name = "KY Newport";
	public $type = "L";
	public $withheld = true;
	public $title = "Newport Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["KY Newport Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.025));
	 }
}
class _ky_newport_wage_base extends Tax { 
	public $name = "KY Newport Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Newport Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_nicholas_co_ extends Tax { 
	public $name = "KY Nicholas Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Nicholas Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_nicholasville extends Tax { 
	public $name = "KY Nicholasville";
	public $type = "L";
	public $withheld = true;
	public $title = "Nicholasville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_nortonville extends Tax { 
	public $name = "KY Nortonville";
	public $type = "L";
	public $withheld = true;
	public $title = "Nortonville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_oak_grove extends Tax { 
	public $name = "KY Oak Grove";
	public $type = "L";
	public $withheld = true;
	public $title = "Oak Grove Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_ohio_co_ extends Tax { 
	public $name = "KY Ohio Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Ohio Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_olive_hill extends Tax { 
	public $name = "KY Olive Hill";
	public $type = "L";
	public $withheld = true;
	public $title = "Olive Hill Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_owensboro extends Tax { 
	public $name = "KY Owensboro";
	public $type = "L";
	public $withheld = true;
	public $title = "Owensboro Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0178));
	 }
}
class _ky_owenton extends Tax { 
	public $name = "KY Owenton";
	public $type = "L";
	public $withheld = true;
	public $title = "Owenton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_paducah extends Tax { 
	public $name = "KY Paducah";
	public $type = "L";
	public $withheld = true;
	public $title = "Paducah Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_paintsville extends Tax { 
	public $name = "KY Paintsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Paintsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_paris extends Tax { 
	public $name = "KY Paris";
	public $type = "L";
	public $withheld = true;
	public $title = "Paris Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_park_city extends Tax { 
	public $name = "KY Park City";
	public $type = "L";
	public $withheld = true;
	public $title = "Park City Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_park_hills extends Tax { 
	public $name = "KY Park Hills";
	public $type = "L";
	public $withheld = true;
	public $title = "Park Hills Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,((50000*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))))*(0.015));
	 }
}
class _ky_pendleton_co_ extends Tax { 
	public $name = "KY Pendleton Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Pendleton County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_perry_co_ extends Tax { 
	public $name = "KY Perry Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Perry County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_perryville extends Tax { 
	public $name = "KY Perryville";
	public $type = "L";
	public $withheld = true;
	public $title = "Perryville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_pikeville extends Tax { 
	public $name = "KY Pikeville";
	public $type = "L";
	public $withheld = true;
	public $title = "Pikeville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_pineville extends Tax { 
	public $name = "KY Pineville";
	public $type = "L";
	public $withheld = true;
	public $title = "Pineville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_pioneer_village extends Tax { 
	public $name = "KY Pioneer Village";
	public $type = "L";
	public $withheld = true;
	public $title = "Pioneer Village Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_powell_co_ extends Tax { 
	public $name = "KY Powell Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Powell Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_prestonsburg extends Tax { 
	public $name = "KY Prestonsburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Prestonsburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_princeton extends Tax { 
	public $name = "KY Princeton";
	public $type = "L";
	public $withheld = true;
	public $title = "Princeton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_pulaski_co_ extends Tax { 
	public $name = "KY Pulaski Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Pulaski Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_raceland extends Tax { 
	public $name = "KY Raceland";
	public $type = "L";
	public $withheld = true;
	public $title = "Raceland Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_radcliff extends Tax { 
	public $name = "KY Radcliff";
	public $type = "L";
	public $withheld = true;
	public $title = "Radcliff Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0225));
	 }
}
class _ky_richmond extends Tax { 
	public $name = "KY Richmond";
	public $type = "L";
	public $withheld = true;
	public $title = "Richmond Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_robertson_co_ extends Tax { 
	public $name = "KY Robertson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Robertson Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_rockcastle_co_ extends Tax { 
	public $name = "KY Rockcastle Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Rockcastle Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_rowan_co_ extends Tax { 
	public $name = "KY Rowan Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Rowan Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_russell extends Tax { 
	public $name = "KY Russell";
	public $type = "L";
	public $withheld = true;
	public $title = "Russell Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0115));
	 }
}
class _ky_russell_county extends Tax { 
	public $name = "KY Russell County";
	public $type = "L";
	public $withheld = true;
	public $title = "Russell County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_russell_springs extends Tax { 
	public $name = "KY Russell Springs";
	public $type = "L";
	public $withheld = true;
	public $title = "Russell Springs Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_russellville extends Tax { 
	public $name = "KY Russellville";
	public $type = "L";
	public $withheld = true;
	public $title = "Russellville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_ryland_heights extends Tax { 
	public $name = "KY Ryland Heights";
	public $type = "L";
	public $withheld = true;
	public $title = "Ryland Heights Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_salyersville extends Tax { 
	public $name = "KY Salyersville";
	public $type = "L";
	public $withheld = true;
	public $title = "Salyersville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_scott_co__nonresident extends Tax { 
	public $name = "KY Scott Co. Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Scott Co. Kentucky Nonresident Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_scott_co__resident extends Tax { 
	public $name = "KY Scott Co. Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "Scott Co. Kentucky Resident Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_scottsville extends Tax { 
	public $name = "KY Scottsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Scottsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_shelby_co_ extends Tax { 
	public $name = "KY Shelby Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Shelby Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Shelbyville\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_shelbyville extends Tax { 
	public $name = "KY Shelbyville";
	public $type = "L";
	public $withheld = true;
	public $title = "Shelbyville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_shepherdsville extends Tax { 
	public $name = "KY Shepherdsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Shepherdsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_shively extends Tax { 
	public $name = "KY Shively";
	public $type = "L";
	public $withheld = true;
	public $title = "Shively Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_silver_grove extends Tax { 
	public $name = "KY Silver Grove";
	public $type = "L";
	public $withheld = true;
	public $title = "Silver Grove Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_simpson_co_ extends Tax { 
	public $name = "KY Simpson Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Simpson Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Shelbyville\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["KY Simpson Co. Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.0075));
	 }
}
class _ky_simpson_co__wage_base extends Tax { 
	public $name = "KY Simpson Co. Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "KY Simpson Co. Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ky_simpsonville extends Tax { 
	public $name = "KY Simpsonville";
	public $type = "L";
	public $withheld = true;
	public $title = "Simpsonville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_somerset extends Tax { 
	public $name = "KY Somerset";
	public $type = "L";
	public $withheld = true;
	public $title = "Somerset Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.006));
	 }
}
class _ky_southgate extends Tax { 
	public $name = "KY Southgate";
	public $type = "L";
	public $withheld = true;
	public $title = "Southgate Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.025));
	 }
}
class _ky_spencer_co_ extends Tax { 
	public $name = "KY Spencer Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Spencer County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.008));
	 }
}
class _ky_springfield extends Tax { 
	public $name = "KY Springfield";
	public $type = "L";
	public $withheld = true;
	public $title = "Springfield Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_st__matthews extends Tax { 
	public $name = "KY St. Matthews";
	public $type = "L";
	public $withheld = true;
	public $title = "St. Matthews Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_stanford extends Tax { 
	public $name = "KY Stanford";
	public $type = "L";
	public $withheld = true;
	public $title = "Stanford Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0065));
	 }
}
class _ky_stanton extends Tax { 
	public $name = "KY Stanton";
	public $type = "L";
	public $withheld = true;
	public $title = "Stanton Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_taylor_co_ extends Tax { 
	public $name = "KY Taylor Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Taylor Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_taylor_mill extends Tax { 
	public $name = "KY Taylor Mill";
	public $type = "L";
	public $withheld = true;
	public $title = "Taylor Mill Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_taylorsville extends Tax { 
	public $name = "KY Taylorsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Taylorsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.008));
	 }
}
class _ky_todd_co_ extends Tax { 
	public $name = "KY Todd Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Todd Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_tompkinsville extends Tax { 
	public $name = "KY Tompkinsville";
	public $type = "L";
	public $withheld = true;
	public $title = "Tompkinsville Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_unemployment extends Tax { 
	public $name = "KY Unemployment";
	public $type = "O";
	public $title = "Kentucky Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["KY Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ky_unemployment_wage_base extends Tax { 
	public $name = "KY Unemployment Wage Base";
	public $type = "P";
	public $title = "KY Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 11100;
	 }
}
class _ky_union extends Tax { 
	public $name = "KY Union";
	public $type = "L";
	public $withheld = true;
	public $title = "Union Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_union_co_ extends Tax { 
	public $name = "KY Union Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Union Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_vanceburg extends Tax { 
	public $name = "KY Vanceburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Vanceburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_versailles extends Tax { 
	public $name = "KY Versailles";
	public $type = "L";
	public $withheld = true;
	public $title = "Versailles Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_villa_hills extends Tax { 
	public $name = "KY Villa Hills";
	public $type = "L";
	public $withheld = true;
	public $title = "Villa Hills Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_vine_grove extends Tax { 
	public $name = "KY Vine Grove";
	public $type = "L";
	public $withheld = true;
	public $title = "Vine Grove Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_warren_co_ extends Tax { 
	public $name = "KY Warren Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Warren Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (workzip) \"KY Bowling Green\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_warren_co__sd extends Tax { 
	public $name = "KY Warren Co. SD";
	public $type = "L";
	public $withheld = true;
	public $title = "Warren Co. Kentucky School District Tax";
	public $homerule = "#f";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_warsaw extends Tax { 
	public $name = "KY Warsaw";
	public $type = "L";
	public $withheld = true;
	public $title = "Warsaw Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_washington_co_ extends Tax { 
	public $name = "KY Washington Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Washington Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0075));
	 }
}
class _ky_wayne_co_ extends Tax { 
	public $name = "KY Wayne Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Wayne Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.009));
	 }
}
class _ky_west_beuchel extends Tax { 
	public $name = "KY West Beuchel";
	public $type = "L";
	public $withheld = true;
	public $title = "West Bueschel Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["KY West Buechel"]->compute($emp));
	 }
}
class _ky_west_buechel extends Tax { 
	public $name = "KY West Buechel";
	public $type = "L";
	public $withheld = true;
	public $title = "West Buechel Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_west_bueschel extends Tax { 
	public $name = "KY West Bueschel";
	public $type = "L";
	public $withheld = true;
	public $title = "West Bueschel Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["KY West Buechel"]->compute($emp));
	 }
}
class _ky_west_liberty extends Tax { 
	public $name = "KY West Liberty";
	public $type = "L";
	public $withheld = true;
	public $title = "West Liberty Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _ky_west_point extends Tax { 
	public $name = "KY West Point";
	public $type = "L";
	public $withheld = true;
	public $title = "West Point Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_wheelwright extends Tax { 
	public $name = "KY Wheelwright";
	public $type = "L";
	public $withheld = true;
	public $title = "Wheelwright Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 60;
	 }
}
class _ky_whitesburg extends Tax { 
	public $name = "KY Whitesburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Whitesburg Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _ky_whitley_co_ extends Tax { 
	public $name = "KY Whitley Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Whitley County Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _ky_wilder extends Tax { 
	public $name = "KY Wilder";
	public $type = "L";
	public $withheld = true;
	public $title = "Wilder Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0225));
	 }
}
class _ky_wilmore extends Tax { 
	public $name = "KY Wilmore";
	public $type = "L";
	public $withheld = true;
	public $title = "Wilmore Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_winchester extends Tax { 
	public $name = "KY Winchester";
	public $type = "L";
	public $withheld = true;
	public $title = "Winchester Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.02));
	 }
}
class _ky_wolfe_co_ extends Tax { 
	public $name = "KY Wolfe Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Wolfe Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0125));
	 }
}
class _ky_woodford_co_ extends Tax { 
	public $name = "KY Woodford Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Woodford Co. Kentucky Occupational Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.015));
	 }
}
class _la extends Tax { 
	public $name = "LA";
	public $type = "S";
	public $withheld = true;
	public $title = "Louisiana Income Tax";
	public $state_exemption_instructions = "Enter number of dependents from form L-4 line 7.  Do not include the personal exemption.";
	public $miscellaneous_instructions = "Enter 0 for no personal exemption.  Enter 1 for one personal exemption.  Enter 2 for two personal exemptions.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Dependents";
	public $miscellaneous_tag = "Personal exemption";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus>(0))&&(($emp->misc>(1))))?(max(0.0,$taxlist["LA Married"]->compute($emp))):(max(0.0,$taxlist["LA Single"]->compute($emp))));
	 }
}
class _la_a extends Tax { 
	public $name = "LA A";
	public $type = "P";
	public $title = "LA A";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["LA A1"]->compute($emp))*(0.0185));
	 }
}
class _la_a1 extends Tax { 
	public $name = "LA A1";
	public $type = "P";
	public $title = "(X * 4500)+(Y * 1000)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(min($emp->misc,(isTrue(($emp->filingstatus<(3)))?(1):(2))),0)*(4500))+(($emp->stateexemptions*(1000))));
	 }
}
class _la_b extends Tax { 
	public $name = "LA B";
	public $type = "P";
	public $title = "LA B";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((isTrue(($emp->filingstatus<(3)))?(((max(0.0,$taxlist["LA A1"]->compute($emp))-(12500))*(0.0165))):(((max(0.0,$taxlist["LA A1"]->compute($emp))-(25000))*(0.0165)))),0);
	 }
}
class _la_married extends Tax { 
	public $name = "LA Married";
	public $type = "P";
	public $title = "LA Married";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((((($emp->earnings*(0.0185))+((isTrue(($emp->earnings>(25000)))?((($emp->earnings-(25000))*(0.0165))):(0))))+((isTrue(($emp->earnings>(100000)))?((($emp->earnings-(100000))*(0.0075))):(0))))-(max(0.0,$taxlist["LA A"]->compute($emp))))-(max(0.0,$taxlist["LA B"]->compute($emp))));
	 }
}
class _la_single extends Tax { 
	public $name = "LA Single";
	public $type = "P";
	public $title = "LA Single";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((((($emp->earnings*(0.0185))+((isTrue(($emp->earnings>(12500)))?((($emp->earnings-(12500))*(0.0165))):(0))))+((isTrue(($emp->earnings>(50000)))?((($emp->earnings-(50000))*(0.0075))):(0))))-(max(0.0,$taxlist["LA A"]->compute($emp))))-(max(0.0,$taxlist["LA B"]->compute($emp))));
	 }
}
class _la_unemployment extends Tax { 
	public $name = "LA Unemployment";
	public $type = "O";
	public $title = "Louisiana Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["LA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _la_unemployment_wage_base extends Tax { 
	public $name = "LA Unemployment Wage Base";
	public $type = "P";
	public $title = "LA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 7700;
	 }
}
class _ma extends Tax { 
	public $name = "MA";
	public $type = "S";
	public $withheld = true;
	public $title = "Massachusetts Income Tax";
	public $miscellaneous_instructions = "Enter 0 for no blindness exemption.  Enter 1 for one blindness exemption.  Enter 2 for two blindness exemptions.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $miscellaneous_tag = "Blindness exemption";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->earnings<(8000))&&(($emp->stateexemptions>(0))))?(0):((((max(0.0,$taxlist["MA Taxable Income"]->compute($emp))*(0.05))-((isTrue($emp->filingstatus==(1))?(120):(0))))-((min($emp->misc,2)*(110))))));
	 }
}
class _ma_emac extends Tax { 
	public $name = "MA EMAC";
	public $type = "O";
	public $title = "Massachusetts Employer Medical Assistance Contribution";
	public $auxiliary_instructions = "Enter assigned rate.  For example, enter 0.34% as 0.0034";
	public $auxiliary_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MA EMAC Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($emp->aux)),0);
	 }
}
class _ma_emac_wage_base extends Tax { 
	public $name = "MA EMAC Wage Base";
	public $type = "P";
	public $title = "MA EMAC Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 15000;
	 }
}
class _ma_exemption_factors extends Tax { 
	public $name = "MA Exemption Factors";
	public $type = "P";
	public $withheld = true;
	public $title = "MA Exemption Factors";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->stateexemptions)?(max(((1000*($emp->stateexemptions))+(3400)),0)):(0));
	 }
}
class _ma_family_leave extends Tax { 
	public $name = "MA Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "Massachusetts Family Leave Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max((max(0.0,$taxlist["MA Family Leave Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0011));
	 }
}
class _ma_family_leave_employer extends Tax { 
	public $name = "MA Family Leave Employer";
	public $type = "O";
	public $title = "Massachusetts Medical Leave Employer Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ma_family_leave_wage_base extends Tax { 
	public $name = "MA Family Leave Wage Base";
	public $type = "P";
	public $title = "MA Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ma_medical_leave extends Tax { 
	public $name = "MA Medical Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "Massachusetts Medical Leave Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max((max(0.0,$taxlist["MA Medical Leave Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.00208));
	 }
}
class _ma_medical_leave_employer extends Tax { 
	public $name = "MA Medical Leave Employer";
	public $type = "O";
	public $title = "Massachusetts Medical Leave Tax";
	public $miscellaneous_instructions = "Enter 1 if employer has 25 or more employees.";
	public $miscellaneous_tag = "25 employees";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc)?((min($emp->earnings,max((max(0.0,$taxlist["MA Medical Leave Employer Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.00312))):(0));
	 }
}
class _ma_medical_leave_employer_wage_base extends Tax { 
	public $name = "MA Medical Leave Employer Wage Base";
	public $type = "P";
	public $title = "MA Medical and Family Leave Employer Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ma_medical_leave_wage_base extends Tax { 
	public $name = "MA Medical Leave Wage Base";
	public $type = "P";
	public $title = "MA Medical and Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ma_medical_and_family_leave extends Tax { 
	public $name = "MA Medical and Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "Massachusetts Medical and Family Leave Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MA Medical Leave"]->compute($emp))+(max(0.0,$taxlist["MA Family Leave"]->compute($emp))));
	 }
}
class _ma_medical_and_family_leave_employer extends Tax { 
	public $name = "MA Medical and Family Leave Employer";
	public $type = "O";
	public $title = "Massachusetts Medical and Family Leave Tax (Employer)";
	public $miscellaneous_instructions = "Enter 1 if employer has 25 or more employees.";
	public $miscellaneous_tag = "25 employees";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["MA Medical Leave Employer"]->compute($emp));
	 }
}
class _ma_medical_and_family_leave_employer_wage_base extends Tax { 
	public $name = "MA Medical and Family Leave Employer Wage Base";
	public $type = "P";
	public $title = "MA Medical and Family Leave Employer Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ma_medical_and_family_leave_wage_base extends Tax { 
	public $name = "MA Medical and Family Leave Wage Base";
	public $type = "P";
	public $title = "MA Medical and Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _ma_se extends Tax { 
	public $name = "MA SE";
	public $type = "S";
	public $withheld = true;
	public $title = "Massachusetts Income Tax (for SECA employees)";
	public $miscellaneous_instructions = "Enter 0 for no blindness exemption.  Enter 1 for one blindness exemption.  Enter 2 for two blindness exemptions.";
	public $miscellaneous_tag = "Blindness exemption";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->earnings<(8000))&&(($emp->stateexemptions>(0))))?(0):(((((($emp->earnings-(min(max(0.0,$taxlist["SECA"]->compute($emp)),2000)))-(max(0.0,$taxlist["MA Exemption Factors"]->compute($emp))))*(0.053))-((isTrue($emp->filingstatus==(1))?(127.2):(0))))-(($emp->misc*(116.6))))));
	 }
}
class _ma_taxable_income extends Tax { 
	public $name = "MA Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "MA Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((($emp->earnings-(min(($emp->earnings*(0.0765)),max(((26143.79-($emp->ytdearnings))*(0.0765)),0))))-(max(0.0,$taxlist["MA Exemption Factors"]->compute($emp)))),0);
	 }
}
class _ma_unemployment extends Tax { 
	public $name = "MA Unemployment";
	public $type = "O";
	public $title = "Massachusetts Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ma_unemployment_wage_base extends Tax { 
	public $name = "MA Unemployment Wage Base";
	public $type = "P";
	public $title = "MA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 15000;
	 }
}
class _ma_workforce_training_fund extends Tax { 
	public $name = "MA Workforce Training Fund";
	public $type = "O";
	public $title = "Massachusetts Workforce Training Fund";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max(((max(0.0,$taxlist["MA Workforce Training Fund Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.00056));
	 }
}
class _ma_workforce_training_fund_wage_base extends Tax { 
	public $name = "MA Workforce Training Fund Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "MA Workforce Training Fund Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["MA Unemployment Wage Base"]->compute($emp));
	 }
}
class _md extends Tax { 
	public $name = "MD";
	public $type = "S";
	public $withheld = true;
	public $title = "Maryland Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "(not (within (homezip) \"DC\" \"VA\" \"WV\" \"PA\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(4))&&(($emp->filingstatus>(0))))?(tableLookup(max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp)),array(array(0,0.0475,0),array(150000,0.05,7125),array(175000,0.0525,8375),array(225000,0.055,11000),array(300000,0.0575,15125)))):(tableLookup(max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp)),array(array(0,0.0475,0),array(100000,0.05,4750),array(125000,0.0525,6000),array(150000,0.055,8625),array(250000,0.0575,14375)))));
	 }
}
class _md_allegany_co_ extends Tax { 
	public $name = "MD Allegany Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Allegany Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0303));
	 }
}
class _md_anne_arundel_co_ extends Tax { 
	public $name = "MD Anne Arundel Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Anne Arundel Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*((isTrue(($emp->earnings<(50000)))?(0.027):(0.0281))));
	 }
}
class _md_baltimore extends Tax { 
	public $name = "MD Baltimore";
	public $type = "L";
	public $withheld = true;
	public $title = "Baltimore Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_baltimore_co_ extends Tax { 
	public $name = "MD Baltimore Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Baltimore Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_calvert_co_ extends Tax { 
	public $name = "MD Calvert Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Calvert Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.03));
	 }
}
class _md_caroline_co_ extends Tax { 
	public $name = "MD Caroline Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Caroline Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_carroll_co_ extends Tax { 
	public $name = "MD Carroll Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Carroll Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0303));
	 }
}
class _md_cecil_co_ extends Tax { 
	public $name = "MD Cecil Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Cecil Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.028));
	 }
}
class _md_charles_co_ extends Tax { 
	public $name = "MD Charles Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Charles Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0303));
	 }
}
class _md_delaware extends Tax { 
	public $name = "MD Delaware";
	public $type = "S";
	public $withheld = true;
	public $title = "Maryland Income Tax (Delaware/nonreciprocal state employment)";
	public $homerule = "(and (company-resident?) (not (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\")))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_dorchester_co_ extends Tax { 
	public $name = "MD Dorchester Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Dorchester Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_frederick_co_ extends Tax { 
	public $name = "MD Frederick Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Frederick Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*((isTrue(($emp->earnings>(100000))||($emp->filingstatus==(0)||(($emp->filingstatus>(3)))&&(($emp->earnings>(50000)))))?(0.0296):(0.0275))));
	 }
}
class _md_garrett_co_ extends Tax { 
	public $name = "MD Garrett Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Garrett Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0265));
	 }
}
class _md_harford_co_ extends Tax { 
	public $name = "MD Harford Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Harford Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0306));
	 }
}
class _md_howard_co_ extends Tax { 
	public $name = "MD Howard Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Howard Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_kent_co_ extends Tax { 
	public $name = "MD Kent Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Kent Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_minimum_taxable_income extends Tax { 
	public $name = "MD Minimum Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "MD Minimum Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue((max(0.0,$taxlist["MD Taxable Income"]->compute($emp))>(5000)))?(max(0.0,$taxlist["MD Taxable Income"]->compute($emp))):(0));
	 }
}
class _md_montgomery_co_ extends Tax { 
	public $name = "MD Montgomery Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Montgomery Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_nonresident extends Tax { 
	public $name = "MD Nonresident";
	public $type = "S";
	public $withheld = true;
	public $title = "Maryland Nonresident Income Tax";
	public $miscellaneous_instructions = "For local portion only, set to 1";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"MD\" \"VA\" \"WV\" \"PA\" \"DC\"))";
	public $miscellaneous_tag = "Local only";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0225))+((isTrue($emp->misc)?(0):(max(0.0,$taxlist["MD"]->compute($emp))))));
	 }
}
class _md_prince_george_s_co_ extends Tax { 
	public $name = "MD Prince George's Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Prince George's Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_queen_anne_s_co_ extends Tax { 
	public $name = "MD Queen Anne's Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Queen Anne's Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_somerset_co_ extends Tax { 
	public $name = "MD Somerset Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Somerset Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_st__mary_s_co_ extends Tax { 
	public $name = "MD St. Mary's Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "St. Mary's Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.03));
	 }
}
class _md_talbot_co_ extends Tax { 
	public $name = "MD Talbot Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Talbot Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.024));
	 }
}
class _md_taxable_income extends Tax { 
	public $name = "MD Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "MD Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((($emp->earnings-(max(min(($emp->earnings*(0.15)),2350),1550)))-(($emp->stateexemptions*(3200)))),0);
	 }
}
class _md_unemployment extends Tax { 
	public $name = "MD Unemployment";
	public $type = "O";
	public $title = "Maryland Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MD Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _md_unemployment_wage_base extends Tax { 
	public $name = "MD Unemployment Wage Base";
	public $type = "P";
	public $title = "MD Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 8500;
	 }
}
class _md_washington_co_ extends Tax { 
	public $name = "MD Washington Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Washington Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0295));
	 }
}
class _md_wicomico_co_ extends Tax { 
	public $name = "MD Wicomico Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Wicomico Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.032));
	 }
}
class _md_worcester_co_ extends Tax { 
	public $name = "MD Worcester Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Worcester Co. Maryland Local Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"MD\" \"DC\" \"VA\" \"WV\" \"PA\"))";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MD Minimum Taxable Income"]->compute($emp))*(0.0225));
	 }
}
class _me extends Tax { 
	public $name = "ME";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Maine Income Tax";
	public $auxiliary_instructions = "Enter 1 if employee is a nonresident alien, 0 otherwise.";
	public $homerule = "(company-resident?)";
	public $auxiliary_tag = "Nonresident Alien";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(0||($emp->aux))?(($emp->earnings*(0.05))):((isTrue($emp->filingstatus==(3))?(tableLookup(max(0.0,$taxlist["ME Taxable Income"]->compute($emp)),array(array(0,0.058,0),array(49050,0.0675,2845),array(116100,0.0715,7371)))):(tableLookup(max(0.0,$taxlist["ME Taxable Income"]->compute($emp)),array(array(0,0.058,0),array(24500,0.0675,1421),array(58050,0.0715,3686)))))));
	 }
}
class _me_standard_deduction extends Tax { 
	public $name = "ME Standard Deduction";
	public $type = "P";
	public $withheld = true;
	public $title = "ME Standard Deduction";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((isTrue($emp->filingstatus==(3))?((isTrue(($emp->earnings<(183050)))?(24850):((isTrue(($emp->earnings>(333050)))?(0):((((333050-($emp->earnings))/(150000))*(24850))))))):((isTrue(($emp->earnings<(91500)))?(11000):((isTrue(($emp->earnings>(166500)))?(0):((((166500-($emp->earnings))/(75000))*(11000)))))))),0);
	 }
}
class _me_taxable_income extends Tax { 
	public $name = "ME Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "ME Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(($emp->earnings-((($emp->stateexemptions*(4700))+(max(0.0,$taxlist["ME Standard Deduction"]->compute($emp)))))),0);
	 }
}
class _me_unemployment extends Tax { 
	public $name = "ME Unemployment";
	public $type = "O";
	public $title = "Maine Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["ME Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _me_unemployment_wage_base extends Tax { 
	public $name = "ME Unemployment Wage Base";
	public $type = "P";
	public $title = "ME Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 15000;
	 }
}
class _mi extends Tax { 
	public $name = "MI";
	public $type = "S";
	public $withheld = true;
	public $title = "Michigan Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"IL\" \"IN\" \"KY\" \"MN\" \"OH\" \"WI\"))";
	public $workrule = "(not (within (homezip) \"IL\" \"IN\" \"KY\" \"MN\" \"OH\" \"WI\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(max(($emp->stateexemptions*(5400)),0)))*(0.0405));
	 }
}
class _mi_albion extends Tax { 
	public $name = "MI Albion";
	public $type = "L";
	public $withheld = true;
	public $title = "Albion Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form AL-W4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_albion_nr extends Tax { 
	public $name = "MI Albion NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Albion Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form AL-W4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_battle_creek extends Tax { 
	public $name = "MI Battle Creek";
	public $type = "L";
	public $withheld = true;
	public $title = "Battle Creek Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form BC-W4 line 8.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(750))))*(0.01));
	 }
}
class _mi_battle_creek_nr extends Tax { 
	public $name = "MI Battle Creek NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Battle Creek Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form BC-W4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(750))))*(0.005));
	 }
}
class _mi_benton_harbor extends Tax { 
	public $name = "MI Benton Harbor";
	public $type = "L";
	public $withheld = true;
	public $title = "Benton Harbor Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form BC-W4 line 8.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(750))))*(0.01));
	 }
}
class _mi_benton_harbor_nr extends Tax { 
	public $name = "MI Benton Harbor NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Benton Harbor Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form BC-W4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(750))))*(0.005));
	 }
}
class _mi_big_rapids extends Tax { 
	public $name = "MI Big Rapids";
	public $type = "L";
	public $withheld = true;
	public $title = "Big Rapids Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form BRW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_big_rapids_nr extends Tax { 
	public $name = "MI Big Rapids NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Big Rapids Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form BRW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_cities_nonresident extends Tax { 
	public $name = "MI Cities Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Michigan Cities Nonresident Income Tax";
	public $miscellaneous_instructions = "Enter the city income tax rate.";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-($emp->aux))*($emp->misc));
	 }
}
class _mi_cities_resident extends Tax { 
	public $name = "MI Cities Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "Michigan Cities Resident Income Tax";
	public $miscellaneous_instructions = "Enter the city income tax rate.";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-($emp->aux))*($emp->misc));
	 }
}
class _mi_detroit extends Tax { 
	public $name = "MI Detroit";
	public $type = "L";
	public $withheld = true;
	public $title = "Detroit Michigan Resident Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["MI Detroit Resident"]->compute($emp));
	 }
}
class _mi_detroit_nr extends Tax { 
	public $name = "MI Detroit NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Detroit Michigan Nonresident Income Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["MI Detroit Nonresident"]->compute($emp));
	 }
}
class _mi_detroit_nonresident extends Tax { 
	public $name = "MI Detroit Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Detroit Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form DW4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.012));
	 }
}
class _mi_detroit_resident extends Tax { 
	public $name = "MI Detroit Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "Detroit Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form DW4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.024));
	 }
}
class _mi_east_lansing extends Tax { 
	public $name = "MI East Lansing";
	public $type = "L";
	public $withheld = true;
	public $title = "East Lansing Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form LW-4 line 6.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_east_lansing_nonresident extends Tax { 
	public $name = "MI East Lansing Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "East Lansing Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form LW-4 line 6.";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_flint extends Tax { 
	public $name = "MI Flint";
	public $type = "L";
	public $withheld = true;
	public $title = "Flint Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form FW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_flint_nr extends Tax { 
	public $name = "MI Flint NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Flint Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form FW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_grand_rapids extends Tax { 
	public $name = "MI Grand Rapids";
	public $type = "L";
	public $withheld = true;
	public $title = "Grand Rapids Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form CF-W-4 line 8.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.015));
	 }
}
class _mi_grand_rapids_nr extends Tax { 
	public $name = "MI Grand Rapids NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Grand Rapids Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form CF-W-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.0075));
	 }
}
class _mi_grayling extends Tax { 
	public $name = "MI Grayling";
	public $type = "L";
	public $withheld = true;
	public $title = "Grayling Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form GR W-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(3000))))*(0.01));
	 }
}
class _mi_grayling_nr extends Tax { 
	public $name = "MI Grayling NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Grayling Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form GR W-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(3000))))*(0.005));
	 }
}
class _mi_hamtramck extends Tax { 
	public $name = "MI Hamtramck";
	public $type = "L";
	public $withheld = true;
	public $title = "Hamtramck Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form HW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_hamtramck_nr extends Tax { 
	public $name = "MI Hamtramck NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Hamtramck Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form HW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_highland_park extends Tax { 
	public $name = "MI Highland Park";
	public $type = "L";
	public $withheld = true;
	public $title = "Highland Park Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form CF-W-4 line 8.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.02));
	 }
}
class _mi_highland_park_nr extends Tax { 
	public $name = "MI Highland Park NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Highland Park Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form CF-W-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_hudson extends Tax { 
	public $name = "MI Hudson";
	public $type = "L";
	public $withheld = true;
	public $title = "Hudson Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form HU W-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(1000))))*(0.01));
	 }
}
class _mi_hudson_nr extends Tax { 
	public $name = "MI Hudson NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Hudson Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form HU W-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(1000))))*(0.005));
	 }
}
class _mi_ionia extends Tax { 
	public $name = "MI Ionia";
	public $type = "L";
	public $withheld = true;
	public $title = "Ionia Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form IW-4 line 6.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(700))))*(0.01));
	 }
}
class _mi_ionia_nr extends Tax { 
	public $name = "MI Ionia NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Ionia Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form IW-4 line 6.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(700))))*(0.005));
	 }
}
class _mi_jackson extends Tax { 
	public $name = "MI Jackson";
	public $type = "L";
	public $withheld = true;
	public $title = "Jackson Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form JW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_jackson_nr extends Tax { 
	public $name = "MI Jackson NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Jackson Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form JW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_lansing extends Tax { 
	public $name = "MI Lansing";
	public $type = "L";
	public $withheld = true;
	public $title = "Lansing Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form LW-4 line 6.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_lansing_nr extends Tax { 
	public $name = "MI Lansing NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Lansing Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form LW-4 line 6.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_lapeer extends Tax { 
	public $name = "MI Lapeer";
	public $type = "L";
	public $withheld = true;
	public $title = "Lapeer Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form LW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_lapeer_nr extends Tax { 
	public $name = "MI Lapeer NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Lapeer Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form LW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_muskegon extends Tax { 
	public $name = "MI Muskegon";
	public $type = "L";
	public $withheld = true;
	public $title = "Muskegon Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form MW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_muskegon_heights extends Tax { 
	public $name = "MI Muskegon Heights";
	public $type = "L";
	public $withheld = true;
	public $title = "Muskegon Heights Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form MHW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_muskegon_heights_nr extends Tax { 
	public $name = "MI Muskegon Heights NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Muskegon Heights Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form MHW-4 line 7.";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_muskegon_nr extends Tax { 
	public $name = "MI Muskegon NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Muskegon Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form MW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_pontiac extends Tax { 
	public $name = "MI Pontiac";
	public $type = "L";
	public $withheld = true;
	public $title = "Pontiac Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form PW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_pontiac_nr extends Tax { 
	public $name = "MI Pontiac NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Pontiac Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form PW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_port_huron extends Tax { 
	public $name = "MI Port Huron";
	public $type = "L";
	public $withheld = true;
	public $title = "Port Huron Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form PH-W4 line 6.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_port_huron_nr extends Tax { 
	public $name = "MI Port Huron NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Port Huron Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form PH-W4 line 6.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mi_portland extends Tax { 
	public $name = "MI Portland";
	public $type = "L";
	public $withheld = true;
	public $title = "Portland Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form PW-4 line 6.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(1000))))*(0.01));
	 }
}
class _mi_portland_nr extends Tax { 
	public $name = "MI Portland NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Portland Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form PW-4 line 6.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(1000))))*(0.005));
	 }
}
class _mi_saginaw extends Tax { 
	public $name = "MI Saginaw";
	public $type = "L";
	public $withheld = true;
	public $title = "Saginaw Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form SW-4 line 6.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(750))))*(0.015));
	 }
}
class _mi_saginaw_nr extends Tax { 
	public $name = "MI Saginaw NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Saginaw Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form SW-4 line 6.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(750))))*(0.0075));
	 }
}
class _mi_springfield extends Tax { 
	public $name = "MI Springfield";
	public $type = "L";
	public $withheld = true;
	public $title = "Springfield Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form SF W-4 line 8.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(1500))))*(0.01));
	 }
}
class _mi_springfield_nr extends Tax { 
	public $name = "MI Springfield NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Springfield Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form SF W-4 line 8.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(1500))))*(0.005));
	 }
}
class _mi_unemployment extends Tax { 
	public $name = "MI Unemployment";
	public $type = "O";
	public $title = "Michigan Unemployment Tax";
	public $auxiliary_instructions = "Enter 1 if employer is considered deliquent";
	public $homerule = "#f";
	public $auxiliary_tag = "Deliquent";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MI Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _mi_unemployment_wage_base extends Tax { 
	public $name = "MI Unemployment Wage Base";
	public $type = "P";
	public $title = "MI Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9500;
	 }
}
class _mi_walker extends Tax { 
	public $name = "MI Walker";
	public $type = "L";
	public $withheld = true;
	public $title = "Walker Michigan Resident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form WW-4 line 7.";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.01));
	 }
}
class _mi_walker_nr extends Tax { 
	public $name = "MI Walker NR";
	public $type = "L";
	public $withheld = true;
	public $title = "Walker Michigan Nonresident Income Tax";
	public $auxiliary_instructions = "Enter exemptions from form WW-4 line 7.";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	public $auxiliary_tag = "Exemptions";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(($emp->aux*(600))))*(0.005));
	 }
}
class _mn extends Tax { 
	public $name = "MN";
	public $type = "S";
	public $withheld = true;
	public $title = "Minnesota Income Tax";
	public $homerule = "(and (company-resident?) (or (not (worktax))(within (workzip) \"MI\" \"ND\")))";
	public $workrule = "(not (within (homezip) \"MI\" \"ND\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["MN Taxable Income"]->compute($emp)),array(array(0,0,0),array(4225,0.0535,0),array(34295,0.068,1608.75),array(102985,0.0785,6279.67),array(187565,0.0985,12919.2)))):(tableLookup(max(0.0,$taxlist["MN Taxable Income"]->compute($emp)),array(array(0,0,0),array(13250,0.0535,0),array(57200,0.068,2351.33),array(187860,0.0785,11236.21),array(318220,0.0985,21469.47)))));
	 }
}
class _mn_taxable_income extends Tax { 
	public $name = "MN Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "MN Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(4800)),0)));
	 }
}
class _mn_unemployment extends Tax { 
	public $name = "MN Unemployment";
	public $type = "O";
	public $title = "Minnesota Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MN Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _mn_unemployment_additional_assessment extends Tax { 
	public $name = "MN Unemployment Additional Assessment";
	public $type = "O";
	public $title = "Minnesota Unemployment Addition Assessment";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["MN Unemployment"]->compute($emp))*(0.14));
	 }
}
class _mn_unemployment_loan_assessment extends Tax { 
	public $name = "MN Unemployment Loan Assessment";
	public $type = "O";
	public $title = "Minnesota Special Assessment (Federal Loan Interest Assessment)";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(0.0,$taxlist["MN Unemployment"]->compute($emp))+(max(0.0,$taxlist["MN Unemployment Additional Assessment"]->compute($emp))))*(0.018));
	 }
}
class _mn_unemployment_wage_base extends Tax { 
	public $name = "MN Unemployment Wage Base";
	public $type = "P";
	public $title = "MN Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 40000;
	 }
}
class _mn_unemployment_workforce_development_fee extends Tax { 
	public $name = "MN Unemployment Workforce Development Fee";
	public $type = "O";
	public $title = "Minnesota Workforce Enhancement Fee";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.001));
	 }
}
class _mo extends Tax { 
	public $name = "MO";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Missouri Income Tax";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["MO Taxable Income"]->compute($emp)),array(array(0,0,0),array(1207.01,0.02,0),array(2414.01,0.025,24.14),array(3621.01,0.03,54.32),array(4828.01,0.035,90.53),array(6035.01,0.04,132.77),array(7242.01,0.045,181.05),array(8449.01,0.0495,235.37)));
	 }
}
class _mo_kansas_city extends Tax { 
	public $name = "MO Kansas City";
	public $type = "L";
	public $withheld = true;
	public $title = "Kansas City Earnings Tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _mo_st__louis extends Tax { 
	public $name = "MO St. Louis";
	public $type = "L";
	public $withheld = true;
	public $title = "St. Louis Earnings Tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _mo_st__louis_employer extends Tax { 
	public $name = "MO St. Louis Employer";
	public $type = "L";
	public $title = "St. Louis Payroll Expense Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _mo_taxable_income extends Tax { 
	public $name = "MO Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "MO Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-((isTrue($emp->filingstatus==(0)||(($emp->filingstatus>(3))))?(13850):((isTrue($emp->filingstatus==(1)||($emp->filingstatus==(2)))?(20800):(27700))))));
	 }
}
class _mo_unemployment extends Tax { 
	public $name = "MO Unemployment";
	public $type = "O";
	public $title = "Missouri Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MO Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _mo_unemployment_wage_base extends Tax { 
	public $name = "MO Unemployment Wage Base";
	public $type = "P";
	public $title = "MO Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 10500;
	 }
}
class _ms extends Tax { 
	public $name = "MS";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Mississippi Income Tax";
	public $miscellaneous_instructions = "Enter total annual amount of exemptions from Form 89-350 line 6, or leave as 0 to compute exemption by filing status.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $miscellaneous_tag = "Exemption amount";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max((($emp->earnings-((isTrue($emp->filingstatus==(0))?(2300):((isTrue($emp->filingstatus==(1))?(3400):((isTrue($emp->filingstatus==(3))?(4600):(2300))))))))-((isTrue($emp->misc)?($emp->misc):((isTrue($emp->filingstatus==(0))?(6000):((isTrue(($emp->filingstatus>(2)))?(12000):(9500)))))))),0),array(array(0,0,0),array(10000,0.05,0)));
	 }
}
class _ms_unemployment extends Tax { 
	public $name = "MS Unemployment";
	public $type = "O";
	public $title = "Mississippi Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MS Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ms_unemployment_wage_base extends Tax { 
	public $name = "MS Unemployment Wage Base";
	public $type = "P";
	public $title = "MS Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 14000;
	 }
}
class _mt extends Tax { 
	public $name = "MT";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Montana Income Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "(not (within (homezip) \"ND\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["MT Taxable Income"]->compute($emp)),array(array(0,0.018,0),array(7630,0.044,137),array(16350,0.06,521),array(130790,0.066,7387)));
	 }
}
class _mt_taxable_income extends Tax { 
	public $name = "MT Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "MT Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(2070)),0)));
	 }
}
class _mt_unemployment extends Tax { 
	public $name = "MT Unemployment";
	public $type = "O";
	public $title = "Montana Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MT Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _mt_unemployment_aft extends Tax { 
	public $name = "MT Unemployment AFT";
	public $type = "O";
	public $title = "Montana Unemployment Administrative Fund Tax";
	public $miscellaneous_instructions = "Enter AFT rate (default is 0.18%)";
	public $homerule = "#f";
	public $miscellaneous_tag = "AFT rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["MT Unemployment AFT Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*((isTrue($emp->misc)?($emp->misc):(0.0018)))),0);
	 }
}
class _mt_unemployment_aft_wage_base extends Tax { 
	public $name = "MT Unemployment AFT Wage Base";
	public $type = "P";
	public $title = "MT Unemployment AFT Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["MT Unemployment Wage Base"]->compute($emp));
	 }
}
class _mt_unemployment_wage_base extends Tax { 
	public $name = "MT Unemployment Wage Base";
	public $type = "P";
	public $title = "MT Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 40500;
	 }
}
class _misc__ extends Tax { 
	public $name = "Misc %";
	public $type = "O";
	public $withheld = true;
	public $title = "Miscellaneous: ";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->misc*($emp->earnings));
	 }
}
class _misc_flat extends Tax { 
	public $name = "Misc Flat";
	public $type = "O";
	public $withheld = true;
	public $title = "Miscellaneous: ";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return $emp->misc;
	 }
}
class _nc extends Tax { 
	public $name = "NC";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "North Carolina Income Tax";
	public $state_exemption_instructions = "Enter allowances from form NC-4 line 1.";
	public $miscellaneous_instructions = "Enter 1 if employee indicates a non-zero amount on form NC-4 NRA line 2, 0 otherwise.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Allowances";
	public $miscellaneous_tag = "Nonresident Alien";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max((($emp->earnings-((isTrue($emp->filingstatus==(1)&&($emp->misc==(0)))?(19125):(12750))))-(($emp->stateexemptions*(2500)))),0)+((isTrue($emp->misc)?(min($emp->earnings,12750)):(0))))*(0.0485));
	 }
}
class _nc_unemployment extends Tax { 
	public $name = "NC Unemployment";
	public $type = "O";
	public $title = "North Carolina Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NC Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _nc_unemployment_wage_base extends Tax { 
	public $name = "NC Unemployment Wage Base";
	public $type = "P";
	public $title = "NC Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 29600;
	 }
}
class _nd extends Tax { 
	public $name = "ND";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "North Dakota Income Tax";
	public $miscellaneous_instructions = "Enter 1 if employee has signed the 2020 Federal W-4";
	public $homerule = "(and (company-resident?) (within (workzip) \"MN\" \"MT\"))";
	public $workrule = "(not (within (homezip) \"MN\" \"MT\"))";
	public $miscellaneous_tag = "2020 W-4";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc)?((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(2)))?(tableLookup($emp->earnings,array(array(0,0,0),array(6925,0.011,0),array(51650,0.0204,491.98),array(115250,0.0227,1789.42),array(232900,0.0264,4460.07),array(498275,0.029,11465.97)))):((isTrue($emp->filingstatus==(1))?(tableLookup($emp->earnings,array(array(0,0,0),array(10400,0.011,0),array(70350,0.0204,659.45),array(165150,0.0227,2593.37),array(260950,0.0264,4768.03),array(501750,0.029,11125.15)))):(tableLookup($emp->earnings,array(array(0,0,0),array(13850,0.011,0),array(51225,0.0204,411.13),array(104125,0.0227,1490.29),array(151400,0.0264,2563.43),array(259525,0.029,5417.93)))))))):((isTrue(($emp->filingstatus<(2)))?(tableLookup(max(0.0,$taxlist["ND Taxable Income"]->compute($emp)),array(array(0,0,0),array(51650,0.0195,0),array(277625,0.025,4406.51)))):(tableLookup(max(0.0,$taxlist["ND Taxable Income"]->compute($emp)),array(array(0,0,0),array(51225,0.0195,0),array(188775,0.025,2682.23)))))));
	 }
}
class _nd_taxable_income extends Tax { 
	public $name = "ND Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "ND Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(($emp->stateexemptions*(4300))));
	 }
}
class _nd_unemployment extends Tax { 
	public $name = "ND Unemployment";
	public $type = "O";
	public $title = "North Dakota Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["ND Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _nd_unemployment_wage_base extends Tax { 
	public $name = "ND Unemployment Wage Base";
	public $type = "P";
	public $title = "ND Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 40800;
	 }
}
class _ne extends Tax { 
	public $name = "NE";
	public $type = "S";
	public $withheld = true;
	public $title = "Nebraska Income Tax";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(2)))?(tableLookup(max(0.0,$taxlist["NE Taxable Income"]->compute($emp)),array(array(0,0,0),array(3060,0.0226,0),array(5990,0.0322,66.22),array(19470,0.0491,500.28),array(28210,0.062,929.41),array(35820,0.0639,1401.23),array(67270,0.0675,3410.89)))):(tableLookup(max(0.0,$taxlist["NE Taxable Income"]->compute($emp)),array(array(0,0,0),array(7530,0.0226,0),array(11610,0.0322,92.21),array(28910,0.0491,649.27),array(44980,0.062,1438.31),array(55810,0.0639,2109.77),array(74010,0.0675,3272.75)))));
	 }
}
class _ne_taxable_income extends Tax { 
	public $name = "NE Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "NE Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(2140)),0)));
	 }
}
class _ne_unemployment extends Tax { 
	public $name = "NE Unemployment";
	public $type = "O";
	public $title = "Nebraska Unemployment Tax";
	public $miscellaneous_instructions = "Enter 1 for Category 20 employer";
	public $homerule = "#f";
	public $miscellaneous_tag = "Category 20";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NE Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ne_unemployment_wage_base extends Tax { 
	public $name = "NE Unemployment Wage Base";
	public $type = "P";
	public $title = "NE Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc)?(24000):(9000));
	 }
}
class _nh extends Tax { 
	public $name = "NH";
	public $type = "S";
	public $title = "New Hampshire";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _nh_unemployment extends Tax { 
	public $name = "NH Unemployment";
	public $type = "O";
	public $title = "New Hampshire Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NH Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _nh_unemployment_administrative_charge extends Tax { 
	public $name = "NH Unemployment Administrative Charge";
	public $type = "O";
	public $title = "New Hampshire Unemployment Administrative Charge";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NH Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.002)),0);
	 }
}
class _nh_unemployment_wage_base extends Tax { 
	public $name = "NH Unemployment Wage Base";
	public $type = "P";
	public $title = "NH Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 14000;
	 }
}
class _nj extends Tax { 
	public $name = "NJ";
	public $type = "S";
	public $withheld = true;
	public $title = "New Jersey Income Tax";
	public $state_exemption_instructions = "Enter allowances from form NJ-W4 line 4.";
	public $miscellaneous_instructions = "Using form NJ-W4 line 3:Enter 1 for Table A.Enter 2 for Table B.Enter 3 for Table C.Enter 4 for Table D.Enter 5 for Table E.Any other value will default to either Table A or B, depending on filing status.";
	public $homerule = "(and (company-resident?) (or (not (worktax)) (within (workzip) \"PA\")))";
	public $workrule = "(not (within (homezip) \"PA\"))";
	public $stateexemption_tag = "Allowances";
	public $miscellaneous_tag = "Table";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->misc==(1))?(max(0.0,$taxlist["NJ Table A"]->compute($emp))):((isTrue($emp->misc==(2))?(max(0.0,$taxlist["NJ Table B"]->compute($emp))):((isTrue($emp->misc==(3))?(tableLookup(max(0.0,$taxlist["NJ Taxable Income"]->compute($emp)),array(array(0,0.015,0),array(20000,0.023,300),array(40000,0.028,760),array(50000,0.035,1040),array(60000,0.056,1390),array(150000,0.066,6430),array(500000,0.099,29530),array(1000000,0.118,79030)))):((isTrue($emp->misc==(4))?(tableLookup(max(0.0,$taxlist["NJ Taxable Income"]->compute($emp)),array(array(0,0.015,0),array(20000,0.027,300),array(40000,0.034,840),array(50000,0.043,1180),array(60000,0.056,1610),array(150000,0.065,6650),array(500000,0.099,29400),array(5000000,0.118,474900)))):((isTrue($emp->misc==(5))?(tableLookup(max(0.0,$taxlist["NJ Taxable Income"]->compute($emp)),array(array(0,0.015,0),array(20000,0.02,300),array(35000,0.058,600),array(100000,0.065,4370),array(500000,0.099,30370),array(5000000,0.118,475870)))):((isTrue($emp->filingstatus==(0)||($emp->filingstatus==(4)))?(max(0.0,$taxlist["NJ Table A"]->compute($emp))):(max(0.0,$taxlist["NJ Table B"]->compute($emp))))))))))))));
	 }
}
class _nj_disability extends Tax { 
	public $name = "NJ Disability";
	public $type = "O";
	public $title = "New Jersey Disability Insurance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _nj_disability_employer extends Tax { 
	public $name = "NJ Disability Employer";
	public $type = "O";
	public $title = "New Jersey Disability Insurance - Employer";
	public $auxiliary_instructions = "Enter the tax rate.  For example 5.4% would be entered as .054";
	public $auxiliary_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NJ Disability Employer Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($emp->aux)),0);
	 }
}
class _nj_disability_employer_wage_base extends Tax { 
	public $name = "NJ Disability Employer Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "NJ Disability Employer Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["NJ Unemployment Wage Base"]->compute($emp));
	 }
}
class _nj_disability_wage_base extends Tax { 
	public $name = "NJ Disability Wage Base";
	public $type = "P";
	public $title = "NJ Disability Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["NJ TDI FLI Wage Base"]->compute($emp));
	 }
}
class _nj_family_leave extends Tax { 
	public $name = "NJ Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "New Jersey Family Leave Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NJ Family Leave Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0006)),0);
	 }
}
class _nj_family_leave_wage_base extends Tax { 
	public $name = "NJ Family Leave Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "NJ Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["NJ TDI FLI Wage Base"]->compute($emp));
	 }
}
class _nj_jersey_city extends Tax { 
	public $name = "NJ Jersey City";
	public $type = "L";
	public $title = "Jersey City New Jersey Payroll Tax";
	public $homerule = "#f";
	public $workrule = "(not (within (homezip) \"NJ Jersey City\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _nj_newark extends Tax { 
	public $name = "NJ Newark";
	public $type = "L";
	public $title = "Newark New Jersey Payroll Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.01));
	 }
}
class _nj_sui_sdi extends Tax { 
	public $name = "NJ SUI/SDI";
	public $type = "O";
	public $withheld = true;
	public $title = "NJ SUI/SDI";
	public $homerule = "#f";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["NJ Unemployment Employee"]->compute($emp))+(max(0.0,$taxlist["NJ Disability"]->compute($emp))));
	 }
}
class _nj_tdi_fli_wage_base extends Tax { 
	public $name = "NJ TDI FLI Wage Base";
	public $type = "P";
	public $title = "NJ Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 156800;
	 }
}
class _nj_table_a extends Tax { 
	public $name = "NJ Table A";
	public $type = "P";
	public $withheld = true;
	public $title = "NJ Table A";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["NJ Taxable Income"]->compute($emp)),array(array(0,0.015,0),array(20000,0.02,300),array(35000,0.039,600),array(40000,0.061,795),array(75000,0.07,2930),array(500000,0.099,32680),array(1000000,0.118,82180)));
	 }
}
class _nj_table_b extends Tax { 
	public $name = "NJ Table B";
	public $type = "P";
	public $withheld = true;
	public $title = "NJ Table B";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["NJ Taxable Income"]->compute($emp)),array(array(0,0.015,0),array(20000,0.02,300),array(50000,0.027,900),array(70000,0.039,1440),array(80000,0.061,1830),array(150000,0.07,6100),array(500000,0.099,30600),array(1000000,0.118,80100)));
	 }
}
class _nj_taxable_income extends Tax { 
	public $name = "NJ Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "NJ Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(1000)),0)));
	 }
}
class _nj_unemployment extends Tax { 
	public $name = "NJ Unemployment";
	public $type = "O";
	public $title = "New Jersey Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NJ Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _nj_unemployment_employee extends Tax { 
	public $name = "NJ Unemployment Employee";
	public $type = "O";
	public $withheld = true;
	public $title = "New Jersey Unemployment Tax, Employee contribution";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NJ Unemployment Employee Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.00425)),0);
	 }
}
class _nj_unemployment_employee_wage_base extends Tax { 
	public $name = "NJ Unemployment Employee Wage Base";
	public $type = "P";
	public $title = "NJ Unemployment Employee Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["NJ Unemployment Wage Base"]->compute($emp));
	 }
}
class _nj_unemployment_wage_base extends Tax { 
	public $name = "NJ Unemployment Wage Base";
	public $type = "P";
	public $title = "NJ Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 41100;
	 }
}
class _nm extends Tax { 
	public $name = "NM";
	public $type = "S";
	public $withheld = true;
	public $title = "New Mexico Income Tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->filingstatus==(3))?(tableLookup($emp->earnings,array(array(0,0,0),array(13850,0.017,0),array(21850,0.032,136),array(29850,0.047,392),array(37850,0.049,768),array(53850,0.049,1552),array(77850,0.049,2728),array(113850,0.049,4492),array(213850,0.049,9392),array(328850,0.059,15027)))):((isTrue($emp->filingstatus==(1))?(tableLookup($emp->earnings,array(array(0,0,0),array(10400,0.017,0),array(18400,0.032,136),array(26400,0.047,392),array(34400,0.049,768),array(50400,0.049,1552),array(74400,0.049,2728),array(110400,0.049,4492),array(210400,0.049,9392),array(325400,0.059,15027)))):(tableLookup($emp->earnings,array(array(0,0,0),array(6925,0.017,0),array(12425,0.032,93.5),array(17925,0.047,269.5),array(22925,0.049,504.5),array(32925,0.049,994.5),array(48925,0.049,1778.5),array(71925,0.049,2905.5),array(131925,0.049,5845.5),array(216925,0.059,10010.5)))))));
	 }
}
class _nm_unemployment extends Tax { 
	public $name = "NM Unemployment";
	public $type = "O";
	public $title = "New Mexico Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NM Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _nm_unemployment_wage_base extends Tax { 
	public $name = "NM Unemployment Wage Base";
	public $type = "P";
	public $title = "NM Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 30100;
	 }
}
class _nm_worker_s_comp_employer_outstanding extends Tax { 
	public $name = "NM Worker's Comp employer outstanding";
	public $type = "P";
	public $title = "NM Worker's Comp employer outstanding";
	public $miscellaneous_instructions = "Enter the quarter-to-date Worker's comp taxes paid for this employee.";
	public $miscellaneous_tag = "QTD Paid";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((2.3-($emp->misc))*($emp->payperiods));
	 }
}
class _nm_worker_s_comp_outstanding extends Tax { 
	public $name = "NM Worker's Comp outstanding";
	public $type = "P";
	public $title = "NM Worker's Comp outstanding";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((2-($emp->misc))*($emp->payperiods));
	 }
}
class _nm_worker_s_compensation extends Tax { 
	public $name = "NM Worker's Compensation";
	public $type = "O";
	public $withheld = true;
	public $title = "New Mexico Worker's Compensation Tax";
	public $miscellaneous_instructions = "Tax withheld this quarter, prior to the current payperiod.";
	public $auxiliary_instructions = "Enter 1 to pro-rate tax, 0 for lump-sum";
	public $miscellaneous_tag = "QTD Tax";
	public $auxiliary_tag = "Prorated";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->aux)?(min(8,max(0.0,$taxlist["NM Worker's Comp outstanding"]->compute($emp)))):(max(0.0,$taxlist["NM Worker's Comp outstanding"]->compute($emp))));
	 }
}
class _nm_worker_s_compensation_employer extends Tax { 
	public $name = "NM Worker's Compensation Employer";
	public $type = "O";
	public $title = "New Mexico Worker's Compensation Tax (Employer)";
	public $miscellaneous_instructions = "Tax liability incurred this quarter, prior to the current payperiod.";
	public $auxiliary_instructions = "Enter 1 to pro-rate tax, 0 for lump-sum";
	public $miscellaneous_tag = "QTD Tax";
	public $auxiliary_tag = "Prorated";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->aux)?(min(9.2,max(0.0,$taxlist["NM Worker's Comp employer outstanding"]->compute($emp)))):(max(0.0,$taxlist["NM Worker's Comp employer outstanding"]->compute($emp))));
	 }
}
class _nv extends Tax { 
	public $name = "NV";
	public $type = "S";
	public $title = "Nevada";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _nv_unemployment extends Tax { 
	public $name = "NV Unemployment";
	public $type = "O";
	public $title = "Nevada Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NV Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _nv_unemployment_cep extends Tax { 
	public $name = "NV Unemployment CEP";
	public $type = "O";
	public $title = "Nevada Career Enhancement Program Tax";
	public $miscellaneous_instructions = "Use rate from Notice of Employer's Contribution, form NUCS 4091 or 4092.";
	public $homerule = "#f";
	public $miscellaneous_tag = "CEP Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max(((max(0.0,$taxlist["NV Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($emp->misc));
	 }
}
class _nv_unemployment_wage_base extends Tax { 
	public $name = "NV Unemployment Wage Base";
	public $type = "P";
	public $title = "NV Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 40100;
	 }
}
class _ny extends Tax { 
	public $name = "NY";
	public $type = "S";
	public $withheld = true;
	public $title = "New York State Income Tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["NY Taxable Income"]->compute($emp)),array(array(0,0.04,0),array(8500,0.045,340),array(11700,0.0525,484),array(13900,0.055,600),array(80650,0.06,4271),array(96800,0.0714,5240),array(107650,0.0764,6014),array(157650,0.065,9832),array(215400,0.1101,13586),array(265400,0.0735,19092),array(1077550,0.1045,112603.98),array(5000000,0.111,555000),array(25000000,0.117,2925000)))):(tableLookup(max(0.0,$taxlist["NY Taxable Income"]->compute($emp)),array(array(0,0.04,0),array(8500,0.045,340),array(11700,0.0525,484),array(13900,0.055,600),array(80650,0.06,4271),array(96800,0.0667,5240),array(107650,0.0717,5963),array(157650,0.0811,9546),array(211550,0.065,13919),array(323200,0.1284,21177),array(373200,0.0735,27599),array(1077550,0.0765,79368),array(2155350,0.1045,225234.08),array(5000000,0.111,555000),array(25000000,0.117,2925000)))));
	 }
}
class _ny_disability extends Tax { 
	public $name = "NY Disability";
	public $type = "O";
	public $withheld = true;
	public $title = "New York Disability Insurance";
	public $miscellaneous_instructions = "Enter week to date tax PRIOR to current payperiod.";
	public $homerule = "#f";
	public $miscellaneous_tag = "WTD tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min(min(max(($emp->earnings*(0.00455)),0),31.2),max(0.0,$taxlist["NY Disability YTDTax limit"]->compute($emp)))-(($emp->misc*($emp->payperiods))));
	 }
}
class _ny_disability_ytdtax_limit extends Tax { 
	public $name = "NY Disability YTDTax limit";
	public $type = "P";
	public $title = "NY Disability YTDTax limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((31.2-($emp->ytdtax))*($emp->payperiods)),0);
	 }
}
class _ny_ecet extends Tax { 
	public $name = "NY ECET";
	public $type = "O";
	public $title = "New York Employer Compensation Expense Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((max(((($emp->ytdearnings-(40000))*($emp->payperiods))+($emp->earnings)),0)*(0.05))-(($emp->ytdtax*($emp->payperiods)))),0);
	 }
}
class _ny_family_leave extends Tax { 
	public $name = "NY Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "New York Paid Family Leave Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min((min($emp->earnings,max(((max(0.0,$taxlist["NY Family Leave Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.00455)),max(0.0,$taxlist["NY Family Leave YTDTax limit"]->compute($emp)));
	 }
}
class _ny_family_leave_wage_base extends Tax { 
	public $name = "NY Family Leave Wage Base";
	public $type = "P";
	public $title = "NY Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 87785.88;
	 }
}
class _ny_family_leave_ytdtax_limit extends Tax { 
	public $name = "NY Family Leave YTDTax limit";
	public $type = "P";
	public $title = "NY Family Leave YTDTax limit";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((399.43-($emp->ytdtax))*($emp->payperiods)),0);
	 }
}
class _ny_mctm extends Tax { 
	public $name = "NY MCTM";
	public $type = "O";
	public $title = "New York Metropolitan Commuter Transportation Mobility tax Zone 2";
	public $miscellaneous_instructions = "Enter tax rate (e.g., .34% is .0034).";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*($emp->misc));
	 }
}
class _ny_mctm_nyc extends Tax { 
	public $name = "NY MCTM NYC";
	public $type = "O";
	public $title = "New York Metropolitan Commuter Transportation Mobility tax Zone 1";
	public $miscellaneous_instructions = "Enter tax rate (e.g., .34% is .0034).";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*($emp->misc));
	 }
}
class _ny_nj_waterfront extends Tax { 
	public $name = "NY NJ Waterfront";
	public $type = "L";
	public $title = "New York-New Jersey Waterfront Payroll Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.016));
	 }
}
class _ny_nyc_nonresident extends Tax { 
	public $name = "NY NYC Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "New York City Nonresident Income Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(($emp->earnings-(tableLookup($emp->earnings,array(array(0,0,999999),array(4000,0,3000),array(10000,0,2000),array(20000,0,1000),array(30000,0,0))))),0)*(0.0025))*(0));
	 }
}
class _ny_nyc_resident extends Tax { 
	public $name = "NY NYC Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "New York City Resident Income Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["NYC Taxable Income"]->compute($emp)),array(array(0,0.0205,0),array(8000,0.028,164),array(8700,0.0325,184),array(15000,0.0395,388),array(25000,0.0415,783),array(60000,0.0425,2236)))):(tableLookup(max(0.0,$taxlist["NYC Taxable Income"]->compute($emp)),array(array(0,0.0205,0),array(8000,0.028,164),array(8700,0.0325,184),array(15000,0.0395,388),array(25000,0.0415,783),array(60000,0.0425,2236)))));
	 }
}
class _ny_re_employment_service_fund extends Tax { 
	public $name = "NY Re-employment Service Fund";
	public $type = "O";
	public $title = "New York Re-employment Service Fund Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NY Re-employment Service Fund Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.00075)),0);
	 }
}
class _ny_re_employment_service_fund_wage_base extends Tax { 
	public $name = "NY Re-employment Service Fund Wage Base";
	public $type = "P";
	public $title = "NY Re-employment Service Fund Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["NY Unemployment Wage Base"]->compute($emp));
	 }
}
class _ny_taxable_income extends Tax { 
	public $name = "NY Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "Negative exemptions are allowed!";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-((isTrue(($emp->filingstatus<(3)))?(7400):(7950))))-(($emp->stateexemptions*(1000))));
	 }
}
class _ny_unemployment extends Tax { 
	public $name = "NY Unemployment";
	public $type = "O";
	public $title = "New York Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["NY Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ny_unemployment_wage_base extends Tax { 
	public $name = "NY Unemployment Wage Base";
	public $type = "P";
	public $title = "NY Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 12300;
	 }
}
class _ny_yonkers_nonresident extends Tax { 
	public $name = "NY Yonkers Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Yonkers Nonresident Income Tax";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(($emp->earnings-(tableLookup($emp->earnings,array(array(0,0,999999),array(4000,0,3000),array(10000,0,2000),array(20000,0,1000),array(30000,0,0))))),0)*(0.005));
	 }
}
class _ny_yonkers_resident extends Tax { 
	public $name = "NY Yonkers Resident";
	public $type = "L";
	public $withheld = true;
	public $title = "Yonkers Resident Income Tax";
	public $homerule = "(company-resident?)";
	public $workrule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["NY Yonkers Taxable Income"]->compute($emp)),array(array(0,0.04,0),array(8500,0.045,340),array(11700,0.0525,484),array(13900,0.0585,600),array(80650,0.0625,4504),array(96800,0.0732,5514),array(107650,0.0782,6308),array(157650,0.0675,10219),array(215400,0.0994,14117),array(265400,0.0735,19085),array(1077550,0.1045,112603.98),array(5000000,0.111,555000),array(25000000,0.117,2925000)))):(tableLookup(max(0.0,$taxlist["NY Yonkers Taxable Income"]->compute($emp)),array(array(0,0.04,0),array(8500,0.045,340),array(11700,0.0525,484),array(13900,0.0585,600),array(80650,0.0625,4504),array(96800,0.0711,5514),array(107650,0.0761,6285),array(157650,0.0804,10090),array(211550,0.0675,14425),array(323200,0.1123,21961),array(373200,0.0735,27576),array(1077550,0.0765,79346),array(215535,0.1045,22523.41),array(5000000,0.111,555000),array(25000000,0.117,2925000)))))*(0.1675));
	 }
}
class _ny_yonkers_taxable_income extends Tax { 
	public $name = "NY Yonkers Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "NY Yonkers Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-((isTrue(($emp->filingstatus<(3)))?(7400):(7950))))-(($emp->stateexemptions*(1000))));
	 }
}
class _nyc_taxable_income extends Tax { 
	public $name = "NYC Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "NYC Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-((isTrue(($emp->filingstatus<(3)))?(5000):(5500))))-(($emp->stateexemptions*(1000))));
	 }
}
class _oh extends Tax { 
	public $name = "OH";
	public $type = "S";
	public $withheld = true;
	public $title = "Ohio Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"IN\" \"KY\" \"MI\" \"PA\" \"WV\"))";
	public $workrule = "(not (within (homezip) \"IN\" \"KY\" \"MI\" \"PA\" \"WV\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (tableLookup(max(0.0,$taxlist["OH Taxable Income"]->compute($emp)),array(array(0,0.00501,0),array(5000,0.01001,25.05),array(10000,0.02005,75.1),array(15000,0.02505,175.35),array(20000,0.0299,300.6),array(100000,0.0441,2692.6)))*(1.001));
	 }
}
class _oh_city extends Tax { 
	public $name = "OH City";
	public $type = "L";
	public $withheld = true;
	public $title = "Ohio City Income Tax";
	public $miscellaneous_instructions = "Enter tax rate (e.g., 2% is .02).";
	public $auxiliary_instructions = "Enter minimum monthly wage an employee must earn to be liable for this tax.  Enter 0 or blank if no minimum applies.";
	public $homerule = "#f";
	public $miscellaneous_tag = "Rate";
	public $auxiliary_tag = "Min taxable";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->earnings>(($emp->stateexemptions*(12)))))?(($emp->earnings*($emp->misc))):(0));
	 }
}
class _oh_sdit extends Tax { 
	public $name = "OH SDIT";
	public $type = "L";
	public $withheld = true;
	public $title = "Ohio School District Income Tax";
	public $miscellaneous_instructions = "Enter tax rate (e.g., 2% is .02).";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*($emp->misc));
	 }
}
class _oh_taxable_income extends Tax { 
	public $name = "OH Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "OH Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(650)),0)));
	 }
}
class _oh_unemployment extends Tax { 
	public $name = "OH Unemployment";
	public $type = "O";
	public $title = "Ohio Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["OH Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _oh_unemployment_wage_base extends Tax { 
	public $name = "OH Unemployment Wage Base";
	public $type = "P";
	public $title = "OH Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9000;
	 }
}
class _ok extends Tax { 
	public $name = "OK";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "Oklahoma Income Tax";
	public $state_exemption_instructions = "Enter number of exemptions from form OK W-4 line 5.";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Allowances";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(2)))?(tableLookup(max(0.0,$taxlist["OK Taxable Income"]->compute($emp)),array(array(0,0,0),array(6350,0.0025,0),array(7350,0.0075,2.5),array(8850,0.0175,13.75),array(10100,0.0275,35.63),array(11250,0.0375,67.25),array(13550,0.0475,153.5)))):(tableLookup(max(0.0,$taxlist["OK Taxable Income"]->compute($emp)),array(array(0,0,0),array(12700,0.0025,0),array(14700,0.0075,5),array(17700,0.0175,27.5),array(20200,0.0275,71.25),array(22500,0.0375,134.5),array(24900,0.0475,224.5)))));
	 }
}
class _ok_taxable_income extends Tax { 
	public $name = "OK Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "OK Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(1000)),0)));
	 }
}
class _ok_unemployment extends Tax { 
	public $name = "OK Unemployment";
	public $type = "O";
	public $title = "Oklahoma Unemployment Tax";
	public $miscellaneous_instructions = "Set to 1 to include Unemployment Tech Fund amount.";
	public $homerule = "#f";
	public $miscellaneous_tag = "Include Tech Fund";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["OK Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*((isTrue($emp->misc)?($unemployment_rate):(($unemployment_rate*(0.95)))))),0);
	 }
}
class _ok_unemployment_technology_fund extends Tax { 
	public $name = "OK Unemployment Technology Fund";
	public $type = "O";
	public $title = "Dropped";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _ok_unemployment_wage_base extends Tax { 
	public $name = "OK Unemployment Wage Base";
	public $type = "P";
	public $title = "OK Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 25700;
	 }
}
class _or extends Tax { 
	public $name = "OR";
	public $type = "S";
	public $withheld = true;
	public $title = "Oregon Income Tax";
	public $state_exemption_instructions = "Enter allowances from Oregon form OR-W-4 line 2.";
	public $auxiliary_instructions = "Enter Federal Income Tax withheld this pay period.";
	public $homerule = "(company-resident?)";
	public $stateexemption_tag = "Allowances";
	public $auxiliary_tag = "Federal Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((isTrue(($emp->earnings<(50000)))?((isTrue(max(0.0,$taxlist["OR Single 3"]->compute($emp)))?(tableLookup(max(0.0,$taxlist["OR BASE"]->compute($emp)),array(array(0,0,0),array(0.01,0.0475,236),array(4050,0.0675,428),array(10200,0.0875,844)))):(tableLookup(max(0.0,$taxlist["OR BASE"]->compute($emp)),array(array(0,0,0),array(0.01,0.0475,236),array(8100,0.0675,621),array(20400,0.0875,1451)))))):((isTrue(max(0.0,$taxlist["OR Single 3"]->compute($emp)))?(tableLookup(max(0.0,$taxlist["OR BASE"]->compute($emp)),array(array(10200,0.0875,608),array(125000,0.099,10653)))):(tableLookup(max(0.0,$taxlist["OR BASE"]->compute($emp)),array(array(20400,0.0875,1215),array(250000,0.099,21305)))))))-((isTrue(($emp->filingstatus<(3)))?((isTrue(($emp->earnings<(100000)))?(max(($emp->stateexemptions*(236)),0)):(0))):((isTrue(($emp->earnings<(200000)))?(($emp->stateexemptions*(236))):(0))))));
	 }
}
class _or_base extends Tax { 
	public $name = "OR BASE";
	public $type = "P";
	public $withheld = true;
	public $title = "OR BASE";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(max(0.0,$taxlist["OR FIT Allowance"]->compute($emp))))-((isTrue(max(0.0,$taxlist["OR Single 3"]->compute($emp)))?(2605):(5210))));
	 }
}
class _or_canby extends Tax { 
	public $name = "OR Canby";
	public $type = "L";
	public $title = "Oregon Canby Transportation Excise tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.006));
	 }
}
class _or_eugene extends Tax { 
	public $name = "OR Eugene";
	public $type = "L";
	public $withheld = true;
	public $title = "Eugene Oregon Employee tax";
	public $miscellaneous_instructions = "Withholding rate (default is 0.44%, use -1 for exempt)";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*((isTrue(($emp->misc<(0)))?(0):((isTrue($emp->misc)?($emp->misc):(0.0044))))));
	 }
}
class _or_eugene_employer extends Tax { 
	public $name = "OR Eugene Employer";
	public $type = "L";
	public $title = "Eugene Oregon Employer Payroll tax";
	public $miscellaneous_instructions = "Employer rate (defaults to 0.21%, set to -1 if exempt)";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*((isTrue(($emp->misc<(0)))?(0):((isTrue($emp->misc)?($emp->misc):(0.0021))))));
	 }
}
class _or_fit_allowance extends Tax { 
	public $name = "OR FIT Allowance";
	public $type = "P";
	public $withheld = true;
	public $title = "OR FIT Allowance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min((isTrue($emp->aux)?(($emp->aux*($emp->payperiods))):(max(0.0,$taxlist["FIT"]->compute($emp)))),(isTrue(max(0.0,$taxlist["OR Single 3"]->compute($emp)))?(tableLookup($emp->earnings,array(array(0,0,7800),array(125000,0,6250),array(130000,0,4700),array(135000,0,3100),array(140000,0,1550),array(145000,0,0)))):(tableLookup($emp->earnings,array(array(0,0,7800),array(250000,0,6250),array(260000,0,4700),array(270000,0,3100),array(280000,0,1550),array(290000,0,0))))));
	 }
}
class _or_family_leave extends Tax { 
	public $name = "OR Family Leave";
	public $type = "O";
	public $withheld = true;
	public $title = "Oregon Family Leave Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["OR Family Leave Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.006)),0);
	 }
}
class _or_family_leave_employer extends Tax { 
	public $name = "OR Family Leave Employer";
	public $type = "O";
	public $title = "Oregon Family Leave Employer Tax";
	public $miscellaneous_instructions = "Enter a non-zero value if employer has 25 or more employees, otherwise 0";
	public $miscellaneous_tag = "25+ Employees";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max(((max(0.0,$taxlist["OR Family Leave Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*((isTrue($emp->misc)?(0.004):(0))));
	 }
}
class _or_family_leave_employer_wage_base extends Tax { 
	public $name = "OR Family Leave Employer Wage Base";
	public $type = "P";
	public $title = "employer wage base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["OR Family Leave Wage Base"]->compute($emp));
	 }
}
class _or_family_leave_wage_base extends Tax { 
	public $name = "OR Family Leave Wage Base";
	public $type = "P";
	public $title = "OR Family Leave Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 132900;
	 }
}
class _or_ltd extends Tax { 
	public $name = "OR LTD";
	public $type = "L";
	public $title = "Oregon Lane County Mass Transportation District tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0078));
	 }
}
class _or_multnomah_co_ extends Tax { 
	public $name = "OR Multnomah Co.";
	public $type = "L";
	public $withheld = true;
	public $title = "Multnomah Oregon County tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+(max(0.0,$taxlist["OR Multnomah YTD"]->compute($emp))))>(400000)))?(((($emp->earnings-(max(((400000-(max(0.0,$taxlist["OR Multnomah YTD"]->compute($emp))))*($emp->payperiods)),0)))*(0.03))+(3000))):((isTrue(((($emp->earnings/($emp->payperiods))+(max(0.0,$taxlist["OR Multnomah YTD"]->compute($emp))))>(200000)))?((($emp->earnings-(max(((200000-(max(0.0,$taxlist["OR Multnomah YTD"]->compute($emp))))*($emp->payperiods)),0)))*(0.015))):(0))));
	 }
}
class _or_multnomah_ytd extends Tax { 
	public $name = "OR Multnomah YTD";
	public $type = "P";
	public $title = "multnomah ytd earnings";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->ytdearnings-(max(0.0,$taxlist["OR FIT Allowance"]->compute($emp))))-(5210));
	 }
}
class _or_portland_metro extends Tax { 
	public $name = "OR Portland Metro";
	public $type = "L";
	public $withheld = true;
	public $title = "Portland Oregon Metro tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(((($emp->earnings/($emp->payperiods))+(max(0.0,$taxlist["OR Multnomah YTD"]->compute($emp))))>(200000)))?((($emp->earnings-(max(((200000-(max(0.0,$taxlist["OR Multnomah YTD"]->compute($emp))))*($emp->payperiods)),0)))*(0.01))):(0));
	 }
}
class _or_sandy extends Tax { 
	public $name = "OR Sandy";
	public $type = "L";
	public $title = "Oregon Sandy Transportation Excise tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.006));
	 }
}
class _or_single_3 extends Tax { 
	public $name = "OR Single 3";
	public $type = "P";
	public $withheld = true;
	public $title = "OR Single 3";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3))&&(($emp->stateexemptions<(3))))?(1):(0));
	 }
}
class _or_south_clackamas extends Tax { 
	public $name = "OR South Clackamas";
	public $type = "L";
	public $title = "South Clackamas Transportation Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _or_transit_tax extends Tax { 
	public $name = "OR Transit Tax";
	public $type = "O";
	public $withheld = true;
	public $title = "Oregon Statewide Transit tax";
	public $homerule = "(company-resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.001));
	 }
}
class _or_tri_met extends Tax { 
	public $name = "OR Tri-Met";
	public $type = "L";
	public $title = "Oregon Tri-County Metropolitan Transportation District tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.008037));
	 }
}
class _or_unemployment extends Tax { 
	public $name = "OR Unemployment";
	public $type = "O";
	public $title = "Oregon Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["OR Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _or_unemployment_wage_base extends Tax { 
	public $name = "OR Unemployment Wage Base";
	public $type = "P";
	public $title = "OR Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 50900;
	 }
}
class _or_wbf extends Tax { 
	public $name = "OR WBF";
	public $type = "O";
	public $withheld = true;
	public $title = "OR Worker's Benefit Fund Assessment";
	public $miscellaneous_instructions = "fraction of assessment to withhold from employee's pay (up to 50%).  e.g., enter 30% as 0.3\nLeave empty for default of 50%.";
	public $auxiliary_instructions = "Indicate number of hours (including partial hours) worked";
	public $miscellaneous_tag = "Employee portion";
	public $auxiliary_tag = "Hours worked";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((((isTrue($emp->misc)?(min($emp->misc,0.5)):(0.5))*(($emp->aux+((isTrue(($emp->aux-($emp->aux)))?(1):(0))))))*(0.022))*($emp->payperiods));
	 }
}
class _or_wbf_employer extends Tax { 
	public $name = "OR WBF Employer";
	public $type = "O";
	public $title = "OR Worker's Benefit Fund Assessment - Employer Portion";
	public $miscellaneous_instructions = "fraction of assessment to withhold from employee's pay (up to 50%).  e.g., enter 30% as 0.3\nLeave empty for default of 50%.";
	public $auxiliary_instructions = "Indicate number of hours (including partial hours) worked";
	public $miscellaneous_tag = "Employer portion";
	public $auxiliary_tag = "Hours worked";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((((1-((isTrue($emp->misc)?(min($emp->misc,0.5)):(0.5))))*(($emp->aux+((isTrue(($emp->aux-($emp->aux)))?(1):(0))))))*(0.022))*($emp->payperiods));
	 }
}
class _or_wilsonville extends Tax { 
	public $name = "OR Wilsonville";
	public $type = "L";
	public $title = "Oregon Wilsonville Transportation Excise tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.005));
	 }
}
class _pa extends Tax { 
	public $name = "PA";
	public $type = "S";
	public $withheld = true;
	public $title = "Pennsylvania Income Tax";
	public $homerule = "(and (company-resident?) (or (not (worktax))(within (workzip) \"IN\" \"MD\" \"NJ\" \"OH\" \"VA\" \"WV\")))";
	public $workrule = "(not (within (homezip) \"IN\" \"MD\" \"NJ\" \"OH\" \"VA\" \"WV\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0307));
	 }
}
class _pa_emst extends Tax { 
	public $name = "PA EMST";
	public $type = "L";
	public $withheld = true;
	public $title = "Pennsylvania Emergency and Municipal Services Tax";
	public $miscellaneous_instructions = "Enter annual amount (default is \$52).";
	public $miscellaneous_tag = "Annual amount";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(min((((isTrue($emp->misc)?($emp->misc):(52))-($emp->ytdtax))*($emp->payperiods)),(isTrue($emp->misc)?($emp->misc):(52))),0);
	 }
}
class _pa_local extends Tax { 
	public $name = "PA Local";
	public $type = "L";
	public $withheld = true;
	public $title = "Pennsylvania Local Earned Income Tax";
	public $miscellaneous_instructions = "Enter local rate.  For example, enter 1% as .01";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*($emp->misc));
	 }
}
class _pa_opt extends Tax { 
	public $name = "PA OPT";
	public $type = "L";
	public $withheld = true;
	public $title = "Pennsylvania Occupational Privilege Tax";
	public $miscellaneous_instructions = "Enter annual amount (default is \$10).";
	public $miscellaneous_tag = "Annual amount";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(((isTrue($emp->misc)?($emp->misc):(10))-((isTrue($emp->aux)?($emp->aux):($emp->ytdtax)))),0);
	 }
}
class _pa_pittsburgh_payroll_expense extends Tax { 
	public $name = "PA Pittsburgh Payroll Expense";
	public $type = "L";
	public $title = "Pittsburgh Payroll Expense Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0055));
	 }
}
class _pa_unemployment extends Tax { 
	public $name = "PA Unemployment";
	public $type = "O";
	public $title = "Pennsylvania Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["PA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _pa_unemployment_employee extends Tax { 
	public $name = "PA Unemployment Employee";
	public $type = "O";
	public $withheld = true;
	public $title = "Pennsylvania Unemployment Employee Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0007));
	 }
}
class _pa_unemployment_wage_base extends Tax { 
	public $name = "PA Unemployment Wage Base";
	public $type = "P";
	public $title = "PA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 10000;
	 }
}
class _pr extends Tax { 
	public $name = "PR";
	public $type = "S";
	public $withheld = true;
	public $title = "Puerto Rico Income Tax";
	public $state_exemption_instructions = "Enter the number of dependents from form 499R-4.1 line B.";
	public $miscellaneous_instructions = "Enter 0 if not claiming the personal exemption. Enter 1 if claiming all of the personal exemption.Enter -1 (minus 1) if Married but claiming half of the personal exemption.";
	public $auxiliary_instructions = "Enter number of allowances based on deductions of the employee from form 499R-4.1 section C.  Use the box labelled 'number of allowances' (under 'EMPLOYER'S USE ONLY').";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	public $stateexemption_tag = "Dependents";
	public $miscellaneous_tag = "Personal exemption";
	public $auxiliary_tag = "Allowances";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["PR Taxable Income"]->compute($emp)),array(array(0,0,0),array(9000,0.07,0),array(25000,0.14,1120),array(41500,0.25,3430),array(61500,0.33,8430)));
	 }
}
class _pr_disability extends Tax { 
	public $name = "PR Disability";
	public $type = "O";
	public $withheld = true;
	public $title = "Puerto Rico Disability Insurance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["PR Disability Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.003)),0);
	 }
}
class _pr_disability_employer extends Tax { 
	public $name = "PR Disability Employer";
	public $type = "O";
	public $title = "Puerto Rico Disability Insurance (Employer)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["PR Disability Employer Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.003)),0);
	 }
}
class _pr_disability_employer_wage_base extends Tax { 
	public $name = "PR Disability Employer Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "PR Disability Employer Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["PR Disability Wage Base"]->compute($emp));
	 }
}
class _pr_disability_wage_base extends Tax { 
	public $name = "PR Disability Wage Base";
	public $type = "P";
	public $title = "PR Disability Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9000;
	 }
}
class _pr_nonresident_alien extends Tax { 
	public $name = "PR Nonresident Alien";
	public $type = "S";
	public $withheld = true;
	public $title = "Puerto Rico Income Tax (Nonresident, not U.S. Citizen";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.29));
	 }
}
class _pr_nonresident_citizen extends Tax { 
	public $name = "PR Nonresident Citizen";
	public $type = "S";
	public $withheld = true;
	public $title = "Puerto Rico Income Tax (Nonresident, U.S. Citizen";
	public $homerule = "#f";
	public $workrule = "(resident?)";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.2));
	 }
}
class _pr_taxable_income extends Tax { 
	public $name = "PR Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "PR Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (((($emp->earnings-((isTrue($emp->misc==(0))?(0):((isTrue($emp->filingstatus==(0))?(3500):((isTrue($emp->filingstatus==(3)||($emp->filingstatus==(2))||($emp->filingstatus==(5))&&(($emp->misc>(0)))||($emp->filingstatus==(1)))?(7000):((isTrue($emp->filingstatus==(4)&&(($emp->misc>(0)))||($emp->filingstatus==(4)&&(($emp->misc<(0)))))?(3500):(0))))))))))-((max(($emp->stateexemptions-((isTrue($emp->filingstatus==(1))?(1):(0)))),0)*(2500))))-(max(($emp->aux*(500)),0)))-(tableLookup($emp->earnings,array(array(0,0,2500),array(20000,0,2000),array(24500,0,1500),array(28500,0,1000),array(32500,0,500),array(36500,0,0)))));
	 }
}
class _pr_unemployment extends Tax { 
	public $name = "PR Unemployment";
	public $type = "O";
	public $title = "Puerto Rico Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["PR Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _pr_unemployment_wage_base extends Tax { 
	public $name = "PR Unemployment Wage Base";
	public $type = "P";
	public $title = "PR Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 7000;
	 }
}
class _ri extends Tax { 
	public $name = "RI";
	public $type = "S";
	public $withheld = true;
	public $title = "Rhode Island Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["RI Taxable Income"]->compute($emp)),array(array(0,0.0375,0),array(73450,0.0475,2754.38),array(166950,0.0599,7195.63)));
	 }
}
class _ri_disability extends Tax { 
	public $name = "RI Disability";
	public $type = "O";
	public $withheld = true;
	public $title = "Rhode Island Disability Insurance";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["RI Disability Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.011)),0);
	 }
}
class _ri_disability_wage_base extends Tax { 
	public $name = "RI Disability Wage Base";
	public $type = "P";
	public $title = "RI Disability Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 84000;
	 }
}
class _ri_taxable_income extends Tax { 
	public $name = "RI Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "RI Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(($emp->earnings-((isTrue(($emp->earnings<(260550)))?(($emp->stateexemptions*(1000))):(0)))),0);
	 }
}
class _ri_unemployment extends Tax { 
	public $name = "RI Unemployment";
	public $type = "O";
	public $title = "Rhode Island Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["RI Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ri_unemployment_job_development_fund extends Tax { 
	public $name = "RI Unemployment Job Development Fund";
	public $type = "O";
	public $title = "Rhode Island Job Development Fund";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["RI Unemployment Job Development Fund Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*(0.0021)),0);
	 }
}
class _ri_unemployment_job_development_fund_wage_base extends Tax { 
	public $name = "RI Unemployment Job Development Fund Wage Base";
	public $type = "O";
	public $title = "Rhode Island Job Development Fund Wage Base";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["RI Unemployment Wage Base"]->compute($emp));
	 }
}
class _ri_unemployment_wage_base extends Tax { 
	public $name = "RI Unemployment Wage Base";
	public $type = "P";
	public $title = "RI Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($unemployment_rate>(0.0978)))?(28200):(29700));
	 }
}
class _sc extends Tax { 
	public $name = "SC";
	public $type = "S";
	public $withheld = true;
	public $title = "South Carolina Income Tax";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["SC Taxable Income"]->compute($emp)),array(array(0,0,0),array(3330,0.03,0),array(16680,0.065,400.5)));
	 }
}
class _sc_taxable_income extends Tax { 
	public $name = "SC Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "SC Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-(max(($emp->stateexemptions*(4310)),0)))-(max((isTrue(($emp->stateexemptions>(0)))?(min(($emp->earnings*(0.1)),6475)):(0)),0)));
	 }
}
class _sc_unemployment extends Tax { 
	public $name = "SC Unemployment";
	public $type = "O";
	public $title = "South Carolina Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["SC Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _sc_unemployment_wage_base extends Tax { 
	public $name = "SC Unemployment Wage Base";
	public $type = "P";
	public $title = "SC Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 14000;
	 }
}
class _sd extends Tax { 
	public $name = "SD";
	public $type = "S";
	public $title = "South Dakota";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _sd_unemployment extends Tax { 
	public $name = "SD Unemployment";
	public $type = "O";
	public $title = "South Dakota Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["SD Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _sd_unemployment_wage_base extends Tax { 
	public $name = "SD Unemployment Wage Base";
	public $type = "P";
	public $title = "SD Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 15000;
	 }
}
class _seca extends Tax { 
	public $name = "SECA";
	public $type = "P";
	public $withheld = true;
	public $title = "SECA Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (max(0.0,$taxlist["SECA Med"]->compute($emp))+(max(0.0,$taxlist["SECA SS"]->compute($emp))));
	 }
}
class _seca_med extends Tax { 
	public $name = "SECA Med";
	public $type = "F";
	public $withheld = true;
	public $title = "SECA-Medicare Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.029));
	 }
}
class _seca_ss extends Tax { 
	public $name = "SECA SS";
	public $type = "F";
	public $withheld = true;
	public $title = "SECA-Social Security Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,(max(0.0,$taxlist["SECA SS Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))))*(0.124));
	 }
}
class _seca_ss_wage_base extends Tax { 
	public $name = "SECA SS Wage Base";
	public $type = "P";
	public $withheld = true;
	public $title = "SECA SS Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _tn extends Tax { 
	public $name = "TN";
	public $type = "S";
	public $title = "Tennessee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _tn_unemployment extends Tax { 
	public $name = "TN Unemployment";
	public $type = "O";
	public $title = "Tennessee Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["TN Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _tn_unemployment_wage_base extends Tax { 
	public $name = "TN Unemployment Wage Base";
	public $type = "P";
	public $title = "TN Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 7000;
	 }
}
class _tx extends Tax { 
	public $name = "TX";
	public $type = "S";
	public $title = "Texas";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _tx_texarkana extends Tax { 
	public $name = "TX Texarkana";
	public $type = "P";
	public $withheld = true;
	public $title = "used for zip code lookup";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _tx_unemployment extends Tax { 
	public $name = "TX Unemployment";
	public $type = "O";
	public $title = "Texas Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["TX Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _tx_unemployment_wage_base extends Tax { 
	public $name = "TX Unemployment Wage Base";
	public $type = "P";
	public $title = "TX Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9000;
	 }
}
class _ut extends Tax { 
	public $name = "UT";
	public $type = "S";
	public $withheld = true;
	public $title = "Utah Income Tax";
	public $homerule = "(and (company-resident?) (not (worktax)))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((($emp->earnings*(0.0465))-((isTrue(($emp->filingstatus<(3)))?(max((415-((($emp->earnings-(8371))*(0.013)))),0)):(max((830-((($emp->earnings-(16742))*(0.013)))),0))))),0);
	 }
}
class _ut_unemployment extends Tax { 
	public $name = "UT Unemployment";
	public $type = "O";
	public $title = "Utah Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["UT Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _ut_unemployment_wage_base extends Tax { 
	public $name = "UT Unemployment Wage Base";
	public $type = "P";
	public $title = "UT Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 44800;
	 }
}
class _va extends Tax { 
	public $name = "VA";
	public $type = "S";
	public $withheld = true;
	public $title = "Virginia Income Tax";
	public $miscellaneous_instructions = "Enter number of exemptions for age and blindness from form VA-4 line 7.";
	public $homerule = "(and (company-resident?) (or (not (worktax))(within (workzip) \"DC\" \"KY\" \"MD\" \"PA\" \"WV\")))";
	public $workrule = "(not (within (homezip) \"DC\" \"KY\" \"MD\" \"PA\" \"WV\"))";
	public $miscellaneous_tag = "VA-4 line 7";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["VA Taxable Income"]->compute($emp)),array(array(0,0.02,0),array(3000,0.03,60),array(5000,0.05,120),array(17000,0.0575,720)));
	 }
}
class _va_taxable_income extends Tax { 
	public $name = "VA Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "VA Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((($emp->earnings-(8000))-(max(($emp->stateexemptions*(930)),0)))-(max((min($emp->misc,2)*(800)),0)));
	 }
}
class _va_unemployment extends Tax { 
	public $name = "VA Unemployment";
	public $type = "O";
	public $title = "Virginia Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["VA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _va_unemployment_wage_base extends Tax { 
	public $name = "VA Unemployment Wage Base";
	public $type = "P";
	public $title = "VA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 8000;
	 }
}
class _vi extends Tax { 
	public $name = "VI";
	public $type = "S";
	public $title = "US Virgin Islands";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _vi_unemployment extends Tax { 
	public $name = "VI Unemployment";
	public $type = "O";
	public $title = "Virgin Islands Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["VI Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _vi_unemployment_wage_base extends Tax { 
	public $name = "VI Unemployment Wage Base";
	public $type = "P";
	public $title = "VI Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 30200;
	 }
}
class _vt extends Tax { 
	public $name = "VT";
	public $type = "S";
	public $withheld = true;
	public $title = "Vermont Income Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue(($emp->filingstatus<(3)))?(tableLookup(max(0.0,$taxlist["VT Taxable Income"]->compute($emp)),array(array(0,0,0),array(3500,0.0335,0),array(48900,0.066,1520.9),array(113550,0.076,5787.8),array(233050,0.0875,14869.8)))):(tableLookup(max(0.0,$taxlist["VT Taxable Income"]->compute($emp)),array(array(0,0,0),array(10538,0.0335,0),array(86388,0.066,2540.98),array(193938,0.076,9639.28),array(289988,0.0875,16939.08)))));
	 }
}
class _vt_taxable_income extends Tax { 
	public $name = "VT Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "VT Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(($emp->earnings-(($emp->stateexemptions*(4850)))),0);
	 }
}
class _vt_unemployment extends Tax { 
	public $name = "VT Unemployment";
	public $type = "O";
	public $title = "Vermont Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["VT Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _vt_unemployment_wage_base extends Tax { 
	public $name = "VT Unemployment Wage Base";
	public $type = "P";
	public $title = "VT Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 13500;
	 }
}
class _wa extends Tax { 
	public $name = "WA";
	public $type = "S";
	public $title = "Washington";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _wa_eaf extends Tax { 
	public $name = "WA EAF";
	public $type = "O";
	public $title = "WA Employment Administration Fund";
	public $miscellaneous_instructions = "Enter assigned rate (default is 0.0003)";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (min($emp->earnings,max(((max(0.0,$taxlist["WA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*((isTrue($emp->misc)?($emp->misc):(0.0003))));
	 }
}
class _wa_eaf_wage_base extends Tax { 
	public $name = "WA EAF Wage Base";
	public $type = "P";
	public $title = "WA EAF Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _wa_fmla extends Tax { 
	public $name = "WA FMLA";
	public $type = "O";
	public $withheld = true;
	public $title = "WA Paid Family & Medical Leave Assessment";
	public $miscellaneous_instructions = "Employee withholding portion (default is 100%)";
	public $miscellaneous_tag = "Employee portion";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(0.0,$taxlist["WA FMLA Earnings Under Cap"]->compute($emp))*(0.00582))*((isTrue($emp->misc)?(min(max($emp->misc,0),1)):(1))));
	 }
}
class _wa_fmla_earnings_under_cap extends Tax { 
	public $name = "WA FMLA Earnings Under Cap";
	public $type = "P";
	public $title = "WA FMLA Earnings Under Cap";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return min($emp->earnings,max((max(0.0,$taxlist["WA FMLA Wage Base"]->compute($emp))-(($emp->ytdearnings*($emp->payperiods)))),0));
	 }
}
class _wa_fmla_employer extends Tax { 
	public $name = "WA FMLA Employer";
	public $type = "O";
	public $title = "WA Paid Family & Medical Leave Assessment - Employer";
	public $miscellaneous_instructions = "Employer portion (default is 27.24%)";
	public $miscellaneous_tag = "Family leave portion";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ((max(0.0,$taxlist["WA FMLA Earnings Under Cap"]->compute($emp))*(0.002179))*((isTrue($emp->misc)?(min(max($emp->misc,0),1)):(0))));
	 }
}
class _wa_fmla_employer_wage_base extends Tax { 
	public $name = "WA FMLA Employer Wage Base";
	public $type = "P";
	public $title = "WA FMLA Employer Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _wa_fmla_wage_base extends Tax { 
	public $name = "WA FMLA Wage Base";
	public $type = "P";
	public $title = "WA FMLA Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max(0.0,$taxlist["FICA Wage Base"]->compute($emp));
	 }
}
class _wa_lni extends Tax { 
	public $name = "WA LNI";
	public $type = "O";
	public $withheld = true;
	public $title = "WA Labor & Industries Assessment - Employee withholding";
	public $miscellaneous_instructions = "Employee withholding rate";
	public $auxiliary_instructions = "Indicate number of hours (including partial hours) worked";
	public $miscellaneous_tag = "Employee rate";
	public $auxiliary_tag = "Hours worked";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->misc*(($emp->aux+((isTrue(($emp->aux-($emp->aux)))?(1):(0))))))*($emp->payperiods));
	 }
}
class _wa_lni_employer extends Tax { 
	public $name = "WA LNI Employer";
	public $type = "O";
	public $title = "WA Labor & Industries Assessment - Employer portion";
	public $miscellaneous_instructions = "Employer rate";
	public $auxiliary_instructions = "Indicate number of hours (including partial hours) worked";
	public $miscellaneous_tag = "Employer rate";
	public $auxiliary_tag = "Hours worked";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->misc*(($emp->aux+((isTrue(($emp->aux-($emp->aux)))?(1):(0))))))*($emp->payperiods));
	 }
}
class _wa_long_term_care extends Tax { 
	public $name = "WA Long Term Care";
	public $type = "O";
	public $withheld = true;
	public $title = "Washington Long Term Care Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*(0.0058));
	 }
}
class _wa_seattle extends Tax { 
	public $name = "WA Seattle";
	public $type = "L";
	public $title = "Seattle Washington Payroll Expense Tax";
	public $miscellaneous_instructions = "Employer rate (defaults to 1.7%, set to -1 if exempt)";
	public $miscellaneous_tag = "Rate";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings*((isTrue(($emp->misc<(0)))?(0):((isTrue($emp->misc)?($emp->misc):(0.017))))));
	 }
}
class _wa_unemployment extends Tax { 
	public $name = "WA Unemployment";
	public $type = "O";
	public $title = "Washington Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["WA Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _wa_unemployment_wage_base extends Tax { 
	public $name = "WA Unemployment Wage Base";
	public $type = "P";
	public $title = "WA Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 67600;
	 }
}
class _wi extends Tax { 
	public $name = "WI";
	public $type = "S";
	public $withheld = true;
	public $title = "Wisconsin Income Tax";
	public $homerule = "(and (company-resident?) (within (workzip) \"IL\" \"IN\" \"KY\" \"MI\"))";
	public $workrule = "(not (within (homezip) \"IL\" \"IN\" \"KY\" \"MI\"))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return tableLookup(max(0.0,$taxlist["WI Taxable Income"]->compute($emp)),array(array(0,0.0354,0),array(12760,0.0465,451.7),array(25520,0.053,1045.04),array(280950,0.0765,14582.83)));
	 }
}
class _wi_taxable_income extends Tax { 
	public $name = "WI Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "WI Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (($emp->earnings-((isTrue(($emp->filingstatus<(3)))?(tableLookup($emp->earnings,array(array(0,0,6702),array(17780,-0.12,6702),array(73630,0,0)))):(tableLookup($emp->earnings,array(array(0,0,9461),array(25727,-0.2,9461),array(73032,0,0)))))))-(($emp->stateexemptions*(400))));
	 }
}
class _wi_unemployment extends Tax { 
	public $name = "WI Unemployment";
	public $type = "O";
	public $title = "Wisconsin Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["WI Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _wi_unemployment_wage_base extends Tax { 
	public $name = "WI Unemployment Wage Base";
	public $type = "P";
	public $title = "WI Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 14000;
	 }
}
class _wv extends Tax { 
	public $name = "WV";
	public $type = "S";
	public $round = true;
	public $withheld = true;
	public $title = "West Virginia Income Tax";
	public $auxiliary_instructions = "Enter non-zero value for 2+ jobs or if spouse is employed";
	public $homerule = "(and (company-resident?) (within (workzip) \"KY\" \"MD\" \"OH\" \"PA\" \"VA\"))";
	public $workrule = "(not (within (homezip) \"KY\" \"MD\" \"OH\" \"PA\" \"VA\"))";
	public $auxiliary_tag = "Two earner/jobs";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return (isTrue($emp->filingstatus==(5)||($emp->aux))?(tableLookup(max(0.0,$taxlist["WV Taxable Income"]->compute($emp)),array(array(0,0.0236,0),array(7500,0.0315,177),array(18750,0.0354,531.38),array(30000,0.0472,929.63),array(45000,0.0512,1637.63)))):(tableLookup(max(0.0,$taxlist["WV Taxable Income"]->compute($emp)),array(array(0,0.0236,0),array(10000,0.0315,236),array(25000,0.0354,708.5),array(40000,0.0472,1239.5),array(60000,0.0512,2183.5)))));
	 }
}
class _wv_charleston extends Tax { 
	public $name = "WV Charleston";
	public $type = "L";
	public $withheld = true;
	public $title = "Charleston WV City Service Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 156;
	 }
}
class _wv_fairmont_nonresident extends Tax { 
	public $name = "WV Fairmont Nonresident";
	public $type = "L";
	public $withheld = true;
	public $title = "Fairmont WV City Service Fee";
	public $homerule = "#f";
	public $workrule = "(not (resident?))";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 104;
	 }
}
class _wv_huntington extends Tax { 
	public $name = "WV Huntington";
	public $type = "L";
	public $withheld = true;
	public $title = "Huntington WV City Service Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 156;
	 }
}
class _wv_madison extends Tax { 
	public $name = "WV Madison";
	public $type = "L";
	public $withheld = true;
	public $title = "Madison WV City Service Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 156;
	 }
}
class _wv_morgantown extends Tax { 
	public $name = "WV Morgantown";
	public $type = "L";
	public $withheld = true;
	public $title = "Huntington WV City Service Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 156;
	 }
}
class _wv_parkersburg extends Tax { 
	public $name = "WV Parkersburg";
	public $type = "L";
	public $withheld = true;
	public $title = "Parkersburg WV City Service Fee";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 130;
	 }
}
class _wv_taxable_income extends Tax { 
	public $name = "WV Taxable Income";
	public $type = "P";
	public $withheld = true;
	public $title = "WV Taxable Income";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->earnings-(max(($emp->stateexemptions*(2000)),0)));
	 }
}
class _wv_unemployment extends Tax { 
	public $name = "WV Unemployment";
	public $type = "O";
	public $title = "West Virginia Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["WV Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _wv_unemployment_debt_repayment extends Tax { 
	public $name = "WV Unemployment Debt Repayment";
	public $type = "O";
	public $title = "rescinded";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _wv_unemployment_wage_base extends Tax { 
	public $name = "WV Unemployment Wage Base";
	public $type = "P";
	public $title = "WV Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 9000;
	 }
}
class _wv_weirton extends Tax { 
	public $name = "WV Weirton";
	public $type = "L";
	public $withheld = true;
	public $title = "Weirton WV Occupational Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 104;
	 }
}
class _wv_wheeling extends Tax { 
	public $name = "WV Wheeling";
	public $type = "L";
	public $withheld = true;
	public $title = "Wheeling WV Occupational Tax";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 104;
	 }
}
class _wy extends Tax { 
	public $name = "WY";
	public $type = "S";
	public $title = "Wyoming";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 0;
	 }
}
class _wy_unemployment extends Tax { 
	public $name = "WY Unemployment";
	public $type = "O";
	public $title = "Wyoming Unemployment Tax";
	public $homerule = "#f";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return max((min($emp->earnings,max(((max(0.0,$taxlist["WY Unemployment Wage Base"]->compute($emp))*($emp->payperiods))-(($emp->ytdearnings*($emp->payperiods)))),0))*($unemployment_rate)),0);
	 }
}
class _wy_unemployment_wage_base extends Tax { 
	public $name = "WY Unemployment Wage Base";
	public $type = "P";
	public $title = "WY Unemployment Wage Base";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return 29100;
	 }
}
class _wy_workers_comp extends Tax { 
	public $name = "WY Workers Comp";
	public $type = "O";
	public $title = "Wyoming Worker's Compensation";
	public $miscellaneous_instructions = "Employee rate";
	public $auxiliary_instructions = "For owner or officer, enter 1.";
	public $miscellaneous_tag = "Rate";
	public $auxiliary_tag = "Owner/officer";
	function compute($emp){ 
	  global $taxlist, $unemployment_rate;
	  return ($emp->misc*((isTrue($emp->aux)?((4223.92*(12))):($emp->earnings))));
	 }
}
$taxlist = array(
"AK" => new _ak(),
"AK Unemployment" => new _ak_unemployment(),
"AK Unemployment Employee" => new _ak_unemployment_employee(),
"AK Unemployment Wage Base" => new _ak_unemployment_wage_base(),
"AL" => new _al(),
"AL Attalla" => new _al_attalla(),
"AL Auburn" => new _al_auburn(),
"AL Bear Creek" => new _al_bear_creek(),
"AL Beaverton" => new _al_beaverton(),
"AL Bessemer" => new _al_bessemer(),
"AL Birmingham" => new _al_birmingham(),
"AL Brilliant" => new _al_brilliant(),
"AL Fairfield" => new _al_fairfield(),
"AL Gadsden" => new _al_gadsden(),
"AL Glencoe" => new _al_glencoe(),
"AL Goodwater" => new _al_goodwater(),
"AL Guin" => new _al_guin(),
"AL Hackleburg" => new _al_hackleburg(),
"AL Haleyville" => new _al_haleyville(),
"AL Hamilton" => new _al_hamilton(),
"AL Hobson" => new _al_hobson(),
"AL Irondale" => new _al_irondale(),
"AL Leeds" => new _al_leeds(),
"AL Lynn" => new _al_lynn(),
"AL Macon Co." => new _al_macon_co_(),
"AL Midfield" => new _al_midfield(),
"AL Mosses" => new _al_mosses(),
"AL Mountain Brook" => new _al_mountain_brook(),
"AL Opelika" => new _al_opelika(),
"AL Rainbow City" => new _al_rainbow_city(),
"AL Red Bay" => new _al_red_bay(),
"AL Shorter" => new _al_shorter(),
"AL Southside" => new _al_southside(),
"AL Sulligent" => new _al_sulligent(),
"AL Tarrant" => new _al_tarrant(),
"AL Taxable Income" => new _al_taxable_income(),
"AL Tuskegee" => new _al_tuskegee(),
"AL Unemployment" => new _al_unemployment(),
"AL Unemployment Wage Base" => new _al_unemployment_wage_base(),
"AL Wadley" => new _al_wadley(),
"AR" => new _ar(),
"AR Taxable Income" => new _ar_taxable_income(),
"AR Texarkana" => new _ar_texarkana(),
"AR Unemployment" => new _ar_unemployment(),
"AR Unemployment Wage Base" => new _ar_unemployment_wage_base(),
"AR temp credit" => new _ar_temp_credit(),
"AZ" => new _az(),
"AZ Job Training Tax" => new _az_job_training_tax(),
"AZ Unemployment" => new _az_unemployment(),
"AZ Unemployment Wage Base" => new _az_unemployment_wage_base(),
"Benefits Dependent Care Exclusion" => new _benefits_dependent_care_exclusion(),
"CA" => new _ca(),
"CA Disability" => new _ca_disability(),
"CA Disability SE" => new _ca_disability_se(),
"CA Disability SE Wage Base" => new _ca_disability_se_wage_base(),
"CA Disability Wage Base" => new _ca_disability_wage_base(),
"CA Employment Training Fund" => new _ca_employment_training_fund(),
"CA Employment Training Fund Wage Base" => new _ca_employment_training_fund_wage_base(),
"CA LA" => new _ca_la(),
"CA Married 2" => new _ca_married_2(),
"CA SDI" => new _ca_sdi(),
"CA SDI Wage Base" => new _ca_sdi_wage_base(),
"CA Taxable Income" => new _ca_taxable_income(),
"CA Unemployment" => new _ca_unemployment(),
"CA Unemployment Wage Base" => new _ca_unemployment_wage_base(),
"CO" => new _co(),
"CO Aurora" => new _co_aurora(),
"CO Aurora Employer" => new _co_aurora_employer(),
"CO Denver" => new _co_denver(),
"CO Denver Employer" => new _co_denver_employer(),
"CO Family Leave" => new _co_family_leave(),
"CO Family Leave Employer" => new _co_family_leave_employer(),
"CO Family Leave Employer Wage Base" => new _co_family_leave_employer_wage_base(),
"CO Family Leave Wage Base" => new _co_family_leave_wage_base(),
"CO Glendale" => new _co_glendale(),
"CO Glendale Employer" => new _co_glendale_employer(),
"CO Greenwood Village" => new _co_greenwood_village(),
"CO Greenwood Village Employer" => new _co_greenwood_village_employer(),
"CO Sheridan" => new _co_sheridan(),
"CO Sheridan Employer" => new _co_sheridan_employer(),
"CO Unemployment" => new _co_unemployment(),
"CO Unemployment Wage Base" => new _co_unemployment_wage_base(),
"CT" => new _ct(),
"CT 3% Phase-Out" => new _ct_3__phase_out(),
"CT Family Leave" => new _ct_family_leave(),
"CT Family Leave Wage Base" => new _ct_family_leave_wage_base(),
"CT Personal Tax Credits" => new _ct_personal_tax_credits(),
"CT Recapture" => new _ct_recapture(),
"CT Recapture A D F" => new _ct_recapture_a_d_f(),
"CT Recapture B" => new _ct_recapture_b(),
"CT Recapture C" => new _ct_recapture_c(),
"CT Taxable Income" => new _ct_taxable_income(),
"CT Unemployment" => new _ct_unemployment(),
"CT Unemployment Wage Base" => new _ct_unemployment_wage_base(),
"DC" => new _dc(),
"DC Family Leave" => new _dc_family_leave(),
"DC Taxable Income" => new _dc_taxable_income(),
"DC Unemployment" => new _dc_unemployment(),
"DC Unemployment Wage Base" => new _dc_unemployment_wage_base(),
"DE" => new _de(),
"DE Training Tax Assessment" => new _de_training_tax_assessment(),
"DE Training Tax Assessment Wage Base" => new _de_training_tax_assessment_wage_base(),
"DE Unemployment" => new _de_unemployment(),
"DE Unemployment Wage Base" => new _de_unemployment_wage_base(),
"DE Wilmington" => new _de_wilmington(),
"DE Wilmington Employer" => new _de_wilmington_employer(),
"FICA" => new _fica(),
"FICA Med" => new _fica_med(),
"FICA Med Additional" => new _fica_med_additional(),
"FICA Med Additional Wage Base" => new _fica_med_additional_wage_base(),
"FICA Med employer" => new _fica_med_employer(),
"FICA SS" => new _fica_ss(),
"FICA SS Wage Base" => new _fica_ss_wage_base(),
"FICA SS YTDTax limit" => new _fica_ss_ytdtax_limit(),
"FICA SS employer" => new _fica_ss_employer(),
"FICA SS employer Wage Base" => new _fica_ss_employer_wage_base(),
"FICA Wage Base" => new _fica_wage_base(),
"FICA employer" => new _fica_employer(),
"FIT" => new _fit(),
"FIT Adjusted Earnings" => new _fit_adjusted_earnings(),
"FIT Supplemental under limit" => new _fit_supplemental_under_limit(),
"FL" => new _fl(),
"FL Unemployment" => new _fl_unemployment(),
"FL Unemployment Federal Interest Assessment" => new _fl_unemployment_federal_interest_assessment(),
"FL Unemployment Wage Base" => new _fl_unemployment_wage_base(),
"FUTA" => new _futa(),
"FUTA Credits" => new _futa_credits(),
"FUTA Normal Credit" => new _futa_normal_credit(),
"FUTA Rate" => new _futa_rate(),
"FUTA Taxable Wages" => new _futa_taxable_wages(),
"FUTA Wage Base" => new _futa_wage_base(),
"Federal Adjusted Earnings" => new _federal_adjusted_earnings(),
"Federal Income Tax" => new _federal_income_tax(),
"Federal Supplemental under limit" => new _federal_supplemental_under_limit(),
"Federal Tax Lien Exemption" => new _federal_tax_lien_exemption(),
"GA" => new _ga(),
"GA Taxable Income" => new _ga_taxable_income(),
"GA Unemployment" => new _ga_unemployment(),
"GA Unemployment Wage Base" => new _ga_unemployment_wage_base(),
"HI" => new _hi(),
"HI Disability" => new _hi_disability(),
"HI Disability Wage Limit" => new _hi_disability_wage_limit(),
"HI ETA" => new _hi_eta(),
"HI Taxable Income" => new _hi_taxable_income(),
"HI Unemployment" => new _hi_unemployment(),
"HI Unemployment Wage Base" => new _hi_unemployment_wage_base(),
"IA" => new _ia(),
"IA Unemployment" => new _ia_unemployment(),
"IA Unemployment Wage Base" => new _ia_unemployment_wage_base(),
"ID" => new _id(),
"ID Taxable Income" => new _id_taxable_income(),
"ID Unemployment" => new _id_unemployment(),
"ID Unemployment Wage Base" => new _id_unemployment_wage_base(),
"IL" => new _il(),
"IL Unemployment" => new _il_unemployment(),
"IL Unemployment Wage Base" => new _il_unemployment_wage_base(),
"IN" => new _in(),
"IN Adams Co." => new _in_adams_co_(),
"IN Adams Co. Nonresident" => new _in_adams_co__nonresident(),
"IN Allen Co." => new _in_allen_co_(),
"IN Allen Co. Nonresident" => new _in_allen_co__nonresident(),
"IN Bartholomew Co." => new _in_bartholomew_co_(),
"IN Bartholomew Co. Nonresident" => new _in_bartholomew_co__nonresident(),
"IN Benton Co." => new _in_benton_co_(),
"IN Benton Co. Nonresident" => new _in_benton_co__nonresident(),
"IN Blackford Co." => new _in_blackford_co_(),
"IN Blackford Co. Nonresident" => new _in_blackford_co__nonresident(),
"IN Boone Co." => new _in_boone_co_(),
"IN Boone Co. Nonresident" => new _in_boone_co__nonresident(),
"IN Brown Co." => new _in_brown_co_(),
"IN Brown Co. Nonresident" => new _in_brown_co__nonresident(),
"IN Carroll Co." => new _in_carroll_co_(),
"IN Carroll Co. Nonresident" => new _in_carroll_co__nonresident(),
"IN Cass Co." => new _in_cass_co_(),
"IN Cass Co. Nonresident" => new _in_cass_co__nonresident(),
"IN Clark Co." => new _in_clark_co_(),
"IN Clark Co. Nonresident" => new _in_clark_co__nonresident(),
"IN Clay Co." => new _in_clay_co_(),
"IN Clay Co. Nonresident" => new _in_clay_co__nonresident(),
"IN Clinton Co." => new _in_clinton_co_(),
"IN Clinton Co. Nonresident" => new _in_clinton_co__nonresident(),
"IN County Tax" => new _in_county_tax(),
"IN Crawford Co." => new _in_crawford_co_(),
"IN Crawford Co. Nonresident" => new _in_crawford_co__nonresident(),
"IN Daviess Co." => new _in_daviess_co_(),
"IN Daviess Co. Nonresident" => new _in_daviess_co__nonresident(),
"IN DeKalb Co." => new _in_dekalb_co_(),
"IN DeKalb Co. Nonresident" => new _in_dekalb_co__nonresident(),
"IN Dearborn Co." => new _in_dearborn_co_(),
"IN Dearborn Co. Nonresident" => new _in_dearborn_co__nonresident(),
"IN Decatur Co." => new _in_decatur_co_(),
"IN Decatur Co. Nonresident" => new _in_decatur_co__nonresident(),
"IN Delaware Co." => new _in_delaware_co_(),
"IN Delaware Co. Nonresident" => new _in_delaware_co__nonresident(),
"IN Dubois Co." => new _in_dubois_co_(),
"IN Dubois Co. Nonresident" => new _in_dubois_co__nonresident(),
"IN Elkhart Co." => new _in_elkhart_co_(),
"IN Elkhart Co. Nonresident" => new _in_elkhart_co__nonresident(),
"IN Fayette Co." => new _in_fayette_co_(),
"IN Fayette Co. Nonresident" => new _in_fayette_co__nonresident(),
"IN Floyd Co." => new _in_floyd_co_(),
"IN Floyd Co. Nonresident" => new _in_floyd_co__nonresident(),
"IN Fountain Co." => new _in_fountain_co_(),
"IN Fountain Co. Nonresident" => new _in_fountain_co__nonresident(),
"IN Franklin Co." => new _in_franklin_co_(),
"IN Franklin Co. Nonresident" => new _in_franklin_co__nonresident(),
"IN Fulton Co." => new _in_fulton_co_(),
"IN Fulton Co. Nonresident" => new _in_fulton_co__nonresident(),
"IN Gibson Co." => new _in_gibson_co_(),
"IN Gibson Co. Nonresident" => new _in_gibson_co__nonresident(),
"IN Grant Co." => new _in_grant_co_(),
"IN Grant Co. Nonresident" => new _in_grant_co__nonresident(),
"IN Greene Co." => new _in_greene_co_(),
"IN Greene Co. Nonresident" => new _in_greene_co__nonresident(),
"IN Hamilton Co." => new _in_hamilton_co_(),
"IN Hamilton Co. Nonresident" => new _in_hamilton_co__nonresident(),
"IN Hancock Co." => new _in_hancock_co_(),
"IN Hancock Co. Nonresident" => new _in_hancock_co__nonresident(),
"IN Harrison Co." => new _in_harrison_co_(),
"IN Harrison Co. Nonresident" => new _in_harrison_co__nonresident(),
"IN Hendricks Co." => new _in_hendricks_co_(),
"IN Hendricks Co. Nonresident" => new _in_hendricks_co__nonresident(),
"IN Henry Co." => new _in_henry_co_(),
"IN Henry Co. Nonresident" => new _in_henry_co__nonresident(),
"IN Howard Co." => new _in_howard_co_(),
"IN Howard Co. Nonresident" => new _in_howard_co__nonresident(),
"IN Huntington Co." => new _in_huntington_co_(),
"IN Huntington Co. Nonresident" => new _in_huntington_co__nonresident(),
"IN Jackson Co." => new _in_jackson_co_(),
"IN Jackson Co. Nonresident" => new _in_jackson_co__nonresident(),
"IN Jasper Co." => new _in_jasper_co_(),
"IN Jasper Co. Nonresident" => new _in_jasper_co__nonresident(),
"IN Jay Co." => new _in_jay_co_(),
"IN Jay Co. Nonresident" => new _in_jay_co__nonresident(),
"IN Jefferson Co." => new _in_jefferson_co_(),
"IN Jefferson Co. Nonresident" => new _in_jefferson_co__nonresident(),
"IN Jennings Co." => new _in_jennings_co_(),
"IN Jennings Co. Nonresident" => new _in_jennings_co__nonresident(),
"IN Johnson Co." => new _in_johnson_co_(),
"IN Johnson Co. Nonresident" => new _in_johnson_co__nonresident(),
"IN Knox Co." => new _in_knox_co_(),
"IN Knox Co. Nonresident" => new _in_knox_co__nonresident(),
"IN Kosciusko Co." => new _in_kosciusko_co_(),
"IN Kosciusko Co. Nonresident" => new _in_kosciusko_co__nonresident(),
"IN LaGrange Co." => new _in_lagrange_co_(),
"IN LaGrange Co. Nonresident" => new _in_lagrange_co__nonresident(),
"IN LaPorte Co." => new _in_laporte_co_(),
"IN LaPorte Co. Nonresident" => new _in_laporte_co__nonresident(),
"IN Lake Co." => new _in_lake_co_(),
"IN Lake Co. Nonresident" => new _in_lake_co__nonresident(),
"IN Lawrence Co." => new _in_lawrence_co_(),
"IN Lawrence Co. Nonresident" => new _in_lawrence_co__nonresident(),
"IN Madison Co." => new _in_madison_co_(),
"IN Madison Co. Nonresident" => new _in_madison_co__nonresident(),
"IN Marion Co." => new _in_marion_co_(),
"IN Marion Co. Nonresident" => new _in_marion_co__nonresident(),
"IN Marshall Co." => new _in_marshall_co_(),
"IN Marshall Co. Nonresident" => new _in_marshall_co__nonresident(),
"IN Martin Co." => new _in_martin_co_(),
"IN Martin Co. Nonresident" => new _in_martin_co__nonresident(),
"IN Miami Co." => new _in_miami_co_(),
"IN Miami Co. Nonresident" => new _in_miami_co__nonresident(),
"IN Monroe Co." => new _in_monroe_co_(),
"IN Monroe Co. Nonresident" => new _in_monroe_co__nonresident(),
"IN Montgomery Co." => new _in_montgomery_co_(),
"IN Montgomery Co. Nonresident" => new _in_montgomery_co__nonresident(),
"IN Morgan Co." => new _in_morgan_co_(),
"IN Morgan Co. Nonresident" => new _in_morgan_co__nonresident(),
"IN Newton Co." => new _in_newton_co_(),
"IN Newton Co. Nonresident" => new _in_newton_co__nonresident(),
"IN Noble Co." => new _in_noble_co_(),
"IN Noble Co. Nonresident" => new _in_noble_co__nonresident(),
"IN Ohio Co." => new _in_ohio_co_(),
"IN Ohio Co. Nonresident" => new _in_ohio_co__nonresident(),
"IN Orange Co." => new _in_orange_co_(),
"IN Orange Co. Nonresident" => new _in_orange_co__nonresident(),
"IN Owen Co." => new _in_owen_co_(),
"IN Owen Co. Nonresident" => new _in_owen_co__nonresident(),
"IN Parke Co." => new _in_parke_co_(),
"IN Parke Co. Nonresident" => new _in_parke_co__nonresident(),
"IN Perry Co." => new _in_perry_co_(),
"IN Perry Co. Nonresident" => new _in_perry_co__nonresident(),
"IN Pike Co." => new _in_pike_co_(),
"IN Pike Co. Nonresident" => new _in_pike_co__nonresident(),
"IN Porter Co." => new _in_porter_co_(),
"IN Porter Co. Nonresident" => new _in_porter_co__nonresident(),
"IN Posey Co." => new _in_posey_co_(),
"IN Posey Co. Nonresident" => new _in_posey_co__nonresident(),
"IN Pulaski Co." => new _in_pulaski_co_(),
"IN Pulaski Co. Nonresident" => new _in_pulaski_co__nonresident(),
"IN Putnam Co." => new _in_putnam_co_(),
"IN Putnam Co. Nonresident" => new _in_putnam_co__nonresident(),
"IN Randolph Co." => new _in_randolph_co_(),
"IN Randolph Co. Nonresident" => new _in_randolph_co__nonresident(),
"IN Ripley Co." => new _in_ripley_co_(),
"IN Ripley Co. Nonresident" => new _in_ripley_co__nonresident(),
"IN Rush Co." => new _in_rush_co_(),
"IN Rush Co. Nonresident" => new _in_rush_co__nonresident(),
"IN Scott Co." => new _in_scott_co_(),
"IN Scott Co. Nonresident" => new _in_scott_co__nonresident(),
"IN Shelby Co." => new _in_shelby_co_(),
"IN Shelby Co. Nonresident" => new _in_shelby_co__nonresident(),
"IN Spencer Co." => new _in_spencer_co_(),
"IN Spencer Co. Nonresident" => new _in_spencer_co__nonresident(),
"IN St. Joseph Co." => new _in_st__joseph_co_(),
"IN St. Joseph Co. Nonresident" => new _in_st__joseph_co__nonresident(),
"IN Starke Co." => new _in_starke_co_(),
"IN Starke Co. Nonresident" => new _in_starke_co__nonresident(),
"IN Steuben Co." => new _in_steuben_co_(),
"IN Steuben Co. Nonresident" => new _in_steuben_co__nonresident(),
"IN Sullivan Co." => new _in_sullivan_co_(),
"IN Sullivan Co. Nonresident" => new _in_sullivan_co__nonresident(),
"IN Switzerland Co." => new _in_switzerland_co_(),
"IN Switzerland Co. Nonresident" => new _in_switzerland_co__nonresident(),
"IN Taxable Income" => new _in_taxable_income(),
"IN Tippecanoe Co." => new _in_tippecanoe_co_(),
"IN Tippecanoe Co. Nonresident" => new _in_tippecanoe_co__nonresident(),
"IN Tipton Co." => new _in_tipton_co_(),
"IN Tipton Co. Nonresident" => new _in_tipton_co__nonresident(),
"IN Unemployment" => new _in_unemployment(),
"IN Unemployment Wage Base" => new _in_unemployment_wage_base(),
"IN Union Co." => new _in_union_co_(),
"IN Union Co. Nonresident" => new _in_union_co__nonresident(),
"IN Vanderburgh Co." => new _in_vanderburgh_co_(),
"IN Vanderburgh Co. Nonresident" => new _in_vanderburgh_co__nonresident(),
"IN Vermillion Co." => new _in_vermillion_co_(),
"IN Vermillion Co. Nonresident" => new _in_vermillion_co__nonresident(),
"IN Vigo Co." => new _in_vigo_co_(),
"IN Vigo Co. Nonresident" => new _in_vigo_co__nonresident(),
"IN Wabash Co." => new _in_wabash_co_(),
"IN Wabash Co. Nonresident" => new _in_wabash_co__nonresident(),
"IN Warren Co." => new _in_warren_co_(),
"IN Warren Co. Nonresident" => new _in_warren_co__nonresident(),
"IN Warrick Co." => new _in_warrick_co_(),
"IN Warrick Co. Nonresident" => new _in_warrick_co__nonresident(),
"IN Washington Co." => new _in_washington_co_(),
"IN Washington Co. Nonresident" => new _in_washington_co__nonresident(),
"IN Wayne Co." => new _in_wayne_co_(),
"IN Wayne Co. Nonresident" => new _in_wayne_co__nonresident(),
"IN Wells Co." => new _in_wells_co_(),
"IN Wells Co. Nonresident" => new _in_wells_co__nonresident(),
"IN White Co." => new _in_white_co_(),
"IN White Co. Nonresident" => new _in_white_co__nonresident(),
"IN Whitley Co." => new _in_whitley_co_(),
"IN Whitley Co. Nonresident" => new _in_whitley_co__nonresident(),
"KS" => new _ks(),
"KS Taxable Income" => new _ks_taxable_income(),
"KS Unemployment" => new _ks_unemployment(),
"KS Unemployment Wage Base" => new _ks_unemployment_wage_base(),
"KY" => new _ky(),
"KY Adairville" => new _ky_adairville(),
"KY Albany" => new _ky_albany(),
"KY Alexandria" => new _ky_alexandria(),
"KY Alexandria Wage Base" => new _ky_alexandria_wage_base(),
"KY Allen Co." => new _ky_allen_co_(),
"KY Ashland" => new _ky_ashland(),
"KY Auburn" => new _ky_auburn(),
"KY Augusta" => new _ky_augusta(),
"KY Ballard Co." => new _ky_ballard_co_(),
"KY Barbourville" => new _ky_barbourville(),
"KY Bardstown" => new _ky_bardstown(),
"KY Bardwell" => new _ky_bardwell(),
"KY Bath Co." => new _ky_bath_co_(),
"KY Beattyville" => new _ky_beattyville(),
"KY Bellevue" => new _ky_bellevue(),
"KY Benton" => new _ky_benton(),
"KY Benton Nonresident" => new _ky_benton_nonresident(),
"KY Berea" => new _ky_berea(),
"KY Boone Co." => new _ky_boone_co_(),
"KY Boone Co. Mental Health" => new _ky_boone_co__mental_health(),
"KY Boone Co. Payroll" => new _ky_boone_co__payroll(),
"KY Boone Co. SD" => new _ky_boone_co__sd(),
"KY Bourbon Co." => new _ky_bourbon_co_(),
"KY Bowling Green" => new _ky_bowling_green(),
"KY Boyd Co." => new _ky_boyd_co_(),
"KY Boyle Co." => new _ky_boyle_co_(),
"KY Breathitt Co." => new _ky_breathitt_co_(),
"KY Brodhead" => new _ky_brodhead(),
"KY Bromley" => new _ky_bromley(),
"KY Brooksville" => new _ky_brooksville(),
"KY Brownsville" => new _ky_brownsville(),
"KY Burkesville" => new _ky_burkesville(),
"KY Butler Co." => new _ky_butler_co_(),
"KY Cadiz" => new _ky_cadiz(),
"KY Caldwell Co." => new _ky_caldwell_co_(),
"KY Calvert" => new _ky_calvert(),
"KY Camargo" => new _ky_camargo(),
"KY Campbell Co." => new _ky_campbell_co_(),
"KY Campbell Co. Wage Base" => new _ky_campbell_co__wage_base(),
"KY Campbellsville" => new _ky_campbellsville(),
"KY Caneyville" => new _ky_caneyville(),
"KY Carlisle" => new _ky_carlisle(),
"KY Carroll Co." => new _ky_carroll_co_(),
"KY Carrolton" => new _ky_carrolton(),
"KY Catlettsburg" => new _ky_catlettsburg(),
"KY Cave City" => new _ky_cave_city(),
"KY Clark Co." => new _ky_clark_co_(),
"KY Clarkson" => new _ky_clarkson(),
"KY Clay City" => new _ky_clay_city(),
"KY Clay Co." => new _ky_clay_co_(),
"KY Clinton Co." => new _ky_clinton_co_(),
"KY Coal Run Village" => new _ky_coal_run_village(),
"KY Cold Spring" => new _ky_cold_spring(),
"KY Columbia" => new _ky_columbia(),
"KY Corbin" => new _ky_corbin(),
"KY Counties" => new _ky_counties(),
"KY Covington" => new _ky_covington(),
"KY Crescent Springs" => new _ky_crescent_springs(),
"KY Cresent Springs" => new _ky_cresent_springs(),
"KY Crestview Hills" => new _ky_crestview_hills(),
"KY Crestview Hills Wage Base" => new _ky_crestview_hills_wage_base(),
"KY Cumberland Co." => new _ky_cumberland_co_(),
"KY Cumberland Co. SD" => new _ky_cumberland_co__sd(),
"KY Cynthiana" => new _ky_cynthiana(),
"KY Danville" => new _ky_danville(),
"KY Daviess Co." => new _ky_daviess_co_(),
"KY Dawson Springs" => new _ky_dawson_springs(),
"KY Dayton" => new _ky_dayton(),
"KY Dry Ridge" => new _ky_dry_ridge(),
"KY Earlington" => new _ky_earlington(),
"KY Eddyville" => new _ky_eddyville(),
"KY Edgewood" => new _ky_edgewood(),
"KY Edmonton" => new _ky_edmonton(),
"KY Elizabethtown" => new _ky_elizabethtown(),
"KY Elkhorn City" => new _ky_elkhorn_city(),
"KY Elkton" => new _ky_elkton(),
"KY Elsmere" => new _ky_elsmere(),
"KY Eminence" => new _ky_eminence(),
"KY Erlanger" => new _ky_erlanger(),
"KY Estill Co." => new _ky_estill_co_(),
"KY Fayette Co." => new _ky_fayette_co_(),
"KY Fayette Co. SD" => new _ky_fayette_co__sd(),
"KY Flemingsburg" => new _ky_flemingsburg(),
"KY Florence" => new _ky_florence(),
"KY Fort Mitchell" => new _ky_fort_mitchell(),
"KY Fort Thomas" => new _ky_fort_thomas(),
"KY Fort Wright" => new _ky_fort_wright(),
"KY Fort Wright Wage Base" => new _ky_fort_wright_wage_base(),
"KY Frankfort" => new _ky_frankfort(),
"KY Franklin" => new _ky_franklin(),
"KY Franklin Co." => new _ky_franklin_co_(),
"KY Franklin Wage Base" => new _ky_franklin_wage_base(),
"KY Fulton" => new _ky_fulton(),
"KY Gallatin Co." => new _ky_gallatin_co_(),
"KY Gamaliel" => new _ky_gamaliel(),
"KY Garrard Co." => new _ky_garrard_co_(),
"KY Georgetown" => new _ky_georgetown(),
"KY Glasgow" => new _ky_glasgow(),
"KY Grant Co." => new _ky_grant_co_(),
"KY Graves Co." => new _ky_graves_co_(),
"KY Grayson" => new _ky_grayson(),
"KY Grayson Co." => new _ky_grayson_co_(),
"KY Greensburg" => new _ky_greensburg(),
"KY Greenup" => new _ky_greenup(),
"KY Guthrie" => new _ky_guthrie(),
"KY Hancock Co." => new _ky_hancock_co_(),
"KY Harrison Co." => new _ky_harrison_co_(),
"KY Harrodsburg" => new _ky_harrodsburg(),
"KY Hart Co." => new _ky_hart_co_(),
"KY Hartford" => new _ky_hartford(),
"KY Hazard" => new _ky_hazard(),
"KY Henderson" => new _ky_henderson(),
"KY Hickman" => new _ky_hickman(),
"KY Hickman Co." => new _ky_hickman_co_(),
"KY Highland Heights" => new _ky_highland_heights(),
"KY Hillview" => new _ky_hillview(),
"KY Hodgenville" => new _ky_hodgenville(),
"KY Hopkins Co." => new _ky_hopkins_co_(),
"KY Hopkinsville" => new _ky_hopkinsville(),
"KY Horse Cave" => new _ky_horse_cave(),
"KY Independence" => new _ky_independence(),
"KY Jackson" => new _ky_jackson(),
"KY Jackson Co." => new _ky_jackson_co_(),
"KY Jamestown" => new _ky_jamestown(),
"KY Jefferson Co. Nonresident" => new _ky_jefferson_co__nonresident(),
"KY Jefferson Co. Resident" => new _ky_jefferson_co__resident(),
"KY Jeffersontown" => new _ky_jeffersontown(),
"KY Jeffersonville" => new _ky_jeffersonville(),
"KY Jessamine Co." => new _ky_jessamine_co_(),
"KY Johnson Co." => new _ky_johnson_co_(),
"KY Junction City" => new _ky_junction_city(),
"KY Kenton Co." => new _ky_kenton_co_(),
"KY Kenton Co. Wage Base" => new _ky_kenton_co__wage_base(),
"KY Kenton Co. bracket 1" => new _ky_kenton_co__bracket_1(),
"KY Kenton Co. bracket 2" => new _ky_kenton_co__bracket_2(),
"KY Knox Co." => new _ky_knox_co_(),
"KY La Grange" => new _ky_la_grange(),
"KY Lakeside Park" => new _ky_lakeside_park(),
"KY Laurel Co." => new _ky_laurel_co_(),
"KY Lebanon" => new _ky_lebanon(),
"KY Lebanon Junction" => new _ky_lebanon_junction(),
"KY Leitchfield" => new _ky_leitchfield(),
"KY Leslie Co." => new _ky_leslie_co_(),
"KY Lewisburg" => new _ky_lewisburg(),
"KY Lexington" => new _ky_lexington(),
"KY Lincoln Co." => new _ky_lincoln_co_(),
"KY Livingston Co." => new _ky_livingston_co_(),
"KY Logan Co." => new _ky_logan_co_(),
"KY Louisville Nonresident" => new _ky_louisville_nonresident(),
"KY Louisville Resident" => new _ky_louisville_resident(),
"KY Ludlow" => new _ky_ludlow(),
"KY Madison Co." => new _ky_madison_co_(),
"KY Madisonville" => new _ky_madisonville(),
"KY Magoffin Co." => new _ky_magoffin_co_(),
"KY Marion" => new _ky_marion(),
"KY Marion Co." => new _ky_marion_co_(),
"KY Marshall Co. Nonresident" => new _ky_marshall_co__nonresident(),
"KY Marshall Co. Resident" => new _ky_marshall_co__resident(),
"KY Martin" => new _ky_martin(),
"KY Martin Co." => new _ky_martin_co_(),
"KY Mayfield" => new _ky_mayfield(),
"KY Maysville" => new _ky_maysville(),
"KY McCracken Co." => new _ky_mccracken_co_(),
"KY McCreary Co." => new _ky_mccreary_co_(),
"KY McKee" => new _ky_mckee(),
"KY McLean Co." => new _ky_mclean_co_(),
"KY Menifee Co." => new _ky_menifee_co_(),
"KY Mercer Co." => new _ky_mercer_co_(),
"KY Metcalfe Co." => new _ky_metcalfe_co_(),
"KY Middlesboro" => new _ky_middlesboro(),
"KY Midway" => new _ky_midway(),
"KY Millersburg" => new _ky_millersburg(),
"KY Monroe Co." => new _ky_monroe_co_(),
"KY Montgomery Co." => new _ky_montgomery_co_(),
"KY Morehead" => new _ky_morehead(),
"KY Morgan Co." => new _ky_morgan_co_(),
"KY Morgantown" => new _ky_morgantown(),
"KY Mount Olivet" => new _ky_mount_olivet(),
"KY Mount Vernon" => new _ky_mount_vernon(),
"KY Mount Washington" => new _ky_mount_washington(),
"KY Muldraugh" => new _ky_muldraugh(),
"KY Munfordville" => new _ky_munfordville(),
"KY Municipalities" => new _ky_municipalities(),
"KY Murray" => new _ky_murray(),
"KY Nelson Co." => new _ky_nelson_co_(),
"KY Newport" => new _ky_newport(),
"KY Newport Wage Base" => new _ky_newport_wage_base(),
"KY Nicholas Co." => new _ky_nicholas_co_(),
"KY Nicholasville" => new _ky_nicholasville(),
"KY Nortonville" => new _ky_nortonville(),
"KY Oak Grove" => new _ky_oak_grove(),
"KY Ohio Co." => new _ky_ohio_co_(),
"KY Olive Hill" => new _ky_olive_hill(),
"KY Owensboro" => new _ky_owensboro(),
"KY Owenton" => new _ky_owenton(),
"KY Paducah" => new _ky_paducah(),
"KY Paintsville" => new _ky_paintsville(),
"KY Paris" => new _ky_paris(),
"KY Park City" => new _ky_park_city(),
"KY Park Hills" => new _ky_park_hills(),
"KY Pendleton Co." => new _ky_pendleton_co_(),
"KY Perry Co." => new _ky_perry_co_(),
"KY Perryville" => new _ky_perryville(),
"KY Pikeville" => new _ky_pikeville(),
"KY Pineville" => new _ky_pineville(),
"KY Pioneer Village" => new _ky_pioneer_village(),
"KY Powell Co." => new _ky_powell_co_(),
"KY Prestonsburg" => new _ky_prestonsburg(),
"KY Princeton" => new _ky_princeton(),
"KY Pulaski Co." => new _ky_pulaski_co_(),
"KY Raceland" => new _ky_raceland(),
"KY Radcliff" => new _ky_radcliff(),
"KY Richmond" => new _ky_richmond(),
"KY Robertson Co." => new _ky_robertson_co_(),
"KY Rockcastle Co." => new _ky_rockcastle_co_(),
"KY Rowan Co." => new _ky_rowan_co_(),
"KY Russell" => new _ky_russell(),
"KY Russell County" => new _ky_russell_county(),
"KY Russell Springs" => new _ky_russell_springs(),
"KY Russellville" => new _ky_russellville(),
"KY Ryland Heights" => new _ky_ryland_heights(),
"KY Salyersville" => new _ky_salyersville(),
"KY Scott Co. Nonresident" => new _ky_scott_co__nonresident(),
"KY Scott Co. Resident" => new _ky_scott_co__resident(),
"KY Scottsville" => new _ky_scottsville(),
"KY Shelby Co." => new _ky_shelby_co_(),
"KY Shelbyville" => new _ky_shelbyville(),
"KY Shepherdsville" => new _ky_shepherdsville(),
"KY Shively" => new _ky_shively(),
"KY Silver Grove" => new _ky_silver_grove(),
"KY Simpson Co." => new _ky_simpson_co_(),
"KY Simpson Co. Wage Base" => new _ky_simpson_co__wage_base(),
"KY Simpsonville" => new _ky_simpsonville(),
"KY Somerset" => new _ky_somerset(),
"KY Southgate" => new _ky_southgate(),
"KY Spencer Co." => new _ky_spencer_co_(),
"KY Springfield" => new _ky_springfield(),
"KY St. Matthews" => new _ky_st__matthews(),
"KY Stanford" => new _ky_stanford(),
"KY Stanton" => new _ky_stanton(),
"KY Taylor Co." => new _ky_taylor_co_(),
"KY Taylor Mill" => new _ky_taylor_mill(),
"KY Taylorsville" => new _ky_taylorsville(),
"KY Todd Co." => new _ky_todd_co_(),
"KY Tompkinsville" => new _ky_tompkinsville(),
"KY Unemployment" => new _ky_unemployment(),
"KY Unemployment Wage Base" => new _ky_unemployment_wage_base(),
"KY Union" => new _ky_union(),
"KY Union Co." => new _ky_union_co_(),
"KY Vanceburg" => new _ky_vanceburg(),
"KY Versailles" => new _ky_versailles(),
"KY Villa Hills" => new _ky_villa_hills(),
"KY Vine Grove" => new _ky_vine_grove(),
"KY Warren Co." => new _ky_warren_co_(),
"KY Warren Co. SD" => new _ky_warren_co__sd(),
"KY Warsaw" => new _ky_warsaw(),
"KY Washington Co." => new _ky_washington_co_(),
"KY Wayne Co." => new _ky_wayne_co_(),
"KY West Beuchel" => new _ky_west_beuchel(),
"KY West Buechel" => new _ky_west_buechel(),
"KY West Bueschel" => new _ky_west_bueschel(),
"KY West Liberty" => new _ky_west_liberty(),
"KY West Point" => new _ky_west_point(),
"KY Wheelwright" => new _ky_wheelwright(),
"KY Whitesburg" => new _ky_whitesburg(),
"KY Whitley Co." => new _ky_whitley_co_(),
"KY Wilder" => new _ky_wilder(),
"KY Wilmore" => new _ky_wilmore(),
"KY Winchester" => new _ky_winchester(),
"KY Wolfe Co." => new _ky_wolfe_co_(),
"KY Woodford Co." => new _ky_woodford_co_(),
"LA" => new _la(),
"LA A" => new _la_a(),
"LA A1" => new _la_a1(),
"LA B" => new _la_b(),
"LA Married" => new _la_married(),
"LA Single" => new _la_single(),
"LA Unemployment" => new _la_unemployment(),
"LA Unemployment Wage Base" => new _la_unemployment_wage_base(),
"MA" => new _ma(),
"MA EMAC" => new _ma_emac(),
"MA EMAC Wage Base" => new _ma_emac_wage_base(),
"MA Exemption Factors" => new _ma_exemption_factors(),
"MA Family Leave" => new _ma_family_leave(),
"MA Family Leave Employer" => new _ma_family_leave_employer(),
"MA Family Leave Wage Base" => new _ma_family_leave_wage_base(),
"MA Medical Leave" => new _ma_medical_leave(),
"MA Medical Leave Employer" => new _ma_medical_leave_employer(),
"MA Medical Leave Employer Wage Base" => new _ma_medical_leave_employer_wage_base(),
"MA Medical Leave Wage Base" => new _ma_medical_leave_wage_base(),
"MA Medical and Family Leave" => new _ma_medical_and_family_leave(),
"MA Medical and Family Leave Employer" => new _ma_medical_and_family_leave_employer(),
"MA Medical and Family Leave Employer Wage Base" => new _ma_medical_and_family_leave_employer_wage_base(),
"MA Medical and Family Leave Wage Base" => new _ma_medical_and_family_leave_wage_base(),
"MA SE" => new _ma_se(),
"MA Taxable Income" => new _ma_taxable_income(),
"MA Unemployment" => new _ma_unemployment(),
"MA Unemployment Wage Base" => new _ma_unemployment_wage_base(),
"MA Workforce Training Fund" => new _ma_workforce_training_fund(),
"MA Workforce Training Fund Wage Base" => new _ma_workforce_training_fund_wage_base(),
"MD" => new _md(),
"MD Allegany Co." => new _md_allegany_co_(),
"MD Anne Arundel Co." => new _md_anne_arundel_co_(),
"MD Baltimore" => new _md_baltimore(),
"MD Baltimore Co." => new _md_baltimore_co_(),
"MD Calvert Co." => new _md_calvert_co_(),
"MD Caroline Co." => new _md_caroline_co_(),
"MD Carroll Co." => new _md_carroll_co_(),
"MD Cecil Co." => new _md_cecil_co_(),
"MD Charles Co." => new _md_charles_co_(),
"MD Delaware" => new _md_delaware(),
"MD Dorchester Co." => new _md_dorchester_co_(),
"MD Frederick Co." => new _md_frederick_co_(),
"MD Garrett Co." => new _md_garrett_co_(),
"MD Harford Co." => new _md_harford_co_(),
"MD Howard Co." => new _md_howard_co_(),
"MD Kent Co." => new _md_kent_co_(),
"MD Minimum Taxable Income" => new _md_minimum_taxable_income(),
"MD Montgomery Co." => new _md_montgomery_co_(),
"MD Nonresident" => new _md_nonresident(),
"MD Prince George's Co." => new _md_prince_george_s_co_(),
"MD Queen Anne's Co." => new _md_queen_anne_s_co_(),
"MD Somerset Co." => new _md_somerset_co_(),
"MD St. Mary's Co." => new _md_st__mary_s_co_(),
"MD Talbot Co." => new _md_talbot_co_(),
"MD Taxable Income" => new _md_taxable_income(),
"MD Unemployment" => new _md_unemployment(),
"MD Unemployment Wage Base" => new _md_unemployment_wage_base(),
"MD Washington Co." => new _md_washington_co_(),
"MD Wicomico Co." => new _md_wicomico_co_(),
"MD Worcester Co." => new _md_worcester_co_(),
"ME" => new _me(),
"ME Standard Deduction" => new _me_standard_deduction(),
"ME Taxable Income" => new _me_taxable_income(),
"ME Unemployment" => new _me_unemployment(),
"ME Unemployment Wage Base" => new _me_unemployment_wage_base(),
"MI" => new _mi(),
"MI Albion" => new _mi_albion(),
"MI Albion NR" => new _mi_albion_nr(),
"MI Battle Creek" => new _mi_battle_creek(),
"MI Battle Creek NR" => new _mi_battle_creek_nr(),
"MI Benton Harbor" => new _mi_benton_harbor(),
"MI Benton Harbor NR" => new _mi_benton_harbor_nr(),
"MI Big Rapids" => new _mi_big_rapids(),
"MI Big Rapids NR" => new _mi_big_rapids_nr(),
"MI Cities Nonresident" => new _mi_cities_nonresident(),
"MI Cities Resident" => new _mi_cities_resident(),
"MI Detroit" => new _mi_detroit(),
"MI Detroit NR" => new _mi_detroit_nr(),
"MI Detroit Nonresident" => new _mi_detroit_nonresident(),
"MI Detroit Resident" => new _mi_detroit_resident(),
"MI East Lansing" => new _mi_east_lansing(),
"MI East Lansing Nonresident" => new _mi_east_lansing_nonresident(),
"MI Flint" => new _mi_flint(),
"MI Flint NR" => new _mi_flint_nr(),
"MI Grand Rapids" => new _mi_grand_rapids(),
"MI Grand Rapids NR" => new _mi_grand_rapids_nr(),
"MI Grayling" => new _mi_grayling(),
"MI Grayling NR" => new _mi_grayling_nr(),
"MI Hamtramck" => new _mi_hamtramck(),
"MI Hamtramck NR" => new _mi_hamtramck_nr(),
"MI Highland Park" => new _mi_highland_park(),
"MI Highland Park NR" => new _mi_highland_park_nr(),
"MI Hudson" => new _mi_hudson(),
"MI Hudson NR" => new _mi_hudson_nr(),
"MI Ionia" => new _mi_ionia(),
"MI Ionia NR" => new _mi_ionia_nr(),
"MI Jackson" => new _mi_jackson(),
"MI Jackson NR" => new _mi_jackson_nr(),
"MI Lansing" => new _mi_lansing(),
"MI Lansing NR" => new _mi_lansing_nr(),
"MI Lapeer" => new _mi_lapeer(),
"MI Lapeer NR" => new _mi_lapeer_nr(),
"MI Muskegon" => new _mi_muskegon(),
"MI Muskegon Heights" => new _mi_muskegon_heights(),
"MI Muskegon Heights NR" => new _mi_muskegon_heights_nr(),
"MI Muskegon NR" => new _mi_muskegon_nr(),
"MI Pontiac" => new _mi_pontiac(),
"MI Pontiac NR" => new _mi_pontiac_nr(),
"MI Port Huron" => new _mi_port_huron(),
"MI Port Huron NR" => new _mi_port_huron_nr(),
"MI Portland" => new _mi_portland(),
"MI Portland NR" => new _mi_portland_nr(),
"MI Saginaw" => new _mi_saginaw(),
"MI Saginaw NR" => new _mi_saginaw_nr(),
"MI Springfield" => new _mi_springfield(),
"MI Springfield NR" => new _mi_springfield_nr(),
"MI Unemployment" => new _mi_unemployment(),
"MI Unemployment Wage Base" => new _mi_unemployment_wage_base(),
"MI Walker" => new _mi_walker(),
"MI Walker NR" => new _mi_walker_nr(),
"MN" => new _mn(),
"MN Taxable Income" => new _mn_taxable_income(),
"MN Unemployment" => new _mn_unemployment(),
"MN Unemployment Additional Assessment" => new _mn_unemployment_additional_assessment(),
"MN Unemployment Loan Assessment" => new _mn_unemployment_loan_assessment(),
"MN Unemployment Wage Base" => new _mn_unemployment_wage_base(),
"MN Unemployment Workforce Development Fee" => new _mn_unemployment_workforce_development_fee(),
"MO" => new _mo(),
"MO Kansas City" => new _mo_kansas_city(),
"MO St. Louis" => new _mo_st__louis(),
"MO St. Louis Employer" => new _mo_st__louis_employer(),
"MO Taxable Income" => new _mo_taxable_income(),
"MO Unemployment" => new _mo_unemployment(),
"MO Unemployment Wage Base" => new _mo_unemployment_wage_base(),
"MS" => new _ms(),
"MS Unemployment" => new _ms_unemployment(),
"MS Unemployment Wage Base" => new _ms_unemployment_wage_base(),
"MT" => new _mt(),
"MT Taxable Income" => new _mt_taxable_income(),
"MT Unemployment" => new _mt_unemployment(),
"MT Unemployment AFT" => new _mt_unemployment_aft(),
"MT Unemployment AFT Wage Base" => new _mt_unemployment_aft_wage_base(),
"MT Unemployment Wage Base" => new _mt_unemployment_wage_base(),
"Misc %" => new _misc__(),
"Misc Flat" => new _misc_flat(),
"NC" => new _nc(),
"NC Unemployment" => new _nc_unemployment(),
"NC Unemployment Wage Base" => new _nc_unemployment_wage_base(),
"ND" => new _nd(),
"ND Taxable Income" => new _nd_taxable_income(),
"ND Unemployment" => new _nd_unemployment(),
"ND Unemployment Wage Base" => new _nd_unemployment_wage_base(),
"NE" => new _ne(),
"NE Taxable Income" => new _ne_taxable_income(),
"NE Unemployment" => new _ne_unemployment(),
"NE Unemployment Wage Base" => new _ne_unemployment_wage_base(),
"NH" => new _nh(),
"NH Unemployment" => new _nh_unemployment(),
"NH Unemployment Administrative Charge" => new _nh_unemployment_administrative_charge(),
"NH Unemployment Wage Base" => new _nh_unemployment_wage_base(),
"NJ" => new _nj(),
"NJ Disability" => new _nj_disability(),
"NJ Disability Employer" => new _nj_disability_employer(),
"NJ Disability Employer Wage Base" => new _nj_disability_employer_wage_base(),
"NJ Disability Wage Base" => new _nj_disability_wage_base(),
"NJ Family Leave" => new _nj_family_leave(),
"NJ Family Leave Wage Base" => new _nj_family_leave_wage_base(),
"NJ Jersey City" => new _nj_jersey_city(),
"NJ Newark" => new _nj_newark(),
"NJ SUI/SDI" => new _nj_sui_sdi(),
"NJ TDI FLI Wage Base" => new _nj_tdi_fli_wage_base(),
"NJ Table A" => new _nj_table_a(),
"NJ Table B" => new _nj_table_b(),
"NJ Taxable Income" => new _nj_taxable_income(),
"NJ Unemployment" => new _nj_unemployment(),
"NJ Unemployment Employee" => new _nj_unemployment_employee(),
"NJ Unemployment Employee Wage Base" => new _nj_unemployment_employee_wage_base(),
"NJ Unemployment Wage Base" => new _nj_unemployment_wage_base(),
"NM" => new _nm(),
"NM Unemployment" => new _nm_unemployment(),
"NM Unemployment Wage Base" => new _nm_unemployment_wage_base(),
"NM Worker's Comp employer outstanding" => new _nm_worker_s_comp_employer_outstanding(),
"NM Worker's Comp outstanding" => new _nm_worker_s_comp_outstanding(),
"NM Worker's Compensation" => new _nm_worker_s_compensation(),
"NM Worker's Compensation Employer" => new _nm_worker_s_compensation_employer(),
"NV" => new _nv(),
"NV Unemployment" => new _nv_unemployment(),
"NV Unemployment CEP" => new _nv_unemployment_cep(),
"NV Unemployment Wage Base" => new _nv_unemployment_wage_base(),
"NY" => new _ny(),
"NY Disability" => new _ny_disability(),
"NY Disability YTDTax limit" => new _ny_disability_ytdtax_limit(),
"NY ECET" => new _ny_ecet(),
"NY Family Leave" => new _ny_family_leave(),
"NY Family Leave Wage Base" => new _ny_family_leave_wage_base(),
"NY Family Leave YTDTax limit" => new _ny_family_leave_ytdtax_limit(),
"NY MCTM" => new _ny_mctm(),
"NY MCTM NYC" => new _ny_mctm_nyc(),
"NY NJ Waterfront" => new _ny_nj_waterfront(),
"NY NYC Nonresident" => new _ny_nyc_nonresident(),
"NY NYC Resident" => new _ny_nyc_resident(),
"NY Re-employment Service Fund" => new _ny_re_employment_service_fund(),
"NY Re-employment Service Fund Wage Base" => new _ny_re_employment_service_fund_wage_base(),
"NY Taxable Income" => new _ny_taxable_income(),
"NY Unemployment" => new _ny_unemployment(),
"NY Unemployment Wage Base" => new _ny_unemployment_wage_base(),
"NY Yonkers Nonresident" => new _ny_yonkers_nonresident(),
"NY Yonkers Resident" => new _ny_yonkers_resident(),
"NY Yonkers Taxable Income" => new _ny_yonkers_taxable_income(),
"NYC Taxable Income" => new _nyc_taxable_income(),
"OH" => new _oh(),
"OH City" => new _oh_city(),
"OH SDIT" => new _oh_sdit(),
"OH Taxable Income" => new _oh_taxable_income(),
"OH Unemployment" => new _oh_unemployment(),
"OH Unemployment Wage Base" => new _oh_unemployment_wage_base(),
"OK" => new _ok(),
"OK Taxable Income" => new _ok_taxable_income(),
"OK Unemployment" => new _ok_unemployment(),
"OK Unemployment Technology Fund" => new _ok_unemployment_technology_fund(),
"OK Unemployment Wage Base" => new _ok_unemployment_wage_base(),
"OR" => new _or(),
"OR BASE" => new _or_base(),
"OR Canby" => new _or_canby(),
"OR Eugene" => new _or_eugene(),
"OR Eugene Employer" => new _or_eugene_employer(),
"OR FIT Allowance" => new _or_fit_allowance(),
"OR Family Leave" => new _or_family_leave(),
"OR Family Leave Employer" => new _or_family_leave_employer(),
"OR Family Leave Employer Wage Base" => new _or_family_leave_employer_wage_base(),
"OR Family Leave Wage Base" => new _or_family_leave_wage_base(),
"OR LTD" => new _or_ltd(),
"OR Multnomah Co." => new _or_multnomah_co_(),
"OR Multnomah YTD" => new _or_multnomah_ytd(),
"OR Portland Metro" => new _or_portland_metro(),
"OR Sandy" => new _or_sandy(),
"OR Single 3" => new _or_single_3(),
"OR South Clackamas" => new _or_south_clackamas(),
"OR Transit Tax" => new _or_transit_tax(),
"OR Tri-Met" => new _or_tri_met(),
"OR Unemployment" => new _or_unemployment(),
"OR Unemployment Wage Base" => new _or_unemployment_wage_base(),
"OR WBF" => new _or_wbf(),
"OR WBF Employer" => new _or_wbf_employer(),
"OR Wilsonville" => new _or_wilsonville(),
"PA" => new _pa(),
"PA EMST" => new _pa_emst(),
"PA Local" => new _pa_local(),
"PA OPT" => new _pa_opt(),
"PA Pittsburgh Payroll Expense" => new _pa_pittsburgh_payroll_expense(),
"PA Unemployment" => new _pa_unemployment(),
"PA Unemployment Employee" => new _pa_unemployment_employee(),
"PA Unemployment Wage Base" => new _pa_unemployment_wage_base(),
"PR" => new _pr(),
"PR Disability" => new _pr_disability(),
"PR Disability Employer" => new _pr_disability_employer(),
"PR Disability Employer Wage Base" => new _pr_disability_employer_wage_base(),
"PR Disability Wage Base" => new _pr_disability_wage_base(),
"PR Nonresident Alien" => new _pr_nonresident_alien(),
"PR Nonresident Citizen" => new _pr_nonresident_citizen(),
"PR Taxable Income" => new _pr_taxable_income(),
"PR Unemployment" => new _pr_unemployment(),
"PR Unemployment Wage Base" => new _pr_unemployment_wage_base(),
"RI" => new _ri(),
"RI Disability" => new _ri_disability(),
"RI Disability Wage Base" => new _ri_disability_wage_base(),
"RI Taxable Income" => new _ri_taxable_income(),
"RI Unemployment" => new _ri_unemployment(),
"RI Unemployment Job Development Fund" => new _ri_unemployment_job_development_fund(),
"RI Unemployment Job Development Fund Wage Base" => new _ri_unemployment_job_development_fund_wage_base(),
"RI Unemployment Wage Base" => new _ri_unemployment_wage_base(),
"SC" => new _sc(),
"SC Taxable Income" => new _sc_taxable_income(),
"SC Unemployment" => new _sc_unemployment(),
"SC Unemployment Wage Base" => new _sc_unemployment_wage_base(),
"SD" => new _sd(),
"SD Unemployment" => new _sd_unemployment(),
"SD Unemployment Wage Base" => new _sd_unemployment_wage_base(),
"SECA" => new _seca(),
"SECA Med" => new _seca_med(),
"SECA SS" => new _seca_ss(),
"SECA SS Wage Base" => new _seca_ss_wage_base(),
"TN" => new _tn(),
"TN Unemployment" => new _tn_unemployment(),
"TN Unemployment Wage Base" => new _tn_unemployment_wage_base(),
"TX" => new _tx(),
"TX Texarkana" => new _tx_texarkana(),
"TX Unemployment" => new _tx_unemployment(),
"TX Unemployment Wage Base" => new _tx_unemployment_wage_base(),
"UT" => new _ut(),
"UT Unemployment" => new _ut_unemployment(),
"UT Unemployment Wage Base" => new _ut_unemployment_wage_base(),
"VA" => new _va(),
"VA Taxable Income" => new _va_taxable_income(),
"VA Unemployment" => new _va_unemployment(),
"VA Unemployment Wage Base" => new _va_unemployment_wage_base(),
"VI" => new _vi(),
"VI Unemployment" => new _vi_unemployment(),
"VI Unemployment Wage Base" => new _vi_unemployment_wage_base(),
"VT" => new _vt(),
"VT Taxable Income" => new _vt_taxable_income(),
"VT Unemployment" => new _vt_unemployment(),
"VT Unemployment Wage Base" => new _vt_unemployment_wage_base(),
"WA" => new _wa(),
"WA EAF" => new _wa_eaf(),
"WA EAF Wage Base" => new _wa_eaf_wage_base(),
"WA FMLA" => new _wa_fmla(),
"WA FMLA Earnings Under Cap" => new _wa_fmla_earnings_under_cap(),
"WA FMLA Employer" => new _wa_fmla_employer(),
"WA FMLA Employer Wage Base" => new _wa_fmla_employer_wage_base(),
"WA FMLA Wage Base" => new _wa_fmla_wage_base(),
"WA LNI" => new _wa_lni(),
"WA LNI Employer" => new _wa_lni_employer(),
"WA Long Term Care" => new _wa_long_term_care(),
"WA Seattle" => new _wa_seattle(),
"WA Unemployment" => new _wa_unemployment(),
"WA Unemployment Wage Base" => new _wa_unemployment_wage_base(),
"WI" => new _wi(),
"WI Taxable Income" => new _wi_taxable_income(),
"WI Unemployment" => new _wi_unemployment(),
"WI Unemployment Wage Base" => new _wi_unemployment_wage_base(),
"WV" => new _wv(),
"WV Charleston" => new _wv_charleston(),
"WV Fairmont Nonresident" => new _wv_fairmont_nonresident(),
"WV Huntington" => new _wv_huntington(),
"WV Madison" => new _wv_madison(),
"WV Morgantown" => new _wv_morgantown(),
"WV Parkersburg" => new _wv_parkersburg(),
"WV Taxable Income" => new _wv_taxable_income(),
"WV Unemployment" => new _wv_unemployment(),
"WV Unemployment Debt Repayment" => new _wv_unemployment_debt_repayment(),
"WV Unemployment Wage Base" => new _wv_unemployment_wage_base(),
"WV Weirton" => new _wv_weirton(),
"WV Wheeling" => new _wv_wheeling(),
"WY" => new _wy(),
"WY Unemployment" => new _wy_unemployment(),
"WY Unemployment Wage Base" => new _wy_unemployment_wage_base(),
"WY Workers Comp" => new _wy_workers_comp(),
);?>